/* Qidfs_utils.h ******************
*  Part of QPeace software
*  See README file in QPeace distribution for copyright information
*
*  SJ Schwartz   (s.j.schwartz@qmw.ac.uk)
*  Queen Mary, University of London
*  26 November 2000
*  Apr 2001  Added Qan date routines
*  May 2001 Re-organised Qpeace
*  Added promote headers
*
*  Utility routines for generic used, including:
*
*  Setting strings, idfs lineages, etc.
*  Setting conversion tables, getting energy bin limits, ...
*  Opening and positioning idfs file
*  Manipulating time into format needed by and returned from idfs routines
*
*/


#ifndef QIDFS_UTILS_H
#define QIDFS_UTILS_H

#ifdef __cplusplus
extern "C"{
#endif

#include <stdio.h>
#include <string.h>
#include <math.h>
#include <stdlib.h>
#include <limits.h>
#include "ret_codes.h"
#include "user_defs.h"
#include "libVIDF.h"
#include "libbase_idfs.h"
#include "libdb.h"    
#include "libCfg.h" 
#include "libPromote.h"
#include "Qpeace_err_diag.h"  /* error codes, diagnostic print levels */


#define MAX_LEVELS 93           /* this is upper limit; less ok */
#define MAX_TABLES_TO_APPLY 20  /* 10 should be adequate */
#define SC_NAME_LEN 30          /* big enough to hold longest sc name */
#define MAX_PSD_UNITS_STR_LEN 100
#define PEACE_SPIN_PERIOD  4.0  /*rough spin per. in secs so can backup 1spin*/

#define LINEAGE_MAX_STR 100
#define DEG2RAD  0.01745329251994329547437168
#define RAD2DEG 57.29577951308232286464772187

#define N_IDFS_ERR 4
typedef enum IDFS_ERR_E{
  IDFS_OK=0,        //SUCCESS!!!
  IDFS_ERROR,       // generic
  IDFS_PROMOTE_ERR, // error promoting data
  IDFS_INIT_ERR     // initialisation error
 } IDFS_ERR;

//define lineages & sensors

#define BX_GSE 4
#define BY_GSE 5
#define BZ_GSE 6
#define U_PROBE_SC 15
#define EFW_CAA_SCPOT 0
#define SC1_LATITUDE 24
#define SC1_LONGITUDE 25
#define SC2_LATITUDE 26
#define SC2_LONGITUDE 27
#define SC3_LATITUDE 28
#define SC3_LONGITUDE 29
#define SC4_LATITUDE 30
#define SC4_LONGITUDE 31
#define SPIN_MODE 0
#define LMODE 1
#define HMODE 2
#define TSOURCE 0
#define TMODE 1
#define BSOURCE 2
#define BMODE 3
#define FILLVAL 3 

#define N_SENSOR_MODE 3
typedef enum SENSOR_MODE_E{
	DOUBLE_SENSOR_MODE=0,
	SINGLE_SENSOR_MODE_TOP,
	SINGLE_SENSOR_MODE_BOTTOM
	}SENSOR_MODE;


#define N_CPCMOMH_SENSORS 30
typedef enum CPCMOMH_SENSORS_E{
    LCOUNTS=0,  //0
    LDENSITY,     //1
    LVELOCITYX,  //2
    LVELOCITYY, //3
    LVELOCITYZ, //4
    LPXX,              //5
    LPYY,             //6
    LPZZ,             //7
    LPXY,            //8
    LPXZ,            //9
    LPYZ,            //10
    LQX,             //11
    LQY,             //12
    LQZ,             //13
    HCOUNTS,      //14
    HDENSITY,     //15
    HVELOCITYX, //16
    HVELOCITYY, //17
    HVELOCITYZ, //18
    HPXX,              //19
    HPYY,              //20
    HPZZ,              //21
    HPXY,              //22
    HPXZ,              //23
    HPYZ,              //24
    HQX,               //25
    HQY,              //26
    HQZ,               //27
    EBEG,                 //28
    EEND                  //29
 } CPCMOMH_SENSORS;


#define N_CPCMOMF_SENSORS 32
typedef enum CPCMOMF_SENSORS_E{
    TCOUNTS=0,  //0
    TDENSITY,     //1
    TVELOCITYX,  //2
    TVELOCITYY, //3
    TVELOCITYZ, //4
    TPXX,              //5
    TPYY,             //6
    TPZZ,             //7
    TPXY,            //8
    TPXZ,            //9
    TPYZ,            //10
    TQX,             //11
    TQY,             //12
    TQZ,             //13
    TEBEG,       //14
    TEEND,       //15
    BCOUNTS,  //16
    BDENSITY,     //17
    BVELOCITYX,  //18
    BVELOCITYY, //19
    BVELOCITYZ, //20
    BPXX,             //21
    BPYY,             //22
    BPZZ,             //23
    BPXY,            //24
    BPXZ,            //25
    BPYZ,            //26
    BQX,             //27
    BQY,             //28
    BQZ,             //29
    BEBEG,       //30
    BEEND       //31
 } CPCMOMF_SENSORS;    

#define N_CPCDM_SENSORS 5
typedef enum CPCDM_SENSORS_E{
    DBYTES = 0,
    HTIMMON,
    LTIMMON,
    SPOS,    // 3 This is for sunpulse
    TIMLOAD
} CPCDM_SENSORS;

typedef struct IDFS_LINEAGE_S{
   char project[LINEAGE_MAX_STR];
   char mission[LINEAGE_MAX_STR];
   char experiment[LINEAGE_MAX_STR];
   char instrument[LINEAGE_MAX_STR];
   char virtual_inst[LINEAGE_MAX_STR];
   char exten[LINEAGE_MAX_STR];/* this and vsn aren't  part of the lineage*/
   SDDAS_USHORT vsn;      /* but it's useful to carry them around here */
} IDFS_LINEAGE;

typedef enum SC_E{
  CLUSTER1,
  CLUSTER2,
  CLUSTER3,
  CLUSTER4,
  C1234
} SC;

#define N_SC 5

typedef enum PSD_UNITS_E{
   PSD_UNITS_CNTS_ACCUM,
   PSD_UNITS_DIFF_ENERGY_FLUX,
   PSD_UNITS_S3_PER_KM6,
   UNKNOWN_PSD_UNITS
} PSD_UNITS;

#define N_PSD_UNITS 3   /* don't count UNKNOWN */

typedef enum BIN_UNITS_E{
   BIN_UNITS_EV,
   BIN_UNITS_KM_S
} BIN_UNITS;

#define N_BIN_UNITS 2

typedef enum INTERP_QUANTITY_E{
   DIFFERENTIAL,
   INTEGRAL
} INTERP_QUANTITY;

typedef enum PHI_ZERO_E{
   PEACE_TZERO,
   UNKNOWN
} PHI_ZERO_FRAME;

typedef enum IDFS_TIME_COMPARE_E{
  T2_LATER   = -1,
  SAME_TIME,
  T1_LATER
} IDFS_TIME_COMPARE;

typedef struct  IDFS_TIME_S{
   SDDAS_SHORT year;
   SDDAS_SHORT doy;
   SDDAS_LONG  sec;
   SDDAS_LONG  nano;
}  IDFS_TIME;

/* prototpyes */


/**********************************************************************
****  IDFS Time  ******************************************************
**********************************************************************/

void to_idfs_time(int year, int doy, int hr, int min, int sec, long nano, IDFS_TIME *idftp);

void sec_nano_to_msec_nano(long sec, long nanos, long *msec, long *nanoms);

void msec_nano_to_sec_nano(long msec, long nanoms, long *sec, long *nanos);

void to_hms_time(IDFS_TIME *idftp, int *hp, int *mp, int *sp, long *np);

void print_idfs_time(IDFS_TIME *idftp);

IDFS_TIME t1_minus_t2(IDFS_TIME *t1, IDFS_TIME *t2);

IDFS_TIME_COMPARE compare_idfs_time(IDFS_TIME *t1, IDFS_TIME *t2);

void QanYearDoyToDate( int year, int doy, int* month, int* day );

void QanDateToYearDoy(int year, int month, int day, int *doy);

double QEpochSecs( int year, int month, int day,
  int hour, int minute, int second, int msec );
  
double idfs_time_to_epoch_secs(IDFS_TIME *idft);

void epoch_secs_to_idfs_time(double epoch, IDFS_TIME *idftp);

void add_secs_to_idfs_time(IDFS_TIME *idftp, float secs2add);

void set_times_from_idf(IDFS_TIME *btime, IDFS_TIME *etime,
                        IDFS_TIME *sun_sen, struct idf_data *idfp);

/*********************************************************************
****** IDFS File and data structure utils ****************************
*********************************************************************/

SDDAS_SHORT init_idf_file(IDFS_TIME btime, IDFS_TIME etime, SDDAS_ULONG dkey,
     struct idf_data *idfp, SDDAS_USHORT vsn);



#ifdef __cplusplus
}
#endif

#endif  /* end ifndef _QIDFS_UTILS */
