
#ifndef TMC_DATA_H
#define TMC_DATA_H

#include "tmc_routines.h"

  /*

  class tmc_data contains all the raw sequences,
  intermediate sequences and corrected sequences.

  member functions allow manipulation of a tmc_data object
  to take raw on-board moments and supplementary data and solve
  for s/c potential and truncation.

  updated July 2007 to add cpcdm and sunaz bits - sjs

  */
  
class tmc_data
{
public: tmc_data(int SC, bool B, bool L, bool H, bool T) //constructor
	      {

	      spacecraft = SC;
	      bottom = B;
	      leea_overlap = L;
	      heea_overlap = H;
	      top = T;
	      ON_BOARD_FACTOR = 2.0;

	      };

        //member fns
        SENSOR_MODE determine_sensor_mode(int n);
        double get_raw_N(double rN, double Ne);
	      double get_raw_T(double rT, double Te);
	      double get_raw_V(double rV, double Ve);
	      void calculate_estimation_ratios(TRUNC_STRUCT *tsr, double *rN, double *rV, double *rT, double *Ie);
	      void set_lineages();
	      void populate_low_energy_moments(SENSOR_MODE m, double *tmc_EL, double *tmc_EU, double *tmc_N, double tmc_V[3], double tmc_P[3][3], double leea_gain_correction_factor, double heea_gain_correction_factor, double phi_offset, int nIndex, int heea_or_leea, double *tcf, double *bcf, double *ocf);
	      void populate_high_energy_moments(double *tmc_top_N, double tmc_top_v[3], double tmc_top_P[3][3], double tcf, int nIndex);
	      void populate_totally_raw_moments(SENSOR_MODE m, double *raw_N, double raw_v[3], double raw_P[3][3], int nIndex, int heea_or_leea, double leea_gain_correction_factor, double heea_gain_correction_factor, double phi_offset);
	      void get_leea_calibration_info(string name);
	      void get_heea_calibration_info(string name);
	      void fiddle_gain_factors(const char *ctime, double *leea_gf, double *heea_gf);
	      void mask_series(QdRScalarSeq_var in, QdObject_var A, QdObject_var B, SequenceMask mask);
	      void extract_half_spin( QdRScalarSeq_var in, SequenceMask mask, QdRScalarSeq_var out);
	      void add(int nSize, QdRScalarSeq_var qdObj1, QdRScalarSeq_var qdObj2, QdRScalarSeq_var qdOut, int add_or_pass);
	      void null_all();
	      TMC_ERROR get_data(IDFS_TIME start, IDFS_TIME end, bool constant_phi, double phi);
	      QdObject_var check_for_non_conformity(QdObject_var in);
	      TMC_ERROR separate_and_sum_overlap();
	      TMC_ERROR join_all();
	      TMC_ERROR set_up_outputs();
	      TMC_ERROR clean_outputs();
	      TMC_ERROR correct_moments(double lower, double upper, double step, double delta_phi, string *destination, string *logname, bool logging, bool clobber, double const_sc_pot);
	      TMC_ERROR join_a_to_b(QdObjectSeq_var a, QdObject_var b, QdObjectSeq_var aj, int count);
	      TMC_ERROR  shift_time_series( QdObject_var  rTobeShifted );
	      TMC_ERROR  shift_all_time_series( QdObjectSeq_var  inputs, int nSeq );
	      bool use_B_moments(double U, double EL, double phi_offset, double value);
	      bool use_T_moments(double value);
	      bool use_B_moments_at_all(double BN);
	      bool use_T_moments_at_all(QdObjectSeq_var inputs);
	      double gain_correction(const char * ctime, QdRScalarSeq_var calibration);
	      void determine_limits_and_factors(SENSOR_MODE m, double phi_offset, int heea_or_leea, double *EL, double *EU, double *ocf, double *tcf, double *bcf, int n, double leea_gain_correction_factor, double heea_gain_correction_factor);
        void handover_params(TRUNC_STRUCT *s, double v0[3], double auxlat, double auxlong, double sunaz, int N);
	      double* get_phi_sc(int i) const;
	      //access methods
	      double get_j_top_Pij(int i, int j, int N) const;
	      double get_j_bottom_Pij(int i, int j, int N) const;
	      double get_j_ss_overlap_Pij(int i, int j, int N) const;
	      double * get_Ne(int i) const;
	      double * get_Vx(int i) const;
	      double * get_Vy(int i) const;
	      double * get_Vz(int i) const;
	      double * get_VT(int i) const;
        double * get_Te(int i) const;
      	double * get_Te_perp(int i) const;
      	double * get_Te_para(int i) const;
       	int * get_err(int i) const;
	      double * get_p_off_on(int i) const;
	      double * get_p_perp_ratio(int i) const;
	      double * get_rT(int i) const;
	      double * get_rN(int i) const;
	      double * get_rV(int i) const;
	      double * get_Ie(int i) const;
	      int get_size() const{ return cNe->sequence_size(); };
        int get_joined_size() const{ return j_ss_overlap_N->sequence_size(); };
	      const char * current_time(int i) const;
	      const char * current_time_calc(int i) const;
        void get_command_history();
        const char * current_time_from_seq(int i, QdRScalarSeq_var seq) const;
	      ~tmc_data(); //destructor

private:
        //these define the moments -- what s/c and what sum?
        int spacecraft;
	      bool bottom;
	      bool leea_overlap;
	      bool heea_overlap;
	      bool top;

        //obf == 2
	      double ON_BOARD_FACTOR;

        //lineages
	      IDFS_LINEAGE cpcmomf; //full moments B,T
	      IDFS_LINEAGE cpcmomh; //half moments overlap
	      IDFS_LINEAGE fgm;     //fgm b-vector
	      IDFS_LINEAGE efw;     //efw u_probe_sc
	      IDFS_LINEAGE efw_caa; //efw caa negative sc potential
	      IDFS_LINEAGE aux;     //aux data (attitude)
              IDFS_LINEAGE cpcdm;   // core for spos offset

	      //calibration info
        QdRScalarSeq_var leea_gain_correction;
	      QdRScalarSeq_var heea_gain_correction;


	      //the raw moments -- overlap
	      QdRScalarSeq_var overlap_N;
	      QdRScalarSeq_var overlap_Vx;
	      QdRScalarSeq_var overlap_Vy;
	      QdRScalarSeq_var overlap_Vz;
	      QdRScalarSeq_var overlap_Pxx;
	      QdRScalarSeq_var overlap_Pxy;
	      QdRScalarSeq_var overlap_Pxz;
	      QdRScalarSeq_var overlap_Pyy;
	      QdRScalarSeq_var overlap_Pyz;
	      QdRScalarSeq_var overlap_Pzz;
	      QdRScalarSeq_var overlap_EL;
	      QdRScalarSeq_var overlap_EU;
	      QdRScalarSeq_var overlap_spinmode;
        QdRScalarSeq_var overlap_energymode;
	
	      //subsetted and summed      
        QdRScalarSeq_var ss_overlap_N;
	      QdRScalarSeq_var ss_overlap_Vx;
	      QdRScalarSeq_var ss_overlap_Vy;
	      QdRScalarSeq_var ss_overlap_Vz;
	      QdRScalarSeq_var ss_overlap_Pxx;
	      QdRScalarSeq_var ss_overlap_Pxy;
	      QdRScalarSeq_var ss_overlap_Pxz;
	      QdRScalarSeq_var ss_overlap_Pyy;
	      QdRScalarSeq_var ss_overlap_Pyz;
	      QdRScalarSeq_var ss_overlap_Pzz;
	      QdRScalarSeq_var ss_overlap_EL;
	      QdRScalarSeq_var ss_overlap_EU;
	      QdRScalarSeq_var ss_overlap_energymode;

	      //the raw moments -- bottom and top
	      QdRScalarSeq_var bottom_N;
	      QdRScalarSeq_var bottom_Vx;
	      QdRScalarSeq_var bottom_Vy;
	      QdRScalarSeq_var bottom_Vz;
	      QdRScalarSeq_var bottom_Pxx;
	      QdRScalarSeq_var bottom_Pxy;
	      QdRScalarSeq_var bottom_Pxz;
	      QdRScalarSeq_var bottom_Pyy;
	      QdRScalarSeq_var bottom_Pyz;
	      QdRScalarSeq_var bottom_Pzz;
	      QdRScalarSeq_var bottom_EL;
	      QdRScalarSeq_var bottom_EU;
	      QdRScalarSeq_var bottom_energymode;
	      QdRScalarSeq_var bottom_source;

	      QdRScalarSeq_var top_N;
	      QdRScalarSeq_var top_Vx;
	      QdRScalarSeq_var top_Vy;
	      QdRScalarSeq_var top_Vz;
	      QdRScalarSeq_var top_Pxx;
	      QdRScalarSeq_var top_Pxy;
	      QdRScalarSeq_var top_Pxz;
	      QdRScalarSeq_var top_Pyy;
	      QdRScalarSeq_var top_Pyz;
	      QdRScalarSeq_var top_Pzz;
	      QdRScalarSeq_var top_EL;
	      QdRScalarSeq_var top_EU;
	      QdRScalarSeq_var top_energymode;
	      QdRScalarSeq_var top_source;

	      //other data we need
	      QdRMatrixSeq_var fgm_B_vec;
	      QdRScalarSeq_var efw_U_probe_sc;
	      QdRScalarSeq_var aux_latitude;
	      QdRScalarSeq_var aux_longitude;
	      QdRScalarSeq_var cpcdm_spos;

	      //joined data
	      QdRScalarSeq_var j_ss_overlap_N;
	      QdRScalarSeq_var j_ss_overlap_Vx;
	      QdRScalarSeq_var j_ss_overlap_Vy;
	      QdRScalarSeq_var j_ss_overlap_Vz;
	      QdRScalarSeq_var j_ss_overlap_Pxx;
	      QdRScalarSeq_var j_ss_overlap_Pxy;
	      QdRScalarSeq_var j_ss_overlap_Pxz;
	      QdRScalarSeq_var j_ss_overlap_Pyy;
	      QdRScalarSeq_var j_ss_overlap_Pyz;
	      QdRScalarSeq_var j_ss_overlap_Pzz;
	      QdRScalarSeq_var j_ss_overlap_EL;
	      QdRScalarSeq_var j_ss_overlap_EU;
	      QdRScalarSeq_var j_ss_overlap_energymode;

	      QdRScalarSeq_var j_bottom_N;
	      QdRScalarSeq_var j_bottom_Vx;
	      QdRScalarSeq_var j_bottom_Vy;
	      QdRScalarSeq_var j_bottom_Vz;
	      QdRScalarSeq_var j_bottom_Pxx;
	      QdRScalarSeq_var j_bottom_Pxy;
	      QdRScalarSeq_var j_bottom_Pxz;
	      QdRScalarSeq_var j_bottom_Pyy;
	      QdRScalarSeq_var j_bottom_Pyz;
	      QdRScalarSeq_var j_bottom_Pzz;
	      QdRScalarSeq_var j_bottom_EL;
	      QdRScalarSeq_var j_bottom_EU;
	      QdRScalarSeq_var j_bottom_energymode;
	      QdRScalarSeq_var j_bottom_source;

	      QdRScalarSeq_var j_top_N;
	      QdRScalarSeq_var j_top_Vx;
	      QdRScalarSeq_var j_top_Vy;
	      QdRScalarSeq_var j_top_Vz;
	      QdRScalarSeq_var j_top_Pxx;
	      QdRScalarSeq_var j_top_Pxy;
	      QdRScalarSeq_var j_top_Pxz;
	      QdRScalarSeq_var j_top_Pyy;
	      QdRScalarSeq_var j_top_Pyz;
	      QdRScalarSeq_var j_top_Pzz;
	      QdRScalarSeq_var j_top_EL;
	      QdRScalarSeq_var j_top_EU;
	      QdRScalarSeq_var j_top_energymode;
	      QdRScalarSeq_var j_top_source;

	      //other data we need - joined
	      QdRMatrixSeq_var j_fgm_B_vec;
	      QdRScalarSeq_var j_efw_U_probe_sc;
	      QdRScalarSeq_var j_aux_latitude;
	      QdRScalarSeq_var j_aux_longitude;
	      QdRScalarSeq_var j_cpcdm_spos;

	      //corrected moments & output data
	      QdRScalarSeq_var Ne;
	      QdRScalarSeq_var Te;
	      QdRScalarSeq_var Te_perp;
        QdRScalarSeq_var Te_para;
	      QdRMatrixSeq_var Ve;
	      QdRScalarSeq_var VT;
	      QdRScalarSeq_var rN;
        QdRScalarSeq_var rV;
	      QdRScalarSeq_var rT;
      	QdRScalarSeq_var err;
        QdRScalarSeq_var Ie;
	      QdRScalarSeq_var p_off_on;	      
	      QdRScalarSeq_var p_perp_ratio;
	      
	      //clean corrected moments & output data
	      QdRScalarSeq_var cNe;
	      QdRScalarSeq_var cTe;
	      QdRScalarSeq_var cTe_perp;
        QdRScalarSeq_var cTe_para;
	      QdRMatrixSeq_var cVe;
	      QdRScalarSeq_var cVT;
	      QdRScalarSeq_var crN;
        QdRScalarSeq_var crV;
	      QdRScalarSeq_var crT;
      	QdRScalarSeq_var cerr;
        QdRScalarSeq_var cIe;
	      QdRScalarSeq_var cp_off_on;	      
	      QdRScalarSeq_var cp_perp_ratio;
        QdRScalarSeq_var c_phi_sc;
	      
	   };      

#endif 
