
#ifndef TRUNCATED_MOMENTS_H
#define TRUNCATED_MOMENTS_H

#ifdef __cplusplus
extern "C"{
#endif

#include <stdio.h>
#include <stdlib.h>
#include <math.h>

#define Q_ELECTRON         1.6021e-19     /* Coulombs */
#define M_ELECTRON         9.1091e-31     /* kg       */
#define K_BOLTZ            1.38054e-23    /* J/K      */
#define PI                 3.14159265359
#define ONE_OVER_SQRT_PI   0.56418958354776

#define NTRIES 50       /* nonlinear iteration tries */
#define DIAG_TMNEWT 1
#define EPS 1.0e-7      /* Approximate square root of the machine precision.
                          NB In double precision there is much more than this!*/

/* tmnewt defs */
#define NR_END 1 
#define FREE_ARG char*
#define TINY 1.0e-20    /*A small number.*/ 

/* tmnewt error codes */

#define NR_OK 0
#define NR_ERROR 1
#define TMNEWT_LU_ERR 2
#define TMNEWT_TRIALS_EXHAUSTED 3


 
/* Error Codes */

#define N_TRUNC_ERR 10
typedef enum TRUNC_ERR_E{
  TRUNC_OK=0,
  TRUNC_ERROR,     /* generic error */
  TRUNC_TMNEWT_LU_ERR,
  TRUNC_TMNEWT_TRIALS_EXHAUSTED,
  TRUNC_VLVU_RANGE_ERROR,
  TRUNC_VLVU_LT_VMIN,
  TRUNC_OK_GUESSES, 
  TRUNC_GUESSES_EXHAUSTED, 
  TRUNC_NEG_P_DIAG,
  TRUNC_NO_B_MOMS
 } TRUNC_ERR;

/* corresponding error strings  - see truncated_moments_corrector.c */
const char *trunc_err_messages(TRUNC_ERR terr);  /* prototype */

typedef enum TBOOL_E{
  TFALSE = 0,
  TTRUE
} TBOOL;


typedef struct TRUNC_STRUCT_S{
/* structure to pass 'round everything! */

/* physical quantities - Maxwellian                                        */
    double n0;        /* density, in particles/m^3                         */
    double v0;        /* magnitude of velocity in m/s                      */
    double vT;        /* thermal speed in m/s = sqrt (2kT0/me)             */
    double T0;        /* temperature in K                                  */

/* physical quantities - measured moments and parameters                   */
    double vl;       /* lower cutoff of integration in m/s                 */
    double vu;       /* upper cutoff of integration in m/s                 */
    double phi_sc;   /* spacecraft potential in eV                         */
    double Nsc;      /* density in particles/m^3                           */
    double vxsc;     /* x-component of velocity - s/c coords               */
    double vysc;     /* y-component "                                      */
    double vzsc;     /* z-component "                                      */  
    double Psc[3][3];/* stress tensor in s/c coords                        */
/* alternative forms as needed by calculation                              */
    double Vsc;      /* magnitude of velocity - s/c coords [Vzsc in notes] */
    double trPsc;    /* trace of stress tensor                             */
    double VTsc;     /* thermal velocity in m/s = sqrt {
                          2/3 [ trPsc/Nsc me ] - 2/3 [ Vsc^2 ]  } 
			  = sqrt (2kT/m)                                   */

/* intermediary quantities                                                 */
    double curlyE;   /* negative sc pot in vel units: -2ephi_sc/m_e        */
    double vL;       /* lower, upper cutoff corrected for s/c potential    */
    double vU;       /* vL,U^2 = vl,u^2 + curlyE  ; ensure > 0 !!          */

/* normalised quantities                                                   */
    double zeta_sc;  /*magnitude of measured bulk velocity normalised to   */
                     /*    measured thermal speed VTsc, i.e. Vsc/VTsc      */
    double vl_norm;  /*speed in free space arriving at detector at lower   */
                     /*    cutoff and normalised to VTsc                   */
    double vu_norm;  /* ditto for upper limit                              */
    double eps_sc;   /* negative sc potential in velocity units normalised */
                     /*    to VTsc^2                                       */
    double eta;      /* n0/Nsc / sqrt(pi) v0_norm vt_norm                  */
    double v0_norm;  /* v0/Vsc                                             */
    double vt_norm;  /* vT/VTsc                                            */
} TRUNC_STRUCT;

/* prototypes */
double split_tqgauss100(int N, double (*func)(double, double, double, double, double),
     double a,double b, double epssc, double v0, double vt_sq, double zeta);
int tmnewt(int ntrial, double x[], int n, double tolx, double tolf,TRUNC_STRUCT  *tstruct);                   /* adapted NR nonlin root */
void tusrfun( double *x, int n, double *fvec, double **fjac,TRUNC_STRUCT  *tstruct);                  /* nonlin eqns + jacob */
void tvecfun( int n, double *x, double *fvec, TRUNC_STRUCT  *tstruct);
                                                /* nonline eqns        */
void tfdjac(int n, double x[], double fvec[], double **df, void (*tvecfunc)(int, double [], double [],TRUNC_STRUCT *), TRUNC_STRUCT  *tstruct );                 /* numerical jacobian */

void lubksb(double **a, int n, int *indx, double b[]); /* lu decomp */
int ludcmp(double **a, int n, int *indx, double *d);   /* ditto     */
double *dvector(long nl, long nh);                     /* NR vecs, etc. */
int *ivector(long nl, long nh); 
double **dmatrix(long nrl, long nrh, long ncl, long nch);
void free_ivector(int *v, long nl, long nh);
void free_dvector(double *v, long nl, long nh);
void free_dmatrix(double **m, long nrl, long nrh, long ncl, long nch); 
void nrerror(char error_text[]);

double ev2m_per_s(double Eev);
double tqgauss100(double (*func)(double, double, double, double, double), double a, double b, double epssc, double v0, double vt, double zeta);
double integrand1(double v, double epssc, double v0, double vt, double zeta);
double integrand2(double v, double epssc, double v0, double vt, double zeta);
double integrand3(double v, double epssc, double v0, double vt, double zeta);
void transform_to_gse_tensor(double latitude, double longitude, double sunaz,  double Q[3][3]);
void transform_to_gse_vector(double latitude, double longitude, double sunaz,  double Q[3]);
void eplusminusfun(double v, double v0, double zeta, double vt, double *eplus, double *eminus);
void calc_moments_from_maxwellian(TRUNC_STRUCT *tstr);
void calc_maxwellian_nl_params(TRUNC_STRUCT *t, double *etatilda, double *v0tilda, double *vTtilda);
void calc_intermediary(TRUNC_STRUCT *tstr);
void prepare_normalised_nl_inputs(TRUNC_STRUCT *tstr);
void dimensional_from_nl_solution(TRUNC_STRUCT *tstr);
void set_defaults_tstruct(TRUNC_STRUCT *tstr);
void print_trunc_struct(TRUNC_STRUCT *tstr, TBOOL headings);
void guess(double v[], double guess_one, double guess_two, double guess_three);
void initialize_tsr(double TR[3][3], double x);
void sc_to_gse(double TR[3][3], double latitude, double longitude, double sunaz, int output);
void set_Nsc(TRUNC_STRUCT *tstr, double N);
void set_vl(TRUNC_STRUCT *tstr, double E_l);
void set_vu(TRUNC_STRUCT *tstr, double E_u);
void set_phi_sc(TRUNC_STRUCT *tstr, double U);
void set_vxsc(TRUNC_STRUCT *tstr, double vx);
void set_vysc(TRUNC_STRUCT *tstr, double vy);
void set_vzsc(TRUNC_STRUCT *tstr, double vz);
void set_Psc(TRUNC_STRUCT *tstr, double Psr[3][3]);
void set_Vsc(TRUNC_STRUCT *tstr);
void set_VTsc(TRUNC_STRUCT *tstr);
double estimate_anisotropy_T0_perp(TRUNC_STRUCT *tstr, double b_gse[3], double P_gse[3][3], double v_gse[3]);
double estimate_anisotropy_T0_para(TRUNC_STRUCT *tstr, double b_gse[3], double P_gse[3][3], double v_gse[3]);
double trace(double P[3][3]);
double get_p_component(TRUNC_STRUCT *tstr, int i, int j);
void print_tensor(double T[3][3], int output);
TRUNC_ERR normalised_nl_solver(TRUNC_STRUCT *tstruct, double ga, double gb, double gc);
TRUNC_ERR check_nl_inputs(TRUNC_STRUCT *tstruct);
TRUNC_ERR check_p_diagonal(TRUNC_STRUCT *tstr);
void check_physics(TRUNC_STRUCT *tstr);
double lower_energy_lookup(int lookupindex, int mode, int craft, int LEEA_or_HEEA);
double upper_energy_lookup(int lookupindex, int mode, int craft, int LEEA_or_HEEA);
double x_cross_y(double x[3], double y[3], double n[3]);
void special_transform_tensor(double v[3], double b[3], double T[3][3]);
void special_transform_vector(double v[3], double b[3]);
void p_special_transform(TRUNC_STRUCT *tsr, double b_gse[3], double P_gse[3][3],double v_gse[3], double p[3][3]);
void copy_trunc_struct_maxwellians(TRUNC_STRUCT *from, TRUNC_STRUCT *to);

#ifdef __cplusplus
}
#endif

#endif  /* define _TRUNCATED_MOMENTS */



