/* QTMC_idfs_utils.c ******************
*  Part of QPeace software
*  See README file in QPeace distribution for copyright information
*
*  This is a subset of Qidfs_utils prepared for QTMC.
*
*  SJ Schwartz   (s.j.schwartz@qmw.ac.uk)
*  Queen Mary, University of London
*  26 November 2000
*  6 March 2001 added set_times_from_idf routine
*  May 2001 - re-organised Qpeace routines
*  July 2001 - added stuff for sweep export
*  May 2002  - tidied PSD_UNITS_STRING with statics for safety
*  July 2007 - updated and change name to QTMC_...
*
*  Utility routines for generic used, including:
*
*  Setting strings, idfs lineages, etc.
*  Setting conversion tables, getting energy bin limits, ...
*  Opeining and positioning idfs file
*  Manipulating time into format needed by and returned from idfs routines
*
*/



#include "QTMC_idfs_utils.h"


/**********************************************************************
****  IDFS Time  ******************************************************
**********************************************************************/

/**********************************************************************/
void to_idfs_time(int year, int doy, int hr,
               int min, int sec, long nano, IDFS_TIME *idftp){
/*************************************************************************/
/* convert a time to format expected by idfs routines */
/* at present just a dumb routine which takes int
* year, doy, hr, min, sec*/

idftp->year = year;
idftp->doy  = doy;
idftp->sec  = 60*(60*hr + min)+sec;
idftp->nano = nano;

return;
}

/*****************************************************************************/
void sec_nano_to_msec_nano(long sec, long nanos, long *msec, long *nanoms){
/*****************************************************************************/

*msec =1000*sec + (nanos/1000000);
*nanoms = nanos % 1000000;
return;
}

/***************************************************************************/
void msec_nano_to_sec_nano(long msec, long nanoms, long *sec, long *nanos){
/**************************************************************************/

*sec = msec/1000;
*nanos = 1000000*(msec % 1000) + nanoms;
return;
}

/***************************************************************/
void to_hms_time(IDFS_TIME *idftp, int *hp, int *mp, int *sp, long *np){
/***************************************************************/
/* and an inverse dumb routine. Will use QDOS Time soon */
*hp = idftp->sec/3600;
*mp = idftp->sec/60 - *hp*60;
*sp = idftp->sec - 60*(*mp + 60**hp);
*np = idftp->nano;

return ;
}

void print_idfs_time(IDFS_TIME *idftp){
int hr, min, sec;
long nano;
to_hms_time(idftp, &hr, &min, &sec, &nano);
printf("%4d/%3d %2d:%2d:%2d (%ld secs in day) and %ld nanosecs",
         idftp->year, idftp->doy, hr, min, sec, idftp->sec, nano);
return;
}

/******************************************************************/
IDFS_TIME t1_minus_t2(IDFS_TIME *t1, IDFS_TIME *t2){
/******************************************************************/

/* find difference between two IDFS_TIMES
*  since no field can overflow, there is no
*  need to do anything other than difference the
*  independent members
*******************************************/

IDFS_TIME dt_idfs, *dtp=&dt_idfs;

dtp->year = t1->year - t2->year;
dtp->doy  = t1->doy  - t2->doy;
dtp->sec  = t1->sec  - t2->sec;
dtp->nano = t1->nano - t2->nano;

return  dt_idfs;
}

/******************************************************************/
IDFS_TIME_COMPARE compare_idfs_time(IDFS_TIME *t1, IDFS_TIME *t2){
/******************************************************************/

IDFS_TIME dt_idfs;

dt_idfs = t1_minus_t2(t1,t2);

if (dt_idfs.year > 0) return T1_LATER;
   else if (dt_idfs.year < 0 ) return T2_LATER;
/* years same; check days */
if (dt_idfs.doy > 0) return T1_LATER;
   else if (dt_idfs.doy < 0 ) return T2_LATER;
/* days same; check seconds */
if (dt_idfs.sec > 0) return T1_LATER;
   else if (dt_idfs.sec < 0 ) return T2_LATER;
/* secs same; check nanosecs */
if (dt_idfs.nano > 0) return T1_LATER;
   else if (dt_idfs.nano < 0 ) return T2_LATER;
/* nanos same; must be same time */

return SAME_TIME;
}

/* Qan-based routines follow. SJS added () around leap year && || ops
*  and constructed QanDateToYearDoy from QanEpochToYearDoy */ 


/* daytab: number of days in month for normal and leap years 
**         index 1 is Jan 
*/

int qidf_days_in_month[2][13] = {
  { 0, 31, 28, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31 },
  { 0, 31, 29, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31 }
};


void QanYearDoyToDate( int year, int doy, int* month, int* day )
{
  int i, leap;
  leap = ( (year%4 == 0 && year%100 != 0) || year%400 == 0 );
/*
** check that doy is in range for supplied year
*/
  if( (leap && doy > 366 ) || (!leap && doy > 365) || doy < 1 )
  {
    *month = -1; *day = -1;
    return;
  }
  
  for(i=1; doy > qidf_days_in_month[leap][i]; i++ )
     doy -= qidf_days_in_month[leap][i];
  *month = i;
  *day = doy;
}


void QanDateToYearDoy(int year, int month, int day, int *doy)
{
  int leap_year, i;
  leap_year = (( year % 4 == 0 ) && ( year % 100!=0 )) || ( year % 400==0 );
  if (month < 1 || month > 12){
     *doy = -1;
     return;
  }
  for( *doy=0, i=0; i < month ; i++)
    *doy += qidf_days_in_month[leap_year][i];
  *doy += day;
}


/*********************************************************************
****** IDFS File and data structure utils ****************************
*********************************************************************/

/**********************************************************************/
SDDAS_SHORT init_idf_file(IDFS_TIME btime, IDFS_TIME etime, SDDAS_ULONG dkey,
     struct idf_data *idfp, SDDAS_USHORT vsn){
/*********************************************************************/
/* open and position an idfs file */
SDDAS_SHORT ret_val;
SDDAS_CHAR exten[5];

strcpy(exten,"");

ret_val = file_open(dkey,exten, vsn,btime.year,btime.doy, btime.sec, btime.nano,
              etime.year, etime.doy, etime.sec, etime.nano, 0);
if (ret_val != ALL_OKAY){
   printf ("PEACE: Error opening idfs file\n"); return QUTILS_ERROR;
}

ret_val = file_pos ( dkey,exten, vsn,idfp,btime.year, btime.doy, btime.sec, btime.nano,
              etime.year,etime.doy, etime.sec, etime.nano);
if (ret_val != ALL_OKAY){
   printf ("PEACE: Error positioning idfs file\n"); return QUTILS_ERROR;
}

if(DIAG_PRINT){
   printf("init_idf_file ok. Start and stop times sent to file_pos are\n");
   print_idfs_time(&btime);printf("\n");
   print_idfs_time(&etime);printf("\n");
}

return ALL_OKAY;
}

