//  CDFDataConfigurationMap.cc
//
//  Qdos2 - Data Configuration Map
//
//  Janet Barnes
//  Dec 1999
//

#include "CDFDataConfigurationMap.h"
#include "StringRep.h"
#include "StringTokenizer.h"
#include "StringMem.h"


namespace QDOS {

CDF_DCMFile::CDF_DCMFile( CDFDataConfigurationMap* data_map, string& file_name,
    CDF_DCMAliasTable& alias_table, CDF_DCMFileStack& open_files,\
    double data_priority):_data_map(data_map),_file_name(file_name), \
    _open_files(open_files),_alias_table(alias_table),
    _data_priority_offset(data_priority),
    _ostr(NULL),_read_status(CDFDCM_SUCCESS)
{;}

CDF_DCMFile::~CDF_DCMFile()
{;}

void CDF_DCMFile::open()
{
  try
  {
    _file_data.read_from_file(_file_name);
  }
  catch(Exception& e)
  {
    e.print_msg();
  }
  _open_files.push_back(_file_name);

}

void CDF_DCMFile::close()
{
  _file_data.delete_all_records();
  _open_files.pop_back();
}

const char* CDF_DCMFile::get_alias(const char* name)
{
  string& newname=_alias_table[name];
  if(newname.size()==0)
  {
     ExceptionMsg m;
     m << "CDF_DCMFile::get_alias: alias "<< name << " not found";
     throw CDF_DCMException(m);
  }
  return newname.c_str();
}

// +++++++++++++++++++++++++++++++++++++++++++ class CDFDataConfigurationMap
// ::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
// ------------------------------------------------ ~CDFDataConfigurationMap
CDFDataConfigurationMap::~CDFDataConfigurationMap()
{
    for(CDF_DCMDataSourceTable::recursive_iterator itr=\
      _data_source_table->recursive_begin();\
        itr!=_data_source_table->recursive_end();itr++)
    {
        if(itr.type()==N_OBJECT)
            delete itr->second._obj;
    }
    for(CDF_DCMMetaDataTable::recursive_iterator itr=\
      _meta_data_table->recursive_begin();\
        itr!=_meta_data_table->recursive_end();itr++)
    {
        if(itr.type()==N_OBJECT)
            delete itr->second._obj;
    }
  
}
// ++++++++++++++++++++++++++ class CDF_DataConfigurationMap::CDF_DCMRecordTypes
// :::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
// ---------------------------------------------------------- CDF_DCMRecordTypes
CDFDataConfigurationMap::CDF_DCMRecordTypes::CDF_DCMRecordTypes()
{
  set_process_recordtype(DCM_COMMENT_RECORD,new ProcessComment);
  set_process_recordtype(DCM_ALIAS_RECORD,new ProcessAlias);
  set_process_recordtype(DCM_UNALIAS_RECORD,new ProcessUnAlias);
  set_process_recordtype(DCM_MERGE_RECORD,new ProcessMerge);
  set_process_recordtype(DCM_DATA_RECORD,new ProcessData);
  set_process_recordtype(DCM_DATA_INFO_RECORD,new ProcessDataInfo);
  set_process_recordtype(DCM_DATA_DIR_RECORD,new ProcessDataDir);
}
// ++++++++++++++++++++++++++ class CDF_DataConfigurationMap::~CDF_DCMRecordTypes
// :::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
// ---------------------------------------------------------- ~CDF_DCMRecordTypes
CDFDataConfigurationMap::CDF_DCMRecordTypes::~CDF_DCMRecordTypes()
{
   for(map<string,ProcessRecord*>::iterator itr=
      _recordtype.begin();\
        itr!=_recordtype.end();itr++)
    {
        ProcessRecord* proc_rec=itr->second;
	if(proc_rec)
	  delete proc_rec;
    }
}
// ++++++++++++++++++++++++++ class CDFDataConfigurationMap::process_file_data
// ::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
// ---------------------------------------------------------- process_file_data
CDFDCMReadStatus CDFDataConfigurationMap::process_file_data(CDF_DCMFile& cdf_dcm_file,\
    CDF_DCMAliasTable& alias_table, CDF_DCMFileStack& open_files)
{
// process data records in _file_data

    CDFDCMReadStatus res=CDFDCM_SUCCESS;

    #ifdef DIAG
    cout << "processing " << cdf_dcm_file.file_name() << endl;
    #endif

    static CDF_DCMRecordTypes record_types;
    CDF_DCMFileData& file_data=cdf_dcm_file.file_data();
    string null_string=DCM_NULL_STR;

    for(register int i=0;i<file_data.n_records();i++)
    {
        try{
          // ignore badly formed records ... or throw warning ... or error
          string field0=file_data[i][0];
          if (field0!=null_string)
          {
              ProcessRecord* process_record=\
                  record_types.get_process_recordtype(field0);
              if (process_record)
                  process_record->operator()((*this),cdf_dcm_file,i,alias_table,\
                      open_files);
              else
              {
                  ExceptionMsg m;
                  m << "CDF_DataConfigurationMap::process_file_data: " << \
                    "ignoring record # " << i << " - unrecognised record type "\
                      << field0;
                  CDF_DCMWarning warn_msg(m);
              }
	   }
	 }
	 catch(Exception& e)
	 {
	   res=CDFDCM_FAILURE;
	   cerr << "The following error occurred at record " << i << endl;
	   e.print_msg();
	   continue;
         }
     }
     return res;
}
// ++++++++++++++++++++ class CDFDataConfigurationMap::ProcessAlias::operator()
// ::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
// --------------------------------------------------------------- ProcessAlias
void CDFDataConfigurationMap::ProcessAlias::operator()
(CDFDataConfigurationMap& data_map,CDF_DCMFile& cdf_dcm_file, int recnum, \
CDF_DCMAliasTable& alias_table, CDF_DCMFileStack& open_files )
{
// process alias line
    #ifdef DIAG
    cout << "processing alias record..." << endl;
    #endif

    CDF_DCMRecord& rec=cdf_dcm_file.file_data()[recnum];
    string null_string=DCM_NULL_STR;

    if(rec.n_fields()<2)
    {
       ExceptionMsg m;
       m << "Badly formed record " << recnum;
       throw CDF_DCMException(m);
    }

    if (rec[1]!=null_string)
    {
        string alias_name=rec[1];
        string alias_string=null_string;
        int fld_num=2;
        while(fld_num<rec.n_fields() && alias_string==null_string)
            alias_string=rec[fld_num++];

        if(alias_string==null_string)
        {
            ExceptionMsg m;
            m << "CDF_DataConfigurationMap::ProcessAlias: alias < " \
                << alias_name << " > not defined";
            throw CDF_DCMException(m);
        }

        #ifdef DIAG
        cout << "adding " << alias_name << " -> " << alias_string \
            << "to alias table" << endl;
        #endif
        alias_table[alias_name]=alias_string;
    }
    else
    {
        ExceptionMsg m;
        m << "CDF_DataConfigurationMap::ProcessAlias: "\
            << " no alias name supplied";
        CDF_DCMWarning warning(m);
    }
}
// ++++++++++++++++++ class CDFDataConfigurationMap::ProcessUnAlias::operator()
// ::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
// ------------------------------------------------------------ ProcessUnAlias
void CDFDataConfigurationMap::ProcessUnAlias::operator()
(CDFDataConfigurationMap& data_map,CDF_DCMFile& cdf_dcm_file, int recnum,
CDF_DCMAliasTable& alias_table, CDF_DCMFileStack& open_files )
{
// process unalias line - what about an unalias all ?

    #ifdef DIAG
    cout << "processing unalias record..." << endl;
    #endif

    CDF_DCMRecord& rec=cdf_dcm_file.file_data()[recnum];
    string null_string=DCM_NULL_STR;

    if(rec.n_fields()<2)
    {
       ExceptionMsg m;
       m << "Badly formed record " << recnum;
              throw CDF_DCMException(m);
    }

    if (rec[1]!=null_string)
    {
        string alias_name=rec[1];

        size_t num=alias_table.erase(alias_name);
	#ifdef DIAG
        cout << "CDF_DataConfigurationMap::ProcessUnAlias: " << alias_name \
            << " deleted " << num << " entries from alias table" << endl;
	#endif
    }
    else
    {
         ExceptionMsg m;
         m << "CDF_DataConfigurationMap::ProcessUnAlias: "\
            << " no alias name supplied";
         CDF_DCMWarning warning(m);
    }
}
// ++++++++++++++++++++ class CDFDataConfigurationMap::ProcessMerge::operator()
// ::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
// --------------------------------------------------------------- ProcessMerge
void CDFDataConfigurationMap::ProcessMerge::operator()
(CDFDataConfigurationMap& data_map,CDF_DCMFile& cdf_dcm_file, int recnum,
CDF_DCMAliasTable& alias_table, CDF_DCMFileStack& open_files )
{
// process merge line

    #ifdef DIAG
    cout << "processing merge record..." << endl;
    #endif

    CDF_DCMRecord& rec=cdf_dcm_file.file_data()[recnum];
    string null_string=DCM_NULL_STR;

    if(rec.n_fields()<3)
    {
       ExceptionMsg m;
       m << "Badly formed record " << recnum;
              throw CDF_DCMException(m);
    }

    double offset=cdf_dcm_file.data_priority();

    string file_name=rec[2];
    if (file_name!=null_string)
    {
        try
        {
	    double priority=string_to_double(rec[1]);
            offset+=priority;
        }
        catch(...)
        {
	    ExceptionMsg m;
            m << "CDF_DataConfigurationMap::ProcessMerge: " << \
                "error converting field one < " << rec[1] << " > to double";
            throw CDF_DCMException(m);
        }

        #ifdef DIAG
        cout << "current file is " << cdf_dcm_file.file_name();
        #endif
        string curdir=cdf_dcm_file.get_curdir();
        data_map.read_merge_file(file_name,offset,alias_table,open_files,curdir);
    }
    else
    {
        ExceptionMsg m;
        m << "CDF_DataConfigurationMap::ProcessMerge: "\
            << "no file name supplied - ignoring record";
        CDF_DCMWarning warning(m);
    }
}
// ++++++++++++++++++ class CDFDataConfigurationMap::ProcessDataDir::operator()
// ::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
// ------------------------------------------------------------- ProcessDataDir
void CDFDataConfigurationMap::ProcessDataDir::operator()
(CDFDataConfigurationMap& data_map,CDF_DCMFile& cdf_dcm_file, int recnum, \
CDF_DCMAliasTable& alias_table, CDF_DCMFileStack& open_files )
{
// process data line

    #ifdef DIAG
    cout << "processing data directory record..." << endl;
    #endif

    CDF_DCMRecord& rec=cdf_dcm_file.file_data()[recnum];
    if(rec.n_fields()<2)
    {
       ExceptionMsg m;
       m << "Badly formed record " << recnum;
              throw CDF_DCMException(m);
    }

    string null_string=DCM_NULL_STR;
    if (rec[1]!=null_string)
      data_map.add_filename(rec[1],cdf_dcm_file.file_name());
}
// +++++++++++++++++++++ class CDFDataConfigurationMap::ProcessData::operator()
// ::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
// ---------------------------------------------------------------- ProcessData
void CDFDataConfigurationMap::ProcessData::operator() \
(CDFDataConfigurationMap& data_map,CDF_DCMFile& cdf_dcm_file, int recnum, \
CDF_DCMAliasTable& alias_table, CDF_DCMFileStack& open_files )
{
// process data line

    #ifdef DIAG
    cout << "processing data record..." << endl;
    #endif

    CDF_DCMRecord& rec=cdf_dcm_file.file_data()[recnum];
    if(rec.n_fields()<12)
    {
       ExceptionMsg m;
       m << "Badly formed record " << recnum;
              throw CDF_DCMException(m);
    }

    string null_string=DCM_NULL_STR;
    if (rec[1]==DCM_CDF_TS_DATA)
    {
         // do alias substitution and environment variable substritution
         for(register int i=1;i<rec.n_fields();i++)
         {
	     string& alias_key=rec[i];
             data_map.replace(alias_key,alias_table);
         }

         // resolve name and see if a CDF_DCMEntryMap already exists
         // for this data_source_name and if not bind it
         string data_source_name=rec[3];
         if(data_source_name!=null_string)
	 {
	   #ifdef DIAG
           cout << "data source name is " << data_source_name << endl;
           #endif
           CDF_DCMEntryMap* cdf_dcm_map;
           try
           {
	     cdf_dcm_map=
	       data_map.get_data_source_table()->resolve(data_source_name);
           }
           catch(NameNotFound& not_found)
           {
             cdf_dcm_map=new CDF_DCMEntryMap;
             data_map.get_data_source_table()->bindx(data_source_name,cdf_dcm_map);
           }

	   // data priority
	   double priority=cdf_dcm_file.data_priority();
	   try{
	     priority+=string_to_double(rec[2]);
           }
	   catch(...)
	   {
              ExceptionMsg m;
    	      m << "Invalid entry at record " << recnum << " Field 2 :: < " <<
	        rec[2] << " could not convert to double";
              throw CDF_DCMException(m);
	   }

	   string data_var_name=rec[4];
	   string time_tags_var=rec[5];
	   string time_tags_type=rec[6];
	   string file_name=rec[7];
           if(file_name!=null_string)
	     data_map.add_filename(file_name,cdf_dcm_file.file_name());
	   Time tint_start,tint_end;
           try{
	     tint_start.set_from_iso(rec[8]);
	   }
	   catch(...)
	   {
              ExceptionMsg m;
    	      m << "Invalid entry at record " << recnum << " Field 8 :: < " <<
	        rec[8] << " could not convert to time";
              throw CDF_DCMException(m);
           }
           try{
	     tint_end.set_from_iso(rec[9]);
	   }
	   catch(...)
	   {
              ExceptionMsg m;
    	      m << "Invalid entry at record " << recnum << " Field 9 :: < " <<
	        rec[9] << " could not convert to time";
              throw CDF_DCMException(m);
           }

	   int num_items;
	   try{
	     num_items=string_to_int(rec[10]);
           }
	   catch(...)
	   {
              ExceptionMsg m;
    	      m << "Invalid entry at record " << recnum << " Field 10 :: < " <<
	        rec[10] << " could not convert to int";
              throw CDF_DCMException(m);
	   }
	   //double resolution;
	   //try{
	   //  resolution=string_to_double(rec[11]);
           //}
	   //catch(...)
	   //{
           //   ExceptionMsg m;
    	   //   m << "Invalid entry at record " << recnum << " Field 11 :: < " <<
	   //     rec[11] << " could not convert to double";
           //   throw CDF_DCMException(m);
	   //}

	   string data_avail;
           if(rec.n_fields()>11)
	     data_avail=rec[11];

	   #ifdef DIAG
           cout << "adding entry for " << data_source_name <<
	     " from " << tint_start << " to " << tint_end << endl;
           #endif
           try{
             cdf_dcm_map->add_entry(new CDF_DCMEntry(
	       data_source_name.c_str(), priority,
               data_var_name.c_str(), time_tags_var.c_str(),
	       time_tags_type.c_str(), file_name.c_str(),
               tint_start.get_time_rep(), tint_end.get_time_rep(), num_items,
	       data_avail.c_str()));
           }
           catch(Exception& e)
	   {
	      ExceptionMsg m;
	      m << "Error in creation of CDF_DCMEntry";
    	      throw CDF_DCMException(m);
	   }

	 }
    }
    else
    {
        ostrstream ost;
        ost << "CDFDataConfigurationMap::ProcessData: "\
            << " Invalid entry at record " << recnum << " Field 2 :: < " << rec[1]
	    << " > expecting " << DCM_CDF_TS_DATA;
        CDF_DCMWarning warning(ost.str());
    }

}
// ++++++++++++++++++ class CDFDataConfigurationMap::ProcessDataInfo::operator()
// :::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
// ------------------------------------------------------------- ProcessDataInfo
void CDFDataConfigurationMap::ProcessDataInfo::operator()
(CDFDataConfigurationMap& data_map,CDF_DCMFile& cdf_dcm_file, int recnum, \
CDF_DCMAliasTable& alias_table, CDF_DCMFileStack& open_files )
{
// process data info line

    #ifdef DIAG
    cout << "processing data information record..." << endl;
    #endif

    CDF_DCMRecord& rec=cdf_dcm_file.file_data()[recnum];
    if(rec.n_fields()<4)
    {
       ExceptionMsg m;
       m << "Badly formed record " << recnum;
              throw CDF_DCMException(m);
    }

    string null_string=DCM_NULL_STR;
    string data_source_name=rec[2];
    if(rec[1]==DCM_CDF_TS_DATA)
    {
      // resolve name and see if a CDF_DCMEntryMap already exists
      // for this data_source_name and if not bind it
      #ifdef DIAG
      cout << "data source name is " << data_source_name << endl;
      #endif

      CDF_DCMEntryMap* cdf_dcm_map;
      try
      {
        cdf_dcm_map=\
          data_map.get_data_source_table()->resolve(data_source_name);
      }
      catch( NameNotFound& not_found)
      {
        cdf_dcm_map=new CDF_DCMEntryMap;
        data_map.get_data_source_table()->bindx(data_source_name,\
          cdf_dcm_map);
      }
      cdf_dcm_map->add_data_info(rec[3]);

    }
    else
    {
      ExceptionMsg m;
      m << "CDFDataConfigurationMap::ProcessDataInfo: "\
        << " Unrecognised DCM DataBase Type < " << rec[1] << \
          " > expecting " << DCM_CDF_TS_DATA;
      CDF_DCMWarning warning(m);
    }
}
// +++++++++++++++++++++ class CDFDataConfigurationMap::get_file_data_intervals
// :::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
// ----------------------------------------------------- get_file_data_intervals
void CDFDataConfigurationMap::get_data_intervals( \
    const string& data_source_name, const TimeInterval& time_interval,\
    CDF_DCMDataIntervalList& file_list)
{
   #ifdef DIAG
   cout << "...fetching file data intervals for " << data_source_name << endl;
   #endif

   CDF_DCMEntryMap *cdf_dcm_map;
   try
   {
       cdf_dcm_map = _data_source_table->resolve(data_source_name);
   }
   catch(...)
   {
       ExceptionMsg m;
       m << "CDFDataConfigurationMap::get_file_data_intervals: "\
           << "data source name < " << data_source_name << " > not found";
       throw CDF_DCMException(m);
   }
   cdf_dcm_map->get_data_intervals(file_list,time_interval);
}
// +++++++++++++++++++++++++++++++ class CDFDataConfigurationMap::get_data_info
// :::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
// --------------------------------------------------------------- get_data_info
void CDFDataConfigurationMap::get_data_info( \
    const string& data_source_name, vector<string>& str,\
    const TimeInterval& tint)
{
   #ifdef DIAG
   cout << "...fetching data info for " << data_source_name << endl;
   #endif

   CDF_DCMEntryMap *cdf_dcm_map;
   try
   {
       cdf_dcm_map = _data_source_table->resolve(data_source_name);
   }
   catch(...)
   {
       ExceptionMsg m;
       m << "CDFDataConfigurationMap::get_data_info: "\
           << "data source name < " << data_source_name << " > not found";
       throw CDF_DCMException(m);
   }
   cdf_dcm_map->get_data_info(str);
   // Also add number of entries
   StringRep string_rep;
   int num_entries=get_num_entries(data_source_name,tint);
   string num_txt("Number of data entries for "+data_source_name+\
      "< " + tint.iso_srep() + " > : " + int_to_string(num_entries));
   str.push_back(num_txt);
}
// ++++++++++++++++++++++++++++++ class CDFDataConfigurationMap::get_time_range
// :::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
// -------------------------------------------------------------- get_time_range
void CDFDataConfigurationMap::get_time_range( const string& data_source_name,\
TimeInterval& tint)
{
    #ifdef DIAG
    cout << "...fetching time range for " << data_source_name << endl;
    #endif

    CDF_DCMEntryMap *cdf_dcm_map;
    try
    {
        cdf_dcm_map=_data_source_table->resolve(data_source_name);
    }
    catch(...)
    {
       ExceptionMsg m;
       m << "CDFDataConfigurationMap::get_time_range: "\
         << "data source name < " << data_source_name << " > not found";
       throw CDF_DCMException(m);
    }

    if(!cdf_dcm_map)
    {
       ExceptionMsg m;
       m << "CDFDataConfigurationMap::get_time_range: "\
         << "data source name < " << data_source_name << \
         " > no data entries found";
       throw CDF_DCMException(m);
    }

    cdf_dcm_map->get_time_range(tint);
}
// +++++++++++++++++++++++++++ class CDFDataConfigurationMap::is_data_available
// :::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
// ----------------------------------------------------------- is_data_available
bool CDFDataConfigurationMap::is_data_available(const string& data_source_name,\
      const TimeInterval& tint)
{
    #ifdef DIAG
    cout << " checking data availability for " << data_source_name \
        << " over time interval " << tint << endl;
    #endif

    CDF_DCMEntryMap *cdf_dcm_map;
    try
    {
        cdf_dcm_map=_data_source_table->resolve(data_source_name);
    }
    catch(Exception& e)
    {
       ExceptionMsg m;
       m << "CDFDataConfigurationMap::is_data_available: data source name < " \
         <<  data_source_name << " > not found";
       throw CDF_DCMException(m);
    }

    if(!cdf_dcm_map)
    {
       ExceptionMsg m;
       m << "CDFDataConfigurationMap::is_data_available: data source name < " \
         << data_source_name << " > no data entries found";
       throw CDF_DCMException(m);
    }

    return cdf_dcm_map->is_data_available(tint);
}

bool CDFDataConfigurationMap::is_data_available(const string& data_source_name,\
const Time& t)
{
    #ifdef DIAG
    cout << " checking data availability for " << data_source_name \
        << " at time " << t << endl;
    #endif

    CDF_DCMEntryMap *cdf_dcm_map;
    try
    {
        cdf_dcm_map=_data_source_table->resolve(data_source_name);
    }
    catch(Exception& e)
    {
        ExceptionMsg m;
        m << "CDFDataConfigurationMap::is_data_available: data source name < " \
          << data_source_name << " > not found";
        throw CDF_DCMException(m);
    }

    if(!cdf_dcm_map)
    {
        ExceptionMsg m;
        m << "CDFDataConfigurationMap::is_data_available: data source name < " \
         << data_source_name << " > no data entries found";
        throw CDF_DCMException(m);
    }

    return cdf_dcm_map->is_data_available(t);
}
// ++++++++++++++++++++++++++++++++++++++++++ class CDFDataConfigurationMap::
// :::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
// -------------------------------------------------------- get_metadata_entry
CDF_DCMMetaDataEntry*
   CDFDataConfigurationMap::get_metadata_entry( const string& data_source_name)
{
    CDF_DCMMetaDataEntry *meta_entry=0;
    try
    {
        meta_entry=_meta_data_table->resolve(data_source_name);
    }
    catch( NameNotFound& e )
    {
      meta_entry = 0;
    }
    return meta_entry;
}
// ++++++++++++++++++++++++++++++++++++++++++ class CDFDataConfigurationMap::
// :::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
// --------------------------------------------------------- get_data_entrymap
CDF_DCMEntryMap* CDFDataConfigurationMap::get_data_entrymap(\
  const string& data_source_name)
{
    CDF_DCMEntryMap *entry_map=0;
    try
    {
        entry_map=_data_source_table->resolve(data_source_name);
    }
    catch( NameNotFound& e )
    {
      entry_map = 0;
    }
    return entry_map;
}
// +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ all_lower
// :::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
// ------------------------------------------------------------------- all_lower
void all_lower(string& str)
{
    for(int i=0;i<str.size();i++)
        if(isupper(str[i]))
            str[i]=tolower(str[i]);
}
// ++++++++++++++++++++++++++++++++ class CDFDataConfigurationMap::get_filename
// :::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
// ---------------------------------------------------------------- get_filename
string CDFDataConfigurationMap::get_filename( const string& file_name,\
    const string& searchpath,CDF_DCMAliasTable& alias_table)
{
    string dir,null_string=DCM_NULL_STR;
    string full_fname=file_name;
    //all_lower(full_fname);
    int slen=full_fname.size();
    string ext=DCM_FEXT;
    //all_lower(ext);
    int sz=ext.size();

    bool found=false;
    if(slen>=sz)
        found=full_fname.substr(slen-sz,sz)==ext;
    if(!found)
        full_fname+=ext;
    if(full_fname.substr(0,1)=="/" || full_fname.substr(0,2)=="./" || \
        full_fname.substr(0,3)=="../")
        return full_fname;
    found=false;
    int i=-1;

    #ifdef DIAG
    cout << "full_fname is " << full_fname << endl;
    #endif

    Name spath(searchpath,DCM_SPATHDELIM);
    #ifdef DIAG
    cout << "spath.srep() is " << spath.srep() << endl;
    #endif

    while(!found && ++i<spath.num_components())
    {
        dir=spath[i].get_id();
        #ifdef DIAG
        cout << " dir is " << dir << endl;
        #endif
        try
        {
            replace(dir,alias_table);
            //all_lower(dir);
            if(null_string!=dir)
            {
                dir=dir+"/";
                dir+=full_fname;
                #ifdef DIAG
                cout << "attempting to open" << dir << endl;
                #endif
                ifstream infile(dir.c_str());
                if(infile)
                {
                    found=true;
                    infile.close();
                }
            }
        }
        catch(...)
        { }
    }
    if(!found)
    {
        ExceptionMsg m;
        m << "CDFDataConfigurationMap::get_filename "\
           << "file < " << file_name << " > not found";
        throw CDF_DCMException(m);
    }
   return dir;
}
// +++++++++++++++++++++++++++++++++++ class CDFDataConfigurationMap::replace
// :::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
// ------------------------------------------------------------------- replace
void CDFDataConfigurationMap::replace(string& str, \
  CDF_DCMAliasTable& alias_table)
{
    char *old_str=StringMem::strdup(str.c_str());
    const char *ptr=old_str;
    char delim[]=DCM_DELIM;

    StringTokenizer st(ptr,delim);
    delete ptr;
    int n_tokens=st.get_n_tokens();
    if(n_tokens<2)
      return;

    char *new_str=st[0];
    size_t len0=strlen(new_str);

    char start_alias[]=DCM_START_ALIAS;
    char end_alias[]=DCM_END_ALIAS;
    char start_environ[]=DCM_START_ENVIRON;
    char end_environ[]=DCM_END_ENVIRON;

    size_t len1=strlen(start_alias);
    size_t len2=strlen(end_alias);

    bool unmatched=false;
    bool noname=false;

    for(register int i=1;i<n_tokens;i++)
    {
       char *token_i=st[i];
       char* ptr_i=token_i;
       const char* var_name=0;
       char* data_name=0;
       if(0==strncmp(ptr_i,start_alias,len1))
       {
          ptr_i+=len1;
          if(NULL==strstr(ptr_i,end_alias))
          {
            unmatched=true;
	    if(token_i)
	    {
	      delete token_i;
	      token_i=NULL;
	    }
            break;
          }
          int n=strcspn(ptr_i,end_alias);
          if(n==0)
            noname=true;
          char *name=StringMem::strndup(ptr_i,n);
          string newname=alias_table[name];
          delete name;
          name=NULL;
          if(newname!="")
            var_name=newname.c_str();
          ptr_i+=n;
          ptr_i+=strspn(ptr_i,end_alias);
          data_name=StringMem::strdup(ptr_i);
      }
      else if(0==strncmp(ptr_i,start_environ,len2))
      {
         ptr_i+=len2;
         if(NULL==strstr(ptr_i,end_environ))
         {
            unmatched=true;
	    if(token_i)
	    {
	      delete token_i;
	      token_i=NULL;
	    }
            break;
         }

         int n=strcspn(ptr_i,end_environ);
         if(n==0)
	 {
           noname=true;
	   if(token_i)
	   {
	     delete token_i;
	     token_i=NULL;
	   }
	   break;
	 }
         char *name=StringMem::strndup(ptr_i,n);
         var_name=getenv(name);
         delete name;
         name=NULL;
         ptr_i+=n;
         ptr_i+=strspn(ptr_i,end_alias);
         data_name=StringMem::strdup(ptr_i);
      }
      else
      {
        string msg(\
          "CDFDataConfigurationMap::ReplaceExternalVariable: expecting ( or [ "
          " at start of string : < ");
        msg+=str+" >";
        CDF_DCMWarning warning(msg);
	if(token_i)
	{
	  delete token_i;
	  token_i=NULL;
	}
        continue;
      }

      if(!var_name)
      {
        ExceptionMsg m;
        m << "CDFDataConfigurationMap::ReplaceExternalVariable: "
          "unrecognised variable name in : < " \
          << str << " >";
	if(token_i)
	{
	  delete token_i;
	  token_i=NULL;
	}
	if(data_name)
	{
	  delete data_name;
	  data_name=NULL;
	}
	delete new_str;
        throw CDF_DCMException(m);
      }
      size_t leni=strlen(var_name);
      size_t lenj=strlen(data_name);
      new_str=
        reinterpret_cast<char*>(
	  StringMem::realloc(new_str,1,len0,len0+leni+lenj+1));
      strcpy(new_str+len0,var_name);
      len0+=leni;
      strcpy(new_str+len0,data_name);
      delete token_i;
      token_i=NULL;
      delete data_name;
      data_name=NULL;
    }
    if(!unmatched && !noname)
      str=new_str;

    delete new_str;
    new_str=NULL;

    if(unmatched)
    {
       ExceptionMsg m;
       m << "CDFDataConfigurationMap::ReplaceExternalVariable: "
         " unmatching delimeters found in : < " \
         << str << " >";
       throw CDF_DCMException(m);
    }
    if(noname)
    {
      ExceptionMsg m;
      m <<  "CDFDataConfigurationMap::ReplaceExternalVariable: "
        "missing name in : < " \
        << str << " >";
      throw CDF_DCMException(m);
    }

}
// +++++++++++++++++++++++++++++++++++ class CDFDataConfigurationMap::read_file
// :::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
// ------------------------------------------------------------------- read_file
CDFDCMReadStatus CDFDataConfigurationMap::read_file( string file_name,\
    double data_priority )
{
    CDF_DCMFileStack *open_files=new CDF_DCMFileStack;
    CDF_DCMAliasTable *alias_table=new CDF_DCMAliasTable;
    if(open_files==NULL || alias_table==NULL)
    {
      cerr << "Unable to allocate memory for alias table and open file directory";
      return CDFDCM_FAILURE;
    }
    CDFDCMReadStatus result=\
      read_merge_file(file_name,data_priority,*alias_table,*open_files);
    delete open_files;
    delete alias_table;
    return result;
}
// +++++++++++++++++++++++++++++ class CDFDataConfigurationMap::read_merge_file
// :::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
// -------------------------------------------------------------- read_merge_file
CDFDCMReadStatus CDFDataConfigurationMap::read_merge_file( \
    const string& file_name,\
    double data_priority, CDF_DCMAliasTable& alias_table, \
    CDF_DCMFileStack& open_files, string curdir)

{
    CDFDCMReadStatus result=CDFDCM_SUCCESS;
    // get search path for file
    string searchpath=curdir;
    #ifdef DIAG
    cout << "current directory is " << curdir << endl;
    #endif
    string null_string=DCM_NULL_STR;
    if(searchpath!=null_string) searchpath+=DCM_SPATHDELIM;
    searchpath+=_search_path;

    //get file name
    string merge_file=get_filename(file_name,searchpath,alias_table);

    // check if already opened
    CDF_DCMFileStack::iterator itr=\
      find(open_files.begin(),open_files.end(),merge_file);
    if(itr!=open_files.end())
    {
      ExceptionMsg m;
      m << "CDFDataConfigurationMao::read_merge_file : circular reference,"\
          << " file < " << merge_file << " > already opened";
      cerr << m << endl;
      return CDFDCM_FAILURE;
    }

    #ifdef DIAG
    cout << "reading file " << merge_file << endl;
    #endif

    CDF_DCMFile cdf_dcm_merge(this,merge_file,alias_table,
      open_files,data_priority);

    // first check if merge file already in this list (wants to be unique?)
    if(_dcm_filenames.end()==find(_dcm_filenames.begin(),_dcm_filenames.end(),
      merge_file))
    {
      _dcm_filenames.push_back(merge_file);
    }

    try{

     cdf_dcm_merge.open();
     result=process_file_data(cdf_dcm_merge,alias_table,open_files);
     cdf_dcm_merge.close();
    }
    catch(Exception& e)
    { cerr << "Error processing file " << merge_file << endl;
      e.print_msg();
      cdf_dcm_merge.close();
      result=CDFDCM_FAILURE;
    }
    return result;
}
// +++++++++++++++++++++++++++++++++++++++++++++++++++++++ class CDF_DCMEntry
// ::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
// ------------------------------------------------------------- CDF_DCMEntry
CDF_DCMEntry::CDF_DCMEntry( const char* data_source_name, double data_priority,\
  const char* data_var_name, const char* time_tags_var,\
  const char* time_tags_type, const char* file_name, \
  double time_interval_start, double time_interval_end,\
  int num_items, const char* data_avail):_data_source_name(data_source_name),\
    _variable_name(data_var_name),_time_tags_variable_name(time_tags_var),\
    _time_tags_type(time_tags_type),_file_name(file_name),_num_items(num_items),
    _data_priority(data_priority)
{
    #ifdef DIAG
    cout << "creating new data entry" << endl;
    #endif

    if(_variable_name=="")
        throw CDF_DCMException(\
            "CDF_DCMEntry: variable name not supplied");
    if(_time_tags_variable_name!="")
    {
        if(_time_tags_type=="")
            throw CDF_DCMException(\
                "CDF_DCMEntry: time tags type not supplied");
        Time start_time,end_time;
        try{
            start_time=Seconds(time_interval_start);
        }
        catch(...)
        {
            ExceptionMsg m;
            m << "CDF_DCMEntry: invalid start time < " << time_interval_start << ">";
            throw CDF_DCMException(m);
        }
        try
        {
            end_time=Seconds(time_interval_end);
        }
        catch(...)
        {
            ExceptionMsg m;
            m <<  "CDF_DCMEntry: invalid end time < " << time_interval_end << ">";
            throw CDF_DCMException(m);
        }
	try{
          _data_availability=DataAvailability(_data_source_name,\
              TimeInterval(start_time,end_time),data_avail);
	}
	catch(...)
	{
	  ExceptionMsg m;
	  m << "CDF_DCMEntry: invalid data availability string < " << data_avail <<
	    ">";
	  throw CDF_DCMException(m);
	}
    }
    else if(0!=strcmp(time_tags_type,"") ||\
       0!=time_interval_start || \
       0!=time_interval_end || \
       0!=strcmp(data_avail,""))
        throw CDF_DCMException(\
            "CDF_DCMEntry: no time tags variable name supplied");
}
// +++++++++++++++++++++++++++++++++++++++++++++++++++++++ class CDF_DCMEntry
// ::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
// -------------------------------------------------------- is_data_available
bool CDF_DCMEntry::is_data_available(const TimeInterval& tint) const
{ return _data_availability.is_data_available(tint);}
bool CDF_DCMEntry::is_data_available(const Time& t) const
{ return _data_availability.is_data_available(t);}

// +++++++++++++++++++++++++++++++++++++++++++++++++++++++ class CDF_DCMEntry
// ::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
// -------------------------------------------------------- get_time_interval
TimeInterval CDF_DCMEntry::get_time_interval() const
{ return _data_availability.get_time_interval();}



// +++++++++++++++++++++++++++++++++++++++++++++++++++++ class CDF_DCMEntryMap
// :::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
// ---------------------------------------------------------- ~CDF_DCMEntryMap
CDF_DCMEntryMap::~CDF_DCMEntryMap()
{
    for(EntryIter itr=_entry_table.begin();itr!=_entry_table.end();\
        itr++)
        delete itr->second;
}
// +++++++++++++++++++++++++++++++++++++++++++++++++++++ class CDF_DCMEntryMap
// :::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
// ----------------------------------------------------------------- add_entry
int CDF_DCMEntryMap::add_entry( CDF_DCMEntry* cdf_data_entry )
{
// what about preventing duplicate entries i.e what is entry keyed on?
    pair<int,bool> res;

    const TimeInterval& tint=cdf_data_entry->get_time_interval();
    int entrynum=_entry_table.size();
    if(!entrynum)
        _time_range=tint;
    else
        _time_range=_time_range.int_union(tint);

    entrynum++;
    _entry_table.insert(make_pair(entrynum,cdf_data_entry));
    _startdates.insert(make_pair(tint.start(),entrynum));
    _enddates.insert(make_pair(tint.end(),entrynum));
    double priority=0.;
    if(!cdf_data_entry)
      priority=cdf_data_entry->_data_priority;
    _priorities.push_back(priority);
    return entrynum;
  }

// +++++++++++++++++++++++++++++++++++++++++ class CDF_DCMEntryMap::next_entry
// :::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
// ---------------------------------------------------------------- next_entry
int CDF_DCMEntryMap::next_entry( const TimeInterval& time_interval)
{
// Find next entry after time_interval.start(). Returns entrynum of
// corresponding entry in map. Returns 0 if no entries found, or first
// entry starts after or at time_interval.end()

    int entrynum=0;
    Time st=time_interval.start();
    Time et=time_interval.end();

    // Find entry of highest priority with a start date < time_interval.start()
    // and an end date > time_interval.start()
//    if(st<=et)
    if(st<et)
    {
        DatesIter s_itr,e_itr;
        set<int> entrynum_1,entrynum_2,entrynum_3;

        e_itr=_startdates.lower_bound(st); // first start date >= st
        s_itr=_startdates.begin();

        for(;s_itr!=e_itr;s_itr++)
            entrynum_1.insert(s_itr->second);

        s_itr=_enddates.upper_bound(st);   // first end date > st
        for(;s_itr!=_enddates.end();s_itr++)
            entrynum_2.insert(s_itr->second);

        // get intersection of entrynums to find entries with start <st
        // and end > st
        set_intersection(entrynum_1.begin(),entrynum_1.end(),\
            entrynum_2.begin(),entrynum_2.end(),\
            inserter(entrynum_3,entrynum_3.begin()));

        if(entrynum_3.size())
        {
            set<int>::iterator itr=entrynum_3.begin();
            entrynum=*itr;

            for(;itr!=entrynum_3.end();itr++)
                if(get_priority(*itr)>get_priority(entrynum))
                    entrynum=*itr;
        }
        else // find first start >= st if no entries found
        {
            s_itr=_startdates.lower_bound(st);
            if (s_itr!=_startdates.end())
            {
                if(s_itr->first<et)
                {
                    e_itr=_startdates.upper_bound(st);
                    // may be more than one entry with same start - get entry
                    // with highest priority
                    entrynum=s_itr->second;
                    double priority=get_priority(entrynum);
                    for(;s_itr!=e_itr;s_itr++)
                        if(get_priority(s_itr->second)>priority)
                            entrynum=s_itr->second;
                }
            }
        }
    }
    return entrynum;
}
// +++++++++++++++++++++++++++++++++++ class CDF_DCMEntryMap::get_file_intervals
// :::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
// ---------------------------------------------------------- get_file_intervals
void CDF_DCMEntryMap::get_data_intervals(CDF_DCMDataIntervalList& filelist)
{
    DatesIter s_itr,e_itr;
    s_itr=_startdates.begin();
    e_itr=_enddates.end();

    Time st=s_itr->first;
    Time et=(--e_itr)->first;

    get_data_intervals( filelist, TimeInterval(st,et) );

}

void CDF_DCMEntryMap::get_data_intervals( CDF_DCMDataIntervalList& filelist, \
    const TimeInterval& time_interval)
{
    #ifdef DIAG
    cout << "...getting file data intervals from entry table " << endl;
    #endif

    DatesIter s_itr,e_itr;
    int entrynum=next_entry(time_interval);

    //CDF_DCMEntry* entry=_entry_table[entrynum];
    Time st=time_interval.start();
    Time et=time_interval.end();
    while(entrynum)
    {
        CDF_DCMEntry* entry=_entry_table[entrynum];
        const TimeInterval& tint=entry->get_time_interval();
        Time start=max( tint.start(), st );
//        Time end  = tint.end();
        Time end  = min( tint.end(), et );
        double priority=entry->_data_priority;

        const CDF_DCMEntry& entryref(*entry);
        // find entry with a start after current start and before
        // current end. of a higher priority
        s_itr=_startdates.upper_bound(start);
        e_itr=_startdates.lower_bound(end);
        while(distance(_startdates.begin(),s_itr)<\
          distance(_startdates.begin(),e_itr) && \
          get_priority(s_itr->second)<=priority)
          s_itr++;
        if(distance(_startdates.begin(),s_itr)<\
          distance(_startdates.begin(),e_itr))
        {
             st=s_itr->first;
             filelist.push_back(
	     CDF_DCMDataInterval( entryref, TimeInterval(start,st) ) );

             // make sure there are no other entries starting on same date
             // with a higher priority
             e_itr=_startdates.upper_bound(st);
             entrynum=s_itr->second;
             priority=get_priority(entrynum);
             for(;s_itr!=e_itr;s_itr++)
                 if(get_priority(s_itr->second)>priority)
                    entrynum=s_itr->second;
         }
         else
         {
         // no entries of a higher priority starting before this one ends
         // find next entry
             filelist.push_back(
	     CDF_DCMDataInterval( entryref, TimeInterval(start,end) ) );
//	     CDF_DCMDataInterval( entryref, TimeInterval(start,min(end,et)) ) );
             if(start==end)
             {
               s_itr=_startdates.upper_bound(end);
               if(s_itr==_startdates.end())
                 end=et;
               else
                end=min(s_itr->first,et);
             }
             entrynum=next_entry(TimeInterval(end,et));
             st=end;
         }
    }
}
// ++++++++++++++++++++++++++++++++++++ class CDF_DCMEntryMap::is_data_available
// :::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
// ----------------------------------------------------------- is_data_available
bool CDF_DCMEntryMap::is_data_available( const Time& t) const
{
     if(!_time_range.contains(t))
         return false;
     bool res;
     for(EntryCIter itr=_entry_table.begin();itr!=_entry_table.end();itr++)
     {
         res=itr->second->is_data_available(t);
         if(res)
             break;
     }
     return res;
}

bool CDF_DCMEntryMap::is_data_available( const TimeInterval& tint) const
{
    if((_time_range.int_intersect(tint)).empty())
        return false;
    bool res;
    for(EntryCIter itr=_entry_table.begin();itr!=_entry_table.end();itr++)
    {
        res=itr->second->is_data_available(tint);
        if(res)
            break;
    }
    return res;
}
// ++++++++++++++++++++++++++++++++++++++++++++++++ class CDF_DCMMetaDataEntry
// :::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
// ------------------------------------------------------ CDF_DCMMetaDataEntry
CDF_DCMMetaDataEntry::CDF_DCMMetaDataEntry(  const char* meta_data_name,\
  const char* file_type, const char* file_name, \
  const char* file_var_name ):_data_source_name(meta_data_name), \
  _file_data_source_name(file_var_name),_file_type(file_type),\
  _file_name(file_name){;}

// +++++++++++++++++++++++++++++++++++++++++++++++++++++++++ class CDF_DCMFile
// :::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
// ---------------------------------------------------------------- get_curdir
string CDF_DCMFile::get_curdir()
{
// make sure this handles ./ and ../
    Name fname(_file_name);
    fname.delete_component(fname.num_components()-1);
    return fname.s_rep();
}

} // namespace QDOS


