// FARdb.h
//
//

#ifndef _FARDB_H_
#define _FARDB_H_

#include <iostream>
#include <string>
#include <vector>
#include <map>
#include <algorithm>
#include <functional>
#include <cstdlib>

namespace FAR {

using namespace std;

class Field : public string {

public:
  Field(void) : string("") {;}
  Field( const char* cs ) : string(cs) {;}
  Field( const string& s ) : string(s) {;}
  Field( const Field& f ) : string(f)  {;}
  operator double() const { return atof( c_str() ); }
};

class Record;
ostream& operator<<(ostream& os,const Record& rec);
class FARData;

class Record {

  vector<Field> fields;
  
  Field null_field;
  friend class FARData;
  
public:

  //Record(void): null_field("") {;}
  Record(int n=0):fields(n),null_field(""){;}
  Record( const Record& r ) : null_field("")
  { int i; for(i=0;i<r.fields.size();i++) fields.push_back( r[i] ); }
  Record( const vector<string>& sv ) : null_field("")
  { int i; for(i=0;i<sv.size();i++) fields.push_back( sv[i] ); }
  Record( const string& s ) : null_field("")
    { fields.push_back( s ); }
  int n_fields(void) const { return fields.size(); }
  bool field_exists( int f ) const { return f>=0 && f<fields.size(); }
  int add_field( const string& f )
    { fields.push_back( f ); return fields.size()-1; }
  void add_fields( vector<string>& ff );
  
  void delete_field( int i )
    { if( i>=0 && i<fields.size() ) fields.erase( fields.begin()+i );  }

  void clear_field( int i ){ if( i>=0 && i<fields.size() ) fields[i] = ""; }
  
  Field& operator[]( int i )
    { return ( i>=0 && i<fields.size() )? fields[i]: null_field; }
  const Field& operator[]( int i ) const
    { return ( i>=0 && i<fields.size() )? fields[i]: null_field; }

  void diag_print(void) const;

};


typedef enum FARCmpType_e {

  FARSORT_NULL,
  FARSORT_ALPHA,
  FARSORT_ALPHA_NO_CASE,
  FARSORT_NUMERIC,
  FARSORT_REV_ALPHA,
  FARSORT_REV_ALPHA_NO_CASE,
  FARSORT_REV_NUMERIC

} FARCmpType;

class FARcompare_records_by_field {

  int n_cmp_f;
  int field_1;
  int field_2;
  FARCmpType cmp_type_1;
  FARCmpType cmp_type_2;
  
public:

  FARcompare_records_by_field( int f, FARCmpType ct )
     : n_cmp_f(1), field_1(f), cmp_type_1(ct)  {;}
  FARcompare_records_by_field( int f_1, FARCmpType ct_1,
                               int f_2, FARCmpType ct_2 )
     : n_cmp_f(2), field_1(f_1), cmp_type_1(ct_1),
                    field_2(f_2), cmp_type_2(ct_2)  {;}

  int operator() ( const Record& r1,  const Record& r2 )const;
  int cmp_rec_1(const Record& r1, const Record& r2 ) const;
  int cmp_rec_2(const Record& r1, const Record& r2 ) const;

};

typedef enum FARWriteFileMode_e {

  FARWRITEFILE_NEW,
  FARWRITEFILE_APPEND

} FARWriteFileMode;


typedef enum FARReadFileMode_e {

  FARREADFILE_REPLACE,
  FARREADFILE_APPEND

} FARReadFileMode;


class FARData {

// data
  vector<Record> records;
  int nfieldsmax;             // largest number of fields in a record

  map<string,int> field_by_name; // fieldname -> field number mappings (no checks!)

// private static functions

  static void trim_whitespace( string& s );
  static bool is_comment_line( const char* s );
  
public:
  FARData(int n=0) : records(n), nfieldsmax(0) {;}
  ~FARData(void) { delete_all_records(); }
  
// public static functions
  static int xtokenize_string( const char *s, const char *delimiters, vector<string>& tokens );
  static int tokenize_string( const char *s, const char *delimiters, vector<string>& tokens );
// ----

  void update_nfieldsmax( int n_f ){ if( n_f > nfieldsmax ) nfieldsmax = n_f; }
  
  int n_fields_max(void) const {return nfieldsmax; }
  int n_records(void) const {return records.size(); }
  int add_record( const Record& rec )
    { records.push_back( rec ); update_nfieldsmax( rec.n_fields() );
      return records.size()-1;}
  
  void xfer_record( int r, FARData& db ){ db.add_record( records[r] ); delete_record(r);}
  
  void delete_record( int r )
    { if( r>=0 && r<records.size() ) records.erase( records.begin() + r ); }
    
  void delete_all_records(void)
    { records.clear(); }
  
  void delete_field( int f );
  
  Record& operator[](int i){ return records[i]; }
  const Record& operator[](int i) const { return records[i]; }
  
  void read_from_file( string file_name,
         FARReadFileMode read_mode=FARREADFILE_REPLACE, string separator="," );
  void append_from_file( string file_name, string separator="," );
  
  void write_to_file( string file_name = "FARdata.csv",
      FARWriteFileMode write_mode=FARWRITEFILE_NEW, string separator="," );

  void sort_records_by_field( const FARcompare_records_by_field& cmp )
    { stable_sort( records.begin(), records.end(), cmp ); }

  int match_record_by_field( int field, string s, int r_start=0, int r_end=-1 ) const;
  int match_record_by_field( int field, double x, int r_start=0, int r_end=-1 ) const;
  
  void set_field_name( int f, string field_name ) { field_by_name[field_name]=f; }
  int  get_field_number( string field_name ){ return field_by_name[field_name]; }
  
  void diag_print(void) const;

};


} // end module FAR

#endif // _FARDB_H_
