//
//  Factory.cc
//
//  Factory class for cloning objects
//
//  David Burgess
//  March 2000
//

#include <vector>
#include "Factory.h"
#include "Matrix.cc"

namespace QDOS {

// String

QdString_var Factory::clone( const string& s )
{ return new QdString(s); }

QdString_var Factory::clone( QdString_var qs_var )
{ return new QdString( *qs_var ); }

// StringSeq

QdStringSeq_var Factory::clone( const vector<string>& s )
{ return new QdStringSeq( s ); }
QdStringSeq_var Factory::clone( const QdStringSeq& s )
{ return new QdStringSeq( s ); }
QdStringSeq_var Factory::clone( QdStringSeq_var qs_var )
{ return new QdStringSeq( *qs_var ); }

// Time

QdTime_var Factory::clone( const Time& t )
{ return new QdTime( t ); }
QdTime_var Factory::clone( QdTime_var t_var )
{ return new QdTime( *t_var ); }

// TimeSeq

QdTimeSeq_var Factory::clone( const TimeSeq& tseq )
{ return new QdTimeSeq( tseq ); }
QdTimeSeq_var Factory::clone( QdTimeSeq_var tseq_var )
{ return new QdTimeSeq( *tseq_var ); }

// TimeInterval

QdTimeInterval_var Factory::clone( const TimeInterval& tintvl )
{ return new QdTimeInterval( tintvl ); }
QdTimeInterval_var Factory::clone( QdTimeInterval_var tintvl_var )
{ return new QdTimeInterval( *tintvl_var ); }

// StringDataBuffer
QdObject_var Factory::clone( StringDataBuffer_var sdbuff )
{
  QdObject_var obj; // return object
  vector<string> v;
  DataBufferDescriptor_var db_desc = sdbuff->descriptor();
  vector<int> dims;
  db_desc->get_dims( dims );
  int nitems = sdbuff->num_items();
  int nitem_elts = sdbuff->num_item_elts();
  if( nitems == 0 ) // nil object!
  { return obj; }
  else if( nitems == 1 ) // single item String or StringMatrix
  {
    sdbuff->get_data( nitem_elts, v );
    if( dims.size() == 0 )  // String
      obj = new QdString( v[0] );
    else if(dims.size() == 1 ) // dims.size()=1 RSatrix
    {
      if(dims[0]==1) // SMatrix of size (1,1) so treat as String
	obj=new QdString( v[0]) ;
      else
      {
        Dim dd(dims);
        obj = new QdSMatrix( dd, v );
      }
    }else                  // dims.size()>0 => RMatrix
    {
       Dim dd(dims);
       obj = new QdSMatrix( dd, v );
    }
  } else // if( nitems > 1 ) // sequence of String or Stringmatrix
  {
    if( dims.size() == 0 )  // StringSeq
    {
      sdbuff->get_data( nitems, v ); // get all values in seq
      obj = new QdStringSeq( v );
    } 
    else if(dims.size() == 1 ) // dims.size()=1 RSatrix
    {
      if(dims[0]==1) // SMatrix of size (1,1) so treat as StringSeq
      {
        sdbuff->get_data( nitems, v ); // get all values in seq
        obj = new QdStringSeq( v );
      }
      else
      {
        QdSMatrixSeq_var sms = new QdSMatrixSeq(nitems,Dim(dims));
        for(int I=0;I<nitems;I++)
        {
          v.clear();
          sdbuff->get_data( nitem_elts,v );
          sms[I].assign_vector(v); 
        }
        obj=sms;
      }
    }else                  // dims.size()>0 => RMatrix
    {
        QdSMatrixSeq_var sms = new QdSMatrixSeq(nitems,Dim(dims));
        for(int I=0;I<nitems;I++)
        {
          v.clear();
          sdbuff->get_data( nitem_elts,v );
          sms[I].assign_vector(v); 
        }
        obj=sms;
    }
  }

  return obj;
} // end clone StringDataBuffer

QdObject_var Factory::clone( NumericDataBuffer_var ndbuff )
{
  QdObject_var obj; // return object
  vector<double> v; // for data from buffer
  DataBufferDescriptor_var db_desc = ndbuff->descriptor();
  vector<int> dims;
  db_desc->get_dims( dims );
  int nitems = db_desc->num_items();
  int nitem_elts = db_desc->num_item_elts();
  if( nitems == 0 ) // nil object!
  { return obj; }
  else if( nitems == 1 ) // single item RScalar or RMatrix
  {
    ndbuff->get_data( nitem_elts, v );
    if( dims.size() == 0 )  // RScalar
      obj = new QdRScalar( v[0] );
    else if(dims.size() == 1 ) // dims.size()=1 RMatrix
    {
      if(dims[0]==1) // RMatrix of size (1,1) so treat as RScalar
	obj=new QdRScalar( v[0]) ;
      else
      {
        Dim dd(dims);
        obj = new QdRMatrix( dd, v );
      }
    }else                  // dims.size()>0 => RMatrix
    {
      Dim dd(dims);
      obj = new QdRMatrix( dd, v );
    }
  } else // if( nitems > 1 ) // sequence of RScalar or RMatrix
  {
    if( dims.size() == 0 )  // RScalarSeq
    {
      ndbuff->get_data( nitems, v ); // get all values in seq
      obj = new QdRScalarSeq( v );
    }
    else if(dims.size() == 1 ) // dims.size()=1 RMatrixSeq
    {
      if(dims[0]==1) // RMatrix of size (1,1) so treat as RScalarSeq
      {
        ndbuff->get_data( nitems, v ); // get all values in seq
        obj = new QdRScalarSeq( v );
      }
      else
      {
        Dim dd(dims);
        QdRMatrixSeq_var rmatseq = new QdRMatrixSeq(nitems,dd);
        obj = rmatseq;
        for(int i=0;i<nitems;i++)
        {
          v.clear();
          ndbuff->get_data( nitem_elts, v );
          rmatseq[i]=RMatrix(dd,v) ;
        }
      }
    }
    else                  // dims.size()>0 => RMatrix
    {
      Dim dd(dims);
      QdRMatrixSeq_var rmatseq = new QdRMatrixSeq(nitems,dd);
      obj = rmatseq;
      for(int i=0;i<nitems;i++)
      {
        v.clear();
        ndbuff->get_data( nitem_elts, v );
        //RMatrix m( dd, v );
        rmatseq[i]=RMatrix(dd,v) ;
      }
    }
  }
  return obj;
} // end clone NumericDataBuffer


// DataBuffer

QdObject_var Factory::clone( DataBuffer_var dbuff )
{
  QdObject_var obj; // return object
  {
    NumericDataBuffer_var ndbuff = NumericDataBuffer_var::narrow( dbuff );

    if( !ndbuff.is_nil() )
    {
      if( ndbuff->num_items() == 0 )
        return obj;
      else
        return clone( ndbuff );
    }
  }
// DataBuffer is StringDataBuffer
  {
    StringDataBuffer_var sdbuff = StringDataBuffer_var::narrow( dbuff );
  
    if( !sdbuff.is_nil() )
    {
      if( sdbuff->num_items() == 0 )
        return obj;
      else
        return clone( sdbuff );
    }
  }

// DataBuffer is DataPtrBuffer
  {
    DataPtrBuffer_var dp_dbuff=DataPtrBuffer_var::narrow(dbuff);
    if(!dp_dbuff.is_nil())
    {
      obj=clone(dp_dbuff->data); // clone the data part
      if(obj.is_nil())
        return obj;
      QdObject_var data_ptr_name_obj=clone(dp_dbuff->data_ptr_name);
      if(data_ptr_name_obj.is_nil())
        return obj;
      obj->set_xref(DATA_PTR_NAME,data_ptr_name_obj);
      return obj;
    }
  }

// DataBuffer is TSDataBuffer
  {
    TSDataBuffer_var ts_dbuff = TSDataBuffer_var::narrow( dbuff );
    if( !ts_dbuff.is_nil() )
    {      
      obj = clone( ts_dbuff->data ); // clone the data part

// if object from data is nil then either zero items in databuffer
//  or unknown data buffer type.
// either way ... return nil object
      if( obj.is_nil() )
        return obj;

      NumericDataBuffer_var ntt_dbuff =
         NumericDataBuffer_var::narrow( ts_dbuff->ttags );
      
      if( ntt_dbuff.is_nil() )
        throw FactoryException(
                ExceptionMsg()
                 << "Factory::clone: Time tags DataBuffer must be Numeric not: "
                 << dbuff->get_type_srep() );

// if there are zero items in data buffer return nil object
      if( ntt_dbuff->num_items() == 0 )
        return obj;

      DataBufferDescriptor_var db_desc = ntt_dbuff->descriptor();
      if( db_desc->num_item_elts() != 1 )
        throw FactoryException(
                ExceptionMsg()
                 << "Factory::clone: Time tags NumericDataBuffer scalar sequence"
                  );
      
      int nitems = db_desc->num_items();
      vector<double> v;
      ntt_dbuff->get_data( nitems, v ); // get all data

      QdTimeSeq_var tt_obj = new QdTimeSeq( nitems ); // create time tags object
      for(int i=0;i<nitems;i++)
        tt_obj[i]=Seconds(v[i]);

      obj->set_xref( TTAGS, tt_obj ); // set cross ref

      return obj;
    }
  }

// could not match data buffer type

  throw FactoryException(
  ExceptionMsg()<< "Factory::clone: Could not clone from DataBuffer_var: type: "
   << dbuff->get_type_srep() );

  return obj;

}

// RScalar

QdRScalar_var Factory::clone( double x )
{ return new QdRScalar( x ); }

QdRScalar_var Factory::clone( const QdRScalar& x )
{ return new QdRScalar( x ); }

QdRScalar_var Factory::clone( QdRScalar_var x_var )
{ return new QdRScalar( *x_var ); }

// RScalarSeq

QdRScalarSeq_var Factory::clone( const vector<double>& v )
{ QdRScalarSeq_var r = new QdRScalarSeq;
 r->assign( v ); return r;}
 
QdRScalarSeq_var Factory::clone( const vector<float>& v )
{ QdRScalarSeq_var r = new QdRScalarSeq;
 r->assign( v ); return r;}

QdRScalarSeq_var Factory::clone( const vector<int>& v )
{ QdRScalarSeq_var r = new QdRScalarSeq;
 r->assign( v ); return r;}
 
QdRScalarSeq_var Factory::clone( const RScalarSeq& rss )
{ return new QdRScalarSeq( rss ); }

QdRScalarSeq_var Factory::clone( QdRScalarSeq_var rss_var )
{ return new QdRScalarSeq( *rss_var ); }

// RMatrix

QdRMatrix_var Factory::clone( const RMatrix& m )
{ return new QdRMatrix( m ); }

QdRMatrix_var Factory::clone( const QdRMatrix_var& m_var )
{ return new QdRMatrix( *m_var ); }

// SMatrix

QdSMatrix_var Factory::clone( const SMatrix& sm )
{ return new QdSMatrix(sm ); }

QdSMatrix_var Factory::clone (const QdSMatrix_var & sm_var)
{ return new QdSMatrix( *sm_var ); }

// RMatrixSeq

QdRMatrixSeq_var Factory::clone( const RMatrixSeq& mseq )
{ return new QdRMatrixSeq( mseq ); }

QdRMatrixSeq_var Factory::clone( const QdRMatrixSeq_var& mseq_var )
{ return new QdRMatrixSeq( *mseq_var ); }

// SMatrixSeq

QdSMatrixSeq_var Factory::clone( const SMatrixSeq& smseq )
{ return new QdSMatrixSeq( smseq ); }

QdSMatrixSeq_var Factory::clone( const QdSMatrixSeq_var& smseq_var )
{ return new QdSMatrixSeq( *smseq_var ); }



} // end namespace QDOS
