//
//  MatrixEig.h
//
//  Qdos2 - Class for calculating eigenvalues/vectors of a symmetric numeric
//  matrix
//
//  Janet Barnes
//  Sept 1999
//

#ifndef QDOS_EIG_H
#define QDOS_EIG_H

#include <algorithm>
#include "Matrix.h"

using namespace std;

namespace QDOS {

// CLAPACK routines for calculating eigenvalue/vectors
extern "C" int dsyev_(char *jobz, char *uplo, long int *n, double *a,
	 long int *lda, double *w, double *work, long int *lwork, long int *info); 
extern "C" long int ilaenv_(long int *ispec, char *name, char *opts,
         long int *n1, long int *n2, long int *n3, long int *n4, 
         long name_len, long opts_len);	 
         
// --- Cover functions for CLAPACK routines
int eigcomp(char *jobz,char *uplo,long int *n,double *a, long int *lda,\
    double *w,long int *info);
long int blcksize(char *fname,long int *n);
         
// +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ class MatrixEig
// Class for calculating eigenvalues/egigenvectors for a rank 2 square
// array
// -----------------------------------------------------------------------------
class MatrixEig
{
 protected:
    size_t _n;                    // no. rows/cols
    ScalarSeq<double> _eigvals;   // _n elt sequence of _rank 0 matrices
    MatrixSeq<double> _eigvecs;   // _n elt sequence of rank 1 matrices
    
 public:
     // --- Constructors
     // Empty constructor
     MatrixEig(){;}
     // Construct eigenvalues/eigenvectors from Matrix
     template<class T>
     MatrixEig(const Matrix<T>& m):_n(m.dims()[0]),\
         _eigvecs(MatrixSeq<double>(m.dims()[0])),\
         _eigvals(ScalarSeq<double>(m.dims()[0]))
     {
         try
         {
             int res;
             // verify that matrix argument is rank 2
             if (m.rank()!=2)
                 throw MatrixRankError();
             //  verify that matrix is symmetric
             if (!m.is_symmetric())
             {
                 throw MatrixDomainError(ExceptionMsg()<<\
                     "Matrix not symmetric");
             }   
             char jobz='V'; // Compute eigenvalues AND eigenvectors
             char uplo='L';
             long int n=_n;             
             Matrix<double> eigvecs(Dim(_n,_n,COL_MAJOR_ORDER));
             eigvecs.copy(m);
             long int lda=_n;
             long int info;
             Matrix<double> eigvals(Dim(_n,COL_MAJOR_ORDER));
             // call CLAPACK routine
             #ifdef DIAG
             cout << "calling CLAPACK eigenvalue/eigenvector routine" << endl;
             #endif
             res=eigcomp(&jobz,&uplo,&n,eigvecs.ptr(),&lda,eigvals.ptr(),&info);
             // convert to MatrixSeq
             for(int i=0;i<n;i++)
             {  
                  _eigvecs[i]=eigvecs.col(i);
                  _eigvals[i]=eigvals.row(i);
             }
                  
         }           
         catch(Exception& e)
         {
             e.msg() << "[MatrixEig<T>]";
             throw e;
         }
     }
     
     // --- Misc functions
     ScalarSeq<double>& eigenvalues() 
     {
         return _eigvals;
     }
     
     MatrixSeq<double>& eigenvectors() 
     {
         return _eigvecs;
     }
     
};

}  // end namespace QDOS        
#endif
