//
//  Time.cc
//
//  Qdos - Time, date and time intervals
//
//  David Burgess
//  July 1999
//

#include "QdTime.h"

#include <stdio.h>  // for sprintf
#include <math.h>  // for fmod

namespace QDOS {

// static variables

int Time::epoch = 2451545;  // epoch is Julian 2000

int Time::qie_days_in_month[2][13] = {
  { 0, 31, 28, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31 },
  { 0, 31, 29, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31 } };
const char* Time::month_names[13] = {
  "", "Jan", "Feb", "Mar", "Apr", "May",
  "Jun", "Jul", "Aug", "Sep", "Oct", "Nov", "Dec" };

const char* Time::full_month_names[13] = { 
  "", "January", "February", "March", "April", "May",
  "June", "July", "August", "September", "October", "November", "December" }; 

const double SecondsFactor::mfac = 1.0;
const double MinutesFactor::mfac = 60.0;
const double HoursFactor::mfac   = 3600.0;
const double DaysFactor::mfac    = 86400.0;
const double mSecondsFactor::mfac= 0.001;

const double TimeUnitsOffset::offset=0.0;
const double cdfEpochOffset::offset=86400.0*(double)(1721060-Time::epoch);

// ++++++++++++++++++++++++++++++++++++++++++++++++++++++ class Time
// ::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
// --------------------------------------------------- cdf_julian_day
// no error checks
// derived from JulianDay in CDF 2.4.11 (epochu.c)

int Time::cdf_julian_day( int y, int m, int d )
{
 return (367*y-7*(y+(m+9)/12)/4-3*((y+(m-9)/7)/100+1)/4+275*m/9+d+1721029);
}
// +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ class Time
// :::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
// ---------------------------------------------------------- set_from_iso
Time& Time::set_from_iso(const string& iso_string)
{
// set from iso string. Assume in one of following forms
// yyyy-mm-dd
// yyyy-mm-ddTHH  (or yyyy-mm-dd HH)
// yyyy-mm-ddTHH:MM (or yyyy-mm-dd HH:MM)
// yyyy-mm-ddTHH:MM:SS.SSS (or yyyy-mm-dd HH:MM:SS.SSS)
// ... to be added later variations on date (together with all time variations)
// dd
// mm-dd
// yy-mm-dd
// yyyy-mm-dd

// Modified by AJA Apr 2007 - this version is 4x faster

      int date1;
      int date2;
      int date3;
      double time1=0.0;
      double time2=0.0;
      double time3=0.0;
      
      int posn = iso_string.find('-'); // work off first separator in case of leading space
      if( posn == -1)  throw BadTime();
      
      date1 = atoi(iso_string.substr(posn-4, 4).c_str() );    
      date2 = atoi(iso_string.substr(posn+1, 2).c_str() );  
      date3 = atoi(iso_string.substr(posn+4, 2).c_str() );
     
      int iso_len = iso_string.size();
      
      if (iso_len > posn+14) {
         time3 = atof(iso_string.substr(posn+13, iso_len-posn-13).c_str() );
	 time2 = atof(iso_string.substr(posn+10, 2).c_str() );
	 time1 = atof(iso_string.substr(posn+7, 2).c_str() );
      }
      else if (iso_len > posn+11){
	 time2 = atof(iso_string.substr(posn+10, 2).c_str() );
	 time1 = atof(iso_string.substr(posn+7, 2).c_str() );
      }
      else if (iso_len > posn+8){
	 time1 = atof(iso_string.substr(posn+7, 2).c_str() );
      }
    
      set_ymd(date1,date2,date3,time1,time2,time3);
      
      /*
      int date[3];
      double time[]={0.,0.,0.};
      Name iso_name(iso_string,"T ");
      int iso_num=iso_name.num_components();

      if(iso_num<1) throw BadTime();
      
      // process date strings
      Name date_name(iso_name[0].get_id(),"-");
      int date_num=date_name.num_components();
      if (date_num!=3)
          throw BadTime();
      for(int i=0;i<date_num;i++)
      {
          istrstream ist(date_name[i].get_id().c_str());
          ist >> date[i];
      }
      
      if(iso_num>1)
      {
          Name time_name(iso_name[1].get_id(),":");
          int time_num=time_name.num_components();
          for(int i=0;i<time_num;i++)
          {
              istrstream ist(time_name[i].get_id().c_str());
              ist >> time[i];
          }
      }
      
      set_ymd(date[0],date[1],date[2],time[0],time[1],time[2]);
      */
      
      
}
// ++++++++++++++++++++++++++++++++++++++++++++++++++++++ class Time
// ::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
// ---------------------------------------------------------- set_ymd
// derived from computeEPOCH in CDF 2.4.11 (epochu.c)

Time& Time::assign_ymd( int year, int month, int day,
                       double hour, double min, double sec )
{
// error checks
  if(    year < 0 || year > 9999   // upper limit from cdf implementation (?)
      || month < 1 || month > 12
      || day < 1 || day > qie_days_in_month[leap_index(year)][month]
      || hour < 0.0 || hour >= 24.0
      || min  < 0.0 || min  >= 60.0
      || sec  < 0.0 || sec  >= 60.0 )
     throw BadTime();
  
  _trep = sec + 60.0 * min + 3600.0 * hour +\
     86400.0 * ( (double) cdf_julian_day (year, month, day) - epoch );

  return *this;
}

// ++++++++++++++++++++++++++++++++++++++++++++++++++++++ class Time
// ::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
// --------------------------------------------------------- set_ydoy
Time& Time::assign_ydoy( int year, int doy,
            double hour, double min, double sec )
{
// error checks
  if(    year < 0 || year > 9999   // upper limit from cdf implementation (?)
      || doy < 1 || doy > (is_leap_year(year)?366:365)
      || hour < 0.0 || hour >= 24.0
      || min  < 0.0 || min  >= 60.0
      || sec  < 0.0 || sec  >= 60.0 )
     throw BadTime();

// calc month and day from doy

  int leap = leap_index(year);
  int month, day=doy;
  for(month=1; day > qie_days_in_month[leap][month]; month++ )
     day -= qie_days_in_month[leap][month];

  _trep = (double) cdf_julian_day (year, month, day) - epoch;
  _trep = _trep * 24.0 + hour;
  _trep = _trep * 60.0 + min;
  _trep = _trep * 60.0 + sec;
     
  return *this;
}

// ++++++++++++++++++++++++++++++++++++++++++++++++++++++ class Time
// ::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
// ---------------------------------------------------- second_of_day
double Time::second_of_day( void ) const
{
  double s=_trep;
  int d=(int)floor(s/86400.0);
  s-=d*86400.0;
  return s;
}
// ++++++++++++++++++++++++++++++++++++++++++++++++++++++ class Time
// ::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
// ---------------------------------------------------- day_of_year
double Time::day_of_year( void ) const
{
  int year, month, day,jd,i,j,k,l,n;
  
  int day_AD=(int)floor(_trep/86400.0);

  jd = (long) (epoch + day_AD);
  l=jd+68569;
  n=4*l/146097;
  l=l-(146097*n+3)/4;
  i=4000*(l+1)/1461001;
  l=l-1461*i/4+31;
  j=80*l/2447;
  k=l-2447*j/80;
  l=j/11;
  j=j+2-12*l;
  i=100*(n-49)+i+l;
  year = i;
  month = j;
  day=k;
  
  double doy=0.0;
  int leap = leap_index(year);
  for( int ii=0; ii<month; ii++ )
    doy += qie_days_in_month[leap][ii];

  doy+=day+(_trep-day_AD*86400.0)/86400.0; 

  return doy;
}

// ++++++++++++++++++++++++++++++++++++++++++++++++++++++ class Time
// ::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
// -------------------------------------------------------- split_dhms
void Time::split_dhms( int& d, int& h, int& m, double& s ) const
{
  s = _trep;                // seconds
  d = (int)floor(s/86400.0);     // whole number of days;
  s -= d*86400.0;           // seconds of day
  h = (int)(s/3600.0);      // whole number of hours
  s -= h*3600.0;            // seconds of hour
  m = (int)(s/60.0);        // whole number of minutes
  s -= m*60.0;              // seconds of minute
}

// ++++++++++++++++++++++++++++++++++++++++++++++++++++++ class Time
// ::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
// -------------------------------------------------------- split_ymd
// derived from EpochBreakdown in CDF 2.4.11 (epochu.c)

void Time::split_ymd( int& year, int& month, int& day,
                int& hour, int& min, int& sec, int& msec ) const
{
  int jd,i,j,k,l,n;
  double second_AD,msec_AD;
  int day_AD;
  split_dhms(day_AD,hour,min,second_AD);
  msec_AD=second_AD*1000+0.5;
  sec=(int)second_AD;
  
  jd = (long) (epoch + day_AD);
  l=jd+68569;
  n=4*l/146097;
  l=l-(146097*n+3)/4;
  i=4000*(l+1)/1461001;
  l=l-1461*i/4+31;
  j=80*l/2447;
  k=l-2447*j/80;
  l=j/11;
  j=j+2-12*l;
  i=100*(n-49)+i+l;
  year = i;
  month = j;
  day = k;
  
  msec   = (int) fmod (msec_AD, (double) 1000.0);
}
// ++++++++++++++++++++++++++++++++++++++++++++++++++++++ class Time
// ::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
// -------------------------------------------------------- split_ymd
// derived from EpochBreakdown in CDF 2.4.11 (epochu.c)

void Time::split_ymd( int& year, int& month, int& day,
                int& hour, int& min, double& sec ) const
{
  int jd,i,j,k,l,n;
  double second;
  int day_AD;
  split_dhms(day_AD,hour,min,sec);
  
  jd = (long) (epoch + day_AD);
  l=jd+68569;
  n=4*l/146097;
  l=l-(146097*n+3)/4;
  i=4000*(l+1)/1461001;
  l=l-1461*i/4+31;
  j=80*l/2447;
  k=l-2447*j/80;
  l=j/11;
  j=j+2-12*l;
  i=100*(n-49)+i+l;
  year = i;
  month = j;
  day = k;
}
// ++++++++++++++++++++++++++++++++++++++++++++++++++++++ class Time
// ::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
// -------------------------------------------------------- split_ymd
// derived from EpochBreakdown in CDF 2.4.11 (epochu.c)

void Time::split_ymd( int& year, int& month, int& day,
                int& hour, int& min, float& sec ) const
{
  double dsec;
  split_ymd( year, month, day, hour, min, dsec);
  sec=dsec;
}
// ++++++++++++++++++++++++++++++++++++++++++++++++++++++ class Time
// ::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
// -------------------------------------------------------- split_ymd
void Time::split_ymd( int& year, int& month, int& day ) const
{
  int h,m,s,ms;
  split_ymd( year, month, day, h, m, s, ms );
}

// ++++++++++++++++++++++++++++++++++++++++++++++++++++++ class Time
// ::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
// ------------------------------------------------------- split_ydoy
void Time::split_ydoy( int& year, int& doy,
                        int& hour, int& min, int& sec, int& msec ) const
{
  int month, day;
  split_ymd( year, month, day, hour, min, sec, msec );
  int leap = leap_index(year);
  doy=0;
  for( int i=0; i<month; i++ )
    doy += qie_days_in_month[leap][i];
  doy+=day;
}
// ++++++++++++++++++++++++++++++++++++++++++++++++++++++ class Time
// ::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
// ------------------------------------------------------- split_ydoy
void Time::split_ydoy( int& year, int& doy,
                        int& hour, int& min, double& sec ) const
{
  int month, day;
  split_ymd( year, month, day, hour, min, sec );
  int leap = leap_index(year);
  doy=0;
  for( int i=0; i<month; i++ )
    doy += qie_days_in_month[leap][i];
  doy+=day;
}
// ++++++++++++++++++++++++++++++++++++++++++++++++++++++ class Time
// ::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
// ------------------------------------------------------- split_ydoy
void Time::split_ydoy( int& year, int& doy,
                        int& hour, int& min, float& sec ) const
{
  double dsec;
  split_ydoy( year, doy, hour, min, dsec );
  sec=dsec;
}

// ++++++++++++++++++++++++++++++++++++++++++++++++++++++ class Time
// ::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
// ------------------------------------------------------- split_ydoy
void Time::split_ydoy( int& year, int& doy ) const
{
  int h,m,s,ms;
  split_ydoy( year, doy, h, m, s, ms );
}

// ++++++++++++++++++++++++++++++++++++++++++++++++++++++ class Time
// ::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
// --------------------------------------------------------- iso_srep
// iso string representation
string Time::iso_srep(const string& t_str, const string& z_str ) const
{
  static char s[100];
  int year, month, day, hour, min, sec, msec;
  year=hour=min=sec=0;
  month=day=msec=1;
  if(get_time_rep()>-6.e10)
    split_ymd( year, month, day, hour, min, sec, msec );
  sprintf( s, "%4.4d-%2.2d-%2.2d%s%2.2d:%2.2d:%2.2d.%3.3d%s",
    year, month, day, t_str.c_str(),hour, min, sec, msec, z_str.c_str() );
  return s;
}

// ++++++++++++++++++++++++++++++++++++++++++++++++++++++ class Time
// ::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
// ------------------------------------------------------- split_srep
// iso string representation
void Time::split_srep( vector<string>& str_components ) const
{
  static char s[100];
  int year, doy, month, day, hour, min, sec, msec;
  split_ymd( year, month, day, hour, min, sec, msec );
  split_ydoy( year, doy);
  str_components.clear();
  sprintf( s, "%4.4d", year   ); str_components.push_back( s );
  sprintf( s, "%3.3d", doy    ); str_components.push_back( s );
  sprintf( s, "%2.2d", month  ); str_components.push_back( s );
  str_components.push_back( full_month_names[month] );
  str_components.push_back( month_names[month] );
  sprintf( s, "%2.2d", day    ); str_components.push_back( s );
  sprintf( s, "%2.2d", hour   ); str_components.push_back( s );
  sprintf( s, "%2.2d", min    ); str_components.push_back( s );
  sprintf( s, "%2.2d", sec    ); str_components.push_back( s );
  sprintf( s, "%3.3d", msec   ); str_components.push_back( s );
}

// +++++++++++++++++++++++++++++++++++++++++++++++++ class QTInterval
// ::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
// ----------------------------------------------------- set_duration
void TimeInterval::set_duration(const Time& d, SetOption option )
{
  Sense sense=get_sense();
  switch(option)
  {
    case SAME_START_TIME:
      if(sense==TIME_SENSE_NEGATIVE)
        _t2=_t1-d;
      else
        _t2=_t1+d;
      break;
    case SAME_END_TIME:
      if(sense==TIME_SENSE_NEGATIVE)
        _t1=_t2+d;
      else
        _t1=_t2-d;
      break;
    case SAME_CENTER_TIME:
      if(sense==TIME_SENSE_NEGATIVE)
      { 
         _t2 = _t2 + 0.5*( _t1 - _t2 - d); 
         _t1 = _t2 + d;
      }
      else
      { 
        _t1 = _t1 + 0.5*( _t2 - _t1 - d); 
        _t2 = _t1 + d;
      }
      break;
    default:
      throw BadTimeIntervalOption("TimeInterval::set_duration(t,option)");
    break;
  }
}

// +++++++++++++++++++++++++++++++++++++++++++++++++ class QTInterval
// ::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
// ----------------------------------------------------- set_duration
void TimeInterval::set_duration( const Time& t, const Time& d, 
                               SetOption option )
{
  switch(option)
  {
    case START_TIME:
      _t1 = t; _t2 = _t1 + d;
      break;
    case END_TIME:
      _t2 = t; _t1 = _t2 - d;
      break;
    case CENTER_TIME:
      _t1 = t - 0.5*d; _t2 = _t1 + d;
    break;
    default:
      throw BadTimeIntervalOption("TimeInterval::set_duration(t1,t2,option)");
    break;
  }
}

// +++++++++++++++++++++++++++++++++++++++++++++++++ class QTInterval
// ::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
// --------------------------------------------------------- contains
bool TimeInterval::contains( const Time& t ) const
{
  bool res;
  Sense sense=get_sense();
  if(sense==TIME_SENSE_NEGATIVE)
    res=(t<=_t1 && _t2<=t);
  else
    res=(_t1<=t && t<=_t2);
  return res;
}
// +++++++++++++++++++++++++++++++++++++++++++++++++ class QTInterval
// ::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
// -------------------------------------------------------- int_union
TimeInterval TimeInterval::int_union( const TimeInterval& tint ) const
{
    if (empty())
        return tint;
    if(tint.empty())
        return *this;    
    return TimeInterval(min(start_abs(),tint.start_abs()),\
        max(end_abs(),tint.end_abs()));
}
// +++++++++++++++++++++++++++++++++++++++++++++++++ class QTInterval
// ::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
// ---------------------------------------------------- int_intersect
TimeInterval TimeInterval::int_intersect( const TimeInterval& tint ) const
{
    TimeInterval res;
    if((tint.start_abs()<end_abs())&& (tint.end_abs() >start_abs()))
        res.set(max(start_abs(),tint.start_abs()),\
            min(end_abs(),tint.end_abs()));
    return res;
}


} // end namespace QDOS

