// PROJECT: Cluster UK CDHF
//
// COMPONENT: Truncated Moment Corrector Plugin for QSAS
//
// MODULE: TruncaMomentCorrector.cc
// LANGUAGE: C++
//
// DESCRIPTION: Takes as input bottom, L1L2 and top truncated moments from PEACE, as well as
//supplementary data to perform a non-linear solving routine to infer the underlying 'correct' Maxwellian
//distribution
//
// ARGUMENTS: [passed via call_list] - See Template (qtpl) for details
//
// RETURN:
//    TYPE: QplugReturnStatus
//    VALUES: see qplug_if.h
//
// HISTORY:  Nora Bounaira  Nov 2003 - First Version, subsequent development Jim Geach

#include "truncated_moments_corrector.h"
#include "tmc_routines.h"
#include <stdio.h>
#include <math.h>
#include <string.h>
#include <stdlib.h>
#include "qgu.h"
#include "qexuiW.h"
#include "Interp.h"
#include "qjui_multijoin.h"
#include "qsas.h"
#include "qplug_if.h"
#include "qdutil.h"
#include "Qdos.h"
#include "qar.h"
#include "Xrefs.h"    

#define DEBUG_LOG

using namespace QSAS;

// Make the function accessible from external modules 
extern "C" QplugReturnStatus QarTruncMomCorrector2( QplugArgList *call_list );

QplugReturnStatus Qartmc_plugin( QplugArgList *call_list )
{
    cout << "\nQSAS - TRUNCATED MOMENTS CORRECTOR PLUGIN" << '\n';      
    cout << "-----------------------------------------" << '\n';
    cout << "         QMUL CLUSTER TEAM 2004           " << '\n';
    cout << endl;
    
    TRUNC_STRUCT tsr;        // Structure to perform the calculations 
    int          nSeq       = 0;  
    double  LowerGuess, UpperGuess, GuessResolution, Phi_offset;
    int SC = 1;
    QplugAppendTextDisplay( "TMC>> EPlug-in version 1.2, Feb 2004\n");
    nSeq = call_list->size();
    
    if ( nSeq < 50 )
    {
	QplugAppendTextDisplay( "TMC>> ERROR Not enough arguments from GUI\n");
	return QPLUG_FAILURE;
    } 
    
    cout << "Testing for B moments... ";
    bool USE_B, USE_T;
    USE_B = use_B_moments_at_all(call_list);
    USE_T = use_T_moments_at_all(call_list);
    
    if(USE_B == TRUE)
	cout << "B moments present." << endl;
    else
	cout << "B moments not-present." << endl;
    
    cout << "Testing for T moments... ";
    if(USE_T == TRUE)
	cout << "T moments present." << endl;
    else
	cout << "T moments not-present." << endl;
    
    if(USE_B==TRUE && USE_T==TRUE)
	nSeq -= 9;
    else if(USE_B==FALSE && USE_T==FALSE)
	nSeq -= 33;
    else if(USE_B==TRUE && USE_T==FALSE)
        nSeq -= 21;
    else //if(USE_B==FALSE && USE_T==TRUE)
        nSeq -=	21;
    
    
    cout << "Separating L1L2 moments." << endl;
    
    int nSize = ( *call_list )[0]->sequence_size();
    
    QdObjectSeq_var LEEAsubset = new QdObjectSeq( 13 ); 
    
    SequenceMask LEEAmask((*call_list)[0]->sequence_size(), true);
    
    QdRScalarSeq_var spinmode = QdRScalarSeq_var::narrow( (*call_list)[ 41 ] );
    
    //Assign Booleans
    for(int i=0; i<(*call_list)[0]->sequence_size(); i++)
    {
	if((int)spinmode[i]==0) 
	    LEEAmask[i]=true;
	else
	    LEEAmask[i]=false;
    }
    
    cout << "Summing L1+L2 moments." << endl;  
    
    if (SumL1L2Moms( call_list, LEEAsubset, LEEAmask ) != QPLUG_SUCCESS)
	return (QPLUG_FAILURE);
    
        QdObjectSeq_var joined = new QdObjectSeq( nSeq );   // Set of time series being joined
    cout << "Joining data." << endl;
    if ( JoinInputSeries( nSeq, call_list, LEEAsubset, joined, USE_B, USE_T ) != QPLUG_SUCCESS )
    {
	return ( QPLUG_FAILURE );
    }
    
    nSize = joined[ 0 ]->sequence_size();
    cout << nSize << endl;
    // Create the output objects and time-stamp them
    cout << "Creating output objects." << endl;  
    QdTimeSeq_var tLine = get_timetags( joined[ 0 ] );
    if ( tLine.is_nil() )
    {
	QplugAppendTextDisplay ("TMC>> ERROR - No time line associated with objects after Join\n");
	return ( QPLUG_FAILURE );
    }
        
    QdRScalarSeq_var  rDensity0 = new QdRScalarSeq( nSize );
    if ( rDensity0.is_nil() )
    {
	QplugAppendTextDisplay ("TMC>> ERROR - Cannot allocate memory for Density output\n");
	return QPLUG_FAILURE;
    }
    
    QdRMatrixSeq_var  rVelocity0 = new QdRMatrixSeq( nSize, Dim(3) );
    if ( rVelocity0.is_nil() )
    {
	QplugAppendTextDisplay ("TMC>> ERROR - Cannot allocate memory for velocity output\n");
	return QPLUG_FAILURE;
    }
    
    QdRScalarSeq_var  rSpeed = new QdRScalarSeq( nSize );
    if ( rSpeed.is_nil() )
    {
	QplugAppendTextDisplay ("TMC>> ERROR - Cannot allocate memory for thermal speed output\n");
	return QPLUG_FAILURE;
    }
    
    QdRScalarSeq_var  rTemperature = new QdRScalarSeq( nSize );
    if ( rTemperature.is_nil() )
    {
	QplugAppendTextDisplay ("TMC>> ERROR - Cannot allocate memory for temperature output\n");
	return QPLUG_FAILURE;
    }
    
    QdRScalarSeq_var  rToPerp  = new QdRScalarSeq( nSize );
    if ( rToPerp.is_nil() )
    {
	QplugAppendTextDisplay ("TMC>> ERROR - Cannot allocate memory for T0 Perpendicular output\n");
	return QPLUG_FAILURE;
    }
    
    QdRScalarSeq_var  rToParal = new QdRScalarSeq( nSize );
    if ( rToPerp.is_nil() )
    {
	QplugAppendTextDisplay ("TMC>> ERROR - Cannot allocate memory for T0 Parallel output\n");
	return QPLUG_FAILURE;
    }
    
    QdRScalarSeq_var  rTratio = new QdRScalarSeq( nSize );
    if ( rTratio.is_nil() )
    {
	QplugAppendTextDisplay ("TMC>> ERROR - Cannot allocate memory for Tr output\n");
	return QPLUG_FAILURE;
    }
    
    QdRScalarSeq_var  rNratio = new QdRScalarSeq( nSize );
    if ( rNratio.is_nil() )
    {
	QplugAppendTextDisplay ("TMC>> ERROR - Cannot allocate memory for Nr output\n");
	return QPLUG_FAILURE;
    }
    
    QdRScalarSeq_var  rVratio = new QdRScalarSeq( nSize );
    if ( rVratio.is_nil() )
    {
	QplugAppendTextDisplay ("TMC>> ERROR - Cannot allocate memory for Vr output\n");
	return QPLUG_FAILURE;
    }
    
    QdRScalarSeq_var  rIe = new QdRScalarSeq( nSize );
    if ( rIe.is_nil() )
    {
	QplugAppendTextDisplay ("TMC>> ERROR - Cannot allocate memory for Ie output\n");
	return QPLUG_FAILURE;
    }
    
    QdRScalarSeq_var  rRetValue = new QdRScalarSeq( nSize );
    if ( rRetValue.is_nil() )
    {
	QplugAppendTextDisplay ("TMC>> ERROR - Cannot allocate memory for the solver ret value\n");
	return QPLUG_FAILURE;
    }
    
    QdRScalarSeq_var  rp_off_on = new QdRScalarSeq( nSize );
    if ( rp_off_on.is_nil() )
    {
	QplugAppendTextDisplay ("TMC>> ERROR - Cannot allocate memory for the p_off_on value\n");
	return QPLUG_FAILURE;
    }
        
    QdRScalarSeq_var  rp_perp_ratio = new QdRScalarSeq( nSize );
    if ( rp_perp_ratio.is_nil() )
    {
	QplugAppendTextDisplay ("TMC>> ERROR - Cannot allocate memory for the p_perp_ratio value\n");
	return QPLUG_FAILURE;
    }
    
    
    QdRScalar_var QdLowerGuess = QdRScalar_var::narrow((*call_list)[42]);
    if ( QdLowerGuess.is_nil() )
    {
	QplugAppendTextDisplay ("TMC>> ERROR - Cannot allocate memory for the solver lower guess\n");
	return QPLUG_FAILURE;
    }
    
    QdRScalar_var QdUpperGuess = QdRScalar_var::narrow((*call_list)[43]);
    if ( QdUpperGuess.is_nil() )
    {
	QplugAppendTextDisplay ("TMC>> ERROR - Cannot allocate memory for the solver upper guess\n");
	return QPLUG_FAILURE;
    }
    
    QdRScalar_var QdGuessResolution = QdRScalar_var::narrow((*call_list)[44]);
    if ( QdGuessResolution.is_nil() )
    {
	QplugAppendTextDisplay ("TMC>> ERROR - Cannot allocate memory for the solver guess resolution\n");
	return QPLUG_FAILURE;
    }
    
    QdRScalar_var QdPhiOffset = QdRScalar_var::narrow((*call_list)[45]);
    if ( QdPhiOffset.is_nil() )
    {
	QplugAppendTextDisplay ("TMC>> ERROR - Cannot allocate memory for the phi offset\n");
	return QPLUG_FAILURE;
    }
    
    QdRScalar_var QdSC = QdRScalar_var::narrow((*call_list)[48]);
    if ( QdSC.is_nil() )
    {
	QplugAppendTextDisplay ("TMC>> ERROR - Cannot allocate memory for the phi offset\n");
	return QPLUG_FAILURE;
    }
    
    QdRScalar_var QdHL = QdRScalar_var::narrow((*call_list)[49]);
    if ( QdHL.is_nil() )
    {
	QplugAppendTextDisplay ("TMC>> ERROR - Cannot allocate memory for the HEEA or LEEA choice\n");
	return QPLUG_FAILURE;
    }
    
    int HorL;
    
    
    
    LowerGuess = (double)(*QdLowerGuess);
    UpperGuess = (double)(*QdUpperGuess);
    GuessResolution = (double)(*QdGuessResolution);
    Phi_offset = (double)(*QdPhiOffset);
    SC = (int)(*QdSC);
    HorL = (int)(*QdHL);
    
    QdRScalarSeq_var Lcalobj, Hcalobj;
    string Lcalname, Hcalname;
    if(SC==1)
    	{Lcalname = "CL1_LEEA.cal"; Hcalname = "CL1_HEEA.cal";}
    else if(SC==2)
    	{Lcalname = "CL2_LEEA.cal"; Hcalname = "CL2_HEEA.cal";}    
    else if(SC==3)
	{Lcalname = "CL3_LEEA.cal"; Hcalname = "CL3_HEEA.cal";}
    else
	{Lcalname = "CL4_LEEA.cal"; Hcalname = "CL4_HEEA.cal";}
    
    Lcalobj = read_n_rtn_cal_file(Lcalname);
    Hcalobj = read_n_rtn_cal_file(Hcalname);
	

    
    cout << "Getting object attributes." << endl;
    rDensity0->copy_xrefs_from( (* call_list)[0] );
    rVelocity0->copy_xrefs_from( (* call_list)[40] );
    rSpeed->copy_xrefs_from( (* call_list)[0] );
    rTemperature->copy_xrefs_from( (* call_list)[0] );
    rToPerp->copy_xrefs_from( (* call_list)[0] );
    rToParal->copy_xrefs_from( (* call_list)[0] );
    rRetValue->copy_xrefs_from( (* call_list)[0] );
    
    // The specific ones should be set individually
    SetDensitySpecificAttributes( rDensity0 );
    SetVelocitySpecificAttributes( rVelocity0 );
    SetSpeedSpecificAttributes( rSpeed );
    SetTemperatureSpecificAttributes( rTemperature );
    SetPerpenTemperatureSpecificAttributes( rToPerp );
    SetParallelTemperatureSpecificAttributes( rToParal );
    SetRetValueSpecificAttributes( rRetValue );
    SetrTSpecificAttributes( rTratio );
    SetrNSpecificAttributes( rNratio );
    SetrVSpecificAttributes( rVratio );
    SetIeSpecificAttributes( rIe );
    
    cout << "Setting time-tags." << endl;
    // Set time tags
    set_timetags( rDensity0, tLine );
    set_timetags( rVelocity0, tLine );
    set_timetags( rSpeed, tLine );
    set_timetags( rTemperature, tLine );
    set_timetags( rToPerp, tLine );
    set_timetags( rToParal, tLine );
    set_timetags( rRetValue, tLine );
    set_timetags( rTratio, tLine );
    set_timetags( rNratio, tLine );
    set_timetags( rVratio, tLine );
    set_timetags( rIe, tLine );
    set_timetags( rp_off_on, tLine );
    set_timetags( rp_perp_ratio, tLine );
    
    // cout << rDensity0->sequence_size() << endl; 
    
    if ( PerformCalculation(  nSize,
			     call_list,
			     joined,
			     LowerGuess,
			     UpperGuess,
			     GuessResolution,
			     rDensity0,
			     rVelocity0,
			     rSpeed,
			     rTemperature,
			     rToPerp,
			     rToParal,
			     rTratio,
			     rNratio,
			     rVratio,
			     rIe,
			     rRetValue, rp_off_on, rp_perp_ratio, USE_B, USE_T, Phi_offset, SC, HorL,  Lcalobj, 
    Hcalobj, "", "debug.log") != QPLUG_SUCCESS )
    {
	QplugAppendTextDisplay ("TMC>> ERROR - Cannot perform calculation on input objects\n");
	return QPLUG_FAILURE;
    }
 
    QplugAppendTextDisplay ("TMC>> Calculation Complete\n");
    
    // ( 9 ) Add output object to  QSAS working list
    call_list->push_back( (QdObject_var) rDensity0 );
    call_list->push_back( (QdObject_var) rVelocity0 );
    call_list->push_back( (QdObject_var) rSpeed );
    call_list->push_back( (QdObject_var) rTemperature );
    call_list->push_back( (QdObject_var) rToParal );
    call_list->push_back( (QdObject_var) rToPerp );
    call_list->push_back( (QdObject_var) rTratio );
    call_list->push_back( (QdObject_var) rNratio );
    call_list->push_back( (QdObject_var) rVratio );
    call_list->push_back( (QdObject_var) rIe );
    call_list->push_back( (QdObject_var) rRetValue );
    call_list->push_back( (QdObject_var) rp_off_on );
    call_list->push_back( (QdObject_var) rp_perp_ratio );
    
    return ( QPLUG_SUCCESS );
}

