/* tmnewt.c
*
*  sjs modified NR routines for truncated_moments_corrector
*
*  Need to pass various params through to nonlinear function solver
*
*  sjs July 2003
*/

#include "truncated_moments_corrector.h" /* safe against multiple includes */
                                         /* needed for TRUNCT_STRUCT */

#include <math.h>
#include <stdlib.h>
#include <stdio.h>


#define FREE {free_dmatrix(fjac,1,n,1,n);free_dvector(fvec,1,n);free_dvector(p,1,n);free_ivector(indx,1,n);} 


int tmnewt(int ntrial, double x[], int n, double tolx, double tolf, TRUNC_STRUCT *tstruct){ 
/* Given an initial guess x[1..n] for a root in n dimensions, take ntrial 
   Newton-Raphson steps to improve the root. Stop if the root converges in 
   either summed absolute variable increments tolx or summed absolute 
   function values tolf. Returns NR_OK on success; 
   NR_ERROR otherwise */

 int k,i,*indx, err;
 double errx,errf,d,*fvec,**fjac,*p;
 indx	= ivector(1,n);
 p	= dvector(1,n);
 fvec	= dvector(1,n);
 fjac	= dmatrix(1,n,1,n);
 for (k=1; k<=ntrial; k++) { 
    tusrfun(x,n,fvec,fjac,tstruct); 	       /* User function supplies function values at x in fvec and Jacobian matrix in fjac.*/
    errf=0.0;
   for (i=1; i<=n; i++) errf += fabs(fvec[i]); /* Check function convergence. */ 
 if (errf <= tolf) {FREE return NR_OK;}
 for (i=1; i<=n; i++) p[i] = -fvec[i];         /* Right-hand side of linear equations.*/
 err= ludcmp(fjac,n,indx,&d); 	               /* Solve linear equations using LU decomposition.*/ 
 if (err != NR_OK) return TMNEWT_LU_ERR;
 lubksb(fjac,n,indx,p);
 errx=0.0;                                     /* Check root convergence. */
 
 for (i=1; i<=n; i++) {                        /* Update solution. */
    errx += fabs(p[i]);
    x[i] += p[i]; }
     
 if (errx <= tolx) {FREE return NR_OK;} 
 
 /* jgeach - M0910 -- added a little extra diagnostic here: output x[1,2,3] to check values all the way through. */
 /* is there a way to see whether things are going wrong? i.e. ridiculous x's, or massive errx? 		 */
 
/* if (DIAG_TMNEWT) printf("mnewt trial %d. errx = %g, errf = %g, x[1] = %g, x[2] = %g, x[3] = %g\n",k,errx,errf,x[1],x[2],x[3]);*/ 
 }
 FREE return TMNEWT_TRIALS_EXHAUSTED; 
}

int ludcmp(double **a, int n, int *indx, double *d) 
/* Given a matrix a[1..n][1..n],this routine replaces t by the LU 
decomposition of a roww se permutation of itself.a and n are input.a
s output,arranged as in equation (2.3.14)above; 
indx[1..n] is an output vector that records the row permutation effected
by partial pivoting; d is output as  1 depending on whether the number 
of row nterchanges was even or odd,respectively.This routine is used in 
combination with lubksb to solve linear equations or invert a matrix.
Returns NR_OK on success; NR_ERROR otherwise */
{ 

int i,imax,j,k; 
double big,dum,sum,temp; 
double *vv;               /* vv stores the implicit scaling of each row. */

vv=dvector(1,n); 
*d=1.0;                   /* No row interchanges yet.*/ 
for (i=1;i<=n;i++) { 
                	  /* Loop over rows to get the implicit scaling information.*/
    big=0.0; 
    for (j=1;j<=n;j++) 
        if ((temp=fabs(a[i][j])) > big) big=temp; 
    if (big == 0.0) { 
                          /* nrerror("Singular matrix in routine ludcmp"); */
       return NR_ERROR;
    } 
                      	  /* No nonzero largest element. */
    vv[i]=1.0/big;        /* Save the scaling.*/ 
} 
for (j=1;j<=n;j++) {      /* This is the loop over columns of Crout's method. */
    for (i=1;i<j;i++) {   /* This is equation (2.3.12)except for i = j . */
	sum=a[i][j]; 
	for (k=1;k<i;k++) sum -= a[i][k]*a[k][j]; 
	a[i][j]=sum; 
    } 
    big=0.0;               /* Initialize for the search for largest pivot element. */
    for (i=j;i<=n;i++) {   /* This is i = j of equation (2.3.12) and 
                              i = j +1 ..N of equation (2.3.13).*/
      sum=a[i][j]; 
      for (k=1;k<j;k++)  sum -= a[i][k]*a[k][j]; 
      a[i][j]=sum; 
      if ( (dum=vv[i]*fabs(sum)) >= big) { 
                           /* Is the figure of merit for the pivot better than the best so far? */
	 big=dum;
	 imax=i; 
      } 
     } 
     if (j != imax) {      /* Do we need to nterchange rows? */
	 for (k=1;k<=n;k++) { /* Yes, do so... */
	     dum=a[imax][k]; 
	     a[imax][k]=a[j][k]; 
	     a[j][k]=dum; 
	 } 
	 *d = -(*d);       /* ...and change the parity of d */ 
	 vv[imax]=vv[j];   /* Also interchange the scale factor.*/ 
     } 
     indx[j]=imax; 
     if (a[j][j] == 0.0) a[j][j]=TINY; /* If the pivot element is zero 
                  the matrix is singular (at least to the precision 
		  of the algorithm).For some applications on singular 
		  matrices, it is desirable to substitute TINY for zero. */
     if (j != n) {         /* Now, finally,divide by the pivot element. */
	 dum=1.0/(a[j][j]); 
	 for (i=j+1;i<=n;i++) a[i][j] *= dum; 
     } 
}                          /*Go back for the next column n the reduction. */
free_dvector(vv,1,n); 
return NR_OK;
}      

void lubksb(double **a, int n, int *indx, double b[]) 
/*  Solves the set of n linear equations A  X = B .Here a[1..n][1..n] is 
input, not as the matrix A but rather as its LU decomposition, determined 
by the routine ludcmp indx[1..n] is input as the permutation vector 
returned by ludcmp b[1..n] is input as the right-hand side vector B ,and 
returns with the solution vector X,a, n and indx are not modified by this 
routine and can be left in place for successive calls with different right-hand
sides b. This routine takes into account the possibility that b will begin 
with many zero elements, so it is efficient for use in matrix inversion.
*/
{   

int i,ii=0,ip,j; 
double sum; 

for (i=1;i<=n;i++) { /* When ii is set to a positive value, it will become 
                     the index of the first nonvanishing element of b. 
		     We now do the forward substitution,equation (2.3.6).
		     The only new wrinkle is to unscramble the permutation 
		     as we go. */
    ip=indx[i]; 
    sum=b[ip]; 
    b[ip]=b[i]; 
    if (ii) for (j=ii;j<=i-1;j++) sum -= a[i][j]*b[j]; 
    else if (sum) ii=i;  /*A nonzero element was encountered, so from now on 
                	  we will have to do the sums n the loop above.*/
    b[i]=sum; 
} 
for (i=n;i>=1;i--) { /* Now we do the backsubstitution,equation (2.3.7).*/ 
   sum=b[i]; 
   for (j=i+1;j<=n;j++) sum -= a[i][j]*b[j]; 
   b[i]=sum/a[i][i]; /* Store a component of the solution vector X . */
} 
} 


double *dvector(long nl, long nh) 
/* allocate a double vector with subscript range v[nl..nh] */ 
{ 
double *v;
v=(double *)malloc((size_t) ((nh-nl+1+NR_END)*sizeof(double))); 
if (!v) {
   nrerror("allocation failure in vector()"); 
   return NULL;
}
return v-nl+NR_END; 
} 

int *ivector(long nl, long nh) 
/* allocate an int vector with subscript range v[nl..nh] */ 
{ 
int *v; v=(int *)malloc((size_t) ((nh-nl+1+NR_END)*sizeof(int))); 
if (!v) {
   nrerror("allocation failure in ivector()"); 
   return NULL;
}
return v-nl+NR_END; 
} 


double **dmatrix(long nrl, long nrh, long ncl, long nch) 
/* allocate a double matrix with subscript range m[nrl..nrh][ncl..nch] */ 
{ 
long i, nrow=nrh-nrl+1,ncol=nch-ncl+1; 
double **m; 
/* allocate pointers to rows */ 
m=(double **) malloc((size_t)((nrow+NR_END)*sizeof(double*))); 
if (!m) {
    nrerror("allocation failure 1 in matrix()");
    return NULL;
} 
m += NR_END; 
m -= nrl; 

/* allocate rows and set pointers to them */ 
m[nrl]=(double *) malloc((size_t)((nrow*ncol+NR_END)*sizeof(double)));
if (!m[nrl]) {
    nrerror("allocation failure 2 in matrix()"); 
    return NULL;
}
m[nrl] += NR_END; 
m[nrl] -= ncl; 
for(i=nrl+1;i<=nrh;i++) m[i]=m[i-1]+ncol; 
/* return pointer to array of pointers to rows */ 
return m; 
}

void free_ivector(int *v, long nl, long nh)
/* free an int vector allocated with ivector() */ 
{ 
free((FREE_ARG) (v+nl-NR_END)); 
}

void free_dvector(double *v, long nl, long nh) 
/* free a double vector allocated with dvector() */ 
{ 
free((FREE_ARG) (v+nl-NR_END)); 
}

void free_dmatrix(double **m, long nrl, long nrh, long ncl, long nch) 
/* free a double matrix allocated by dmatrix() */ 
{
free((FREE_ARG) (m[nrl]+ncl-NR_END));
free((FREE_ARG) (m+nrl-NR_END));
}

void nrerror(char error_text[])
{
printf("Error in NR-related routine: %s\n",error_text);
return;
}

double tqgauss100(double (*func)(double, double, double, double, double),
     double a,double b, double epssc, double v0, double vt_sq, double zeta){
int j;
double dx,xm,xr,ss;

static double w[]={0.031255424e0,0.0312248841e0,0.0311638359e0,0.0310723372e0,
     	0.0309504792e0,  0.0307983793e0,  0.0306161866e0,  0.0304040797e0,
     	0.0301622655e0,  0.0298909787e0,  0.0295904875e0,  0.0292610843e0,
     	0.0289030895e0,  0.0285168551e0,  0.0281027555e0,  0.0276611988e0,
     	0.0271926131e0,  0.0266974587e0,  0.0261762198e0,  0.0256294031e0,
     	0.0250575449e0,  0.0244612023e0,  0.0238409601e0,  0.0231974237e0,
     	0.0225312207e0,  0.0218430031e0,  0.0211334415e0,  0.0204032324e0,
     	0.0196530875e0,  0.0188837405e0,  0.0180959404e0,  0.0172904599e0,
     	0.0164680853e0,  0.0156296212e0,  0.0147758843e0,  0.013907711e0,
     	0.0130259478e0,  0.0121314572e0,  0.0112251136e0,  0.0103078028e0,
     	0.00938041974e0, 0.00844387151e0, 0.00749907317e0, 0.00654694857e0,
     	0.00558842812e0, 0.00462445011e0, 0.00365596125e0, 0.00268392544e0,
     	0.00170939264e0, 0.000734634465e0};

static double x[]={0.0156289842e0,0.0468716808e0,0.0780685842e0,0.109189205e0,
     	0.140203133e0,  0.171080083e0,  0.201789871e0,  0.232302487e0,
     	0.262588114e0,  0.292617202e0,  0.322360337e0,  0.351788521e0,
     	0.380872995e0,  0.409585297e0,  0.437897414e0,  0.465781659e0,
     	0.493210793e0,  0.520157993e0,  0.546597004e0,  0.572501957e0,
     	0.597847462e0,  0.62260884e0,   0.646761894e0,  0.67028302e0,
     	0.693149209e0,  0.715338111e0,  0.736828089e0,  0.757598102e0,
     	0.777627885e0,  0.796897888e0,  0.815389216e0,  0.833083868e0,
     	0.849964499e0,  0.866014659e0,  0.881218672e0,  0.895561635e0,
     	0.909029543e0,  0.921609282e0,  0.933288515e0,  0.944055855e0,
     	0.953900754e0,  0.962813675e0,  0.970785797e0,  0.97780937e0,
     	0.98387754e0,   0.988984406e0,  0.993124962e0,  0.996295154e0,
     	0.998491943e0,  0.999713719e0};

      xm=0.5e0*(b+a);
      xr=0.5e0*(b-a);
      ss=0.e0;
      for(j=0; j< 50; j++){
        dx=xr*x[j];
        ss+=w[j]*((*func)(xm+dx, epssc, v0, vt_sq, zeta) +
	          (*func)(xm-dx, epssc, v0, vt_sq, zeta));
      }
      ss *=xr;
      return ss;
}

double split_tqgauss100(int N, double (*func)(double, double, double, double, double),
     double a,double b, double epssc, double v0, double vt_sq, double zeta)
{

   int i=1;
   double sum=0.;
   double step = (b-a)/N;

   while( (a+i*step) <= b )
   {
    //   printf("i = %d\tfrom: %g\tto: %gsum = %g\n",i, (a+(i-1)*step), (a+i*step),sum);
       sum += tqgauss100(func, (a+(i-1)*step), (a+i*step), epssc, v0, vt_sq, zeta);
       i++;
   }

   return sum;

}

/* Addapted from VG Fortran routine by SJS April 2003
C Adapted (VG 03/2001) from (C) Copr. 1986-92 Numerical Recipes Software 888.
      SUBROUTINE qgaus100(func,a,b,ss)s
      DOUBLE PRECISION a,b,ss,func
      EXTERNAL func
      INTEGER j
      DOUBLE PRECISION dx,xm,xr,w(50),x(50)
      SAVE w,x
     	DATA w/0.031255424d0,0.0312248841d0,0.0311638359d0,0.0310723372d0,
     $	0.0309504792d0,  0.0307983793d0,  0.0306161866d0,  0.0304040797d0,
     $	0.0301622655d0,  0.0298909787d0,  0.0295904875d0,  0.0292610843d0,
     $	0.0289030895d0,  0.0285168551d0,  0.0281027555d0,  0.0276611988d0,
     $	0.0271926131d0,  0.0266974587d0,  0.0261762198d0,  0.0256294031d0,
     $	0.0250575449d0,  0.0244612023d0,  0.0238409601d0,  0.0231974237d0,
     $	0.0225312207d0,  0.0218430031d0,  0.0211334415d0,  0.0204032324d0,
     $	0.0196530875d0,  0.0188837405d0,  0.0180959404d0,  0.0172904599d0,
     $	0.0164680853d0,  0.0156296212d0,  0.0147758843d0,  0.013907711d0,
     $	0.0130259478d0,  0.0121314572d0,  0.0112251136d0,  0.0103078028d0,
     $	0.00938041974d0,  0.00844387151d0,0.00749907317d0,0.00654694857d0,
     $	0.00558842812d0,  0.00462445011d0,0.00365596125d0,0.00268392544d0,
     $	0.00170939264d0,  0.000734634465d0/
     
	DATA x/0.0156289842d0,0.0468716808d0,0.0780685842d0,0.109189205d0,
     $	0.140203133d0,  0.171080083d0,  0.201789871d0,  0.232302487d0,
     $	0.262588114d0,  0.292617202d0,  0.322360337d0,  0.351788521d0,
     $	0.380872995d0,  0.409585297d0,  0.437897414d0,  0.465781659d0,
     $	0.493210793d0,  0.520157993d0,  0.546597004d0,  0.572501957d0,
     $	0.597847462d0,  0.62260884d0,  0.646761894d0,  0.67028302d0,
     $	0.693149209d0,  0.715338111d0,  0.736828089d0,  0.757598102d0,
     $	0.777627885d0,  0.796897888d0,  0.815389216d0,  0.833083868d0,
     $	0.849964499d0,  0.866014659d0,  0.881218672d0,  0.895561635d0,
     $	0.909029543d0,  0.921609282d0,  0.933288515d0,  0.944055855d0,
     $	0.953900754d0,  0.962813675d0,  0.970785797d0,  0.97780937d0,
     $	0.98387754d0,  0.988984406d0,  0.993124962d0,  0.996295154d0,
     $	0.998491943d0,  0.999713719d0/
     
      xm=0.5d0*(b+a)
      xr=0.5d0*(b-a)
      ss=0.d0
      do 11 j=1,50
        dx=xr*x(j)
        ss=ss+w(j)*(func(xm+dx)+func(xm-dx))
11    continue
      ss=xr*ss
      return
      END
C Adapted (VG 03/2001) from (C) Copr. 1986-92 Numerical Recipes Software 888.
*/


