/*********************************************************************************************************************************
*
*  truncated_moments_corrector.c
*
*  Routines to take as input a set of truncated particle moments and
*  estimate the properties of the full population based on a truncated
*  Maxwellian, including the effects of spacecraft potential
*
*  SJ Schwartz, July 2003
*
*  Modified J.Geach  06/10/03 -- working on 'first guess' routine
*		     09/10/03 -- found decent initial guesses based on regression, building into streamlined routine
*		     10/10/03 -- updates
*		     14/10/03 -- updates
*		     03/11/03 -- added sc_to_gse routine to convert between coord systems
*		     04/11/03 -- added more routines to deal with tensor and vector rotations to gse
*		     05/11/03 -- added modifier methods for updating measured and associated values of TRUNC_STRUCT
*		     14/11/03 -- added/improved error codes, modified initial guesses
*		     19/11/03 -- fixed anisotropy bug - yet still get -ve Asc sometimes... investigating
***********************************************************************************************************************************/

#include "truncated_moments_corrector.h"


double GUESS_ONE = 1.0;            /* guess one */
double GUESS_TWO = 0.65;           /* guess two */
double GUESS_THREE = 0.41;         /* guess three */

void copy_trunc_struct_maxwellians(TRUNC_STRUCT *from, TRUNC_STRUCT *to)
{

    to->n0 = from->n0;
    to->T0 = from->T0;
    to->v0 = from->v0;
    to->vxsc = from->vxsc;
    to->vysc = from->vysc;
    to->vzsc = from->vzsc;
    to->vT = from->vT;

return;
}

double x_cross_y(double x[3], double y[3], double n[3])
{

n[0] = x[1]*y[2] - x[2]*y[1];
n[1] = -(x[0]*y[2] - x[2]*y[0]);
n[2] = x[0]*y[1] - x[1]*y[0];

return sqrt(n[0]*n[0] + n[1]*n[1] + n[2]*n[2]);

}

void special_transform_vector(double v[3], double b[3])
{

int i;
double R[3][3];
double b_cross_v[3];
double e1_cross_e2[3];
double v_trans[3];
double norm1, norm2;
double bmag = sqrt(b[0]*b[0] + b[1]*b[1] + b[2]*b[2]);
double b_unit[3];

for(i=0; i<3; i++)
  b_unit[i] = b[i]/bmag;

for(i=0; i<3; i++)
  b[i] = b_unit[i];

norm1 = x_cross_y(b, v, b_cross_v);

for(i=0; i<3; i++)
b_cross_v[i] /= norm1;

norm2 = x_cross_y(b, b_cross_v, e1_cross_e2);

for(i=0; i<3; i++)
e1_cross_e2[i] /= norm2;


R[0][0] = b[0]; R[0][1] = b[1];       R[0][2] = b[2];
R[1][0] = b_cross_v[0];	          R[1][1] = b_cross_v[1];	R[1][2] = b_cross_v[2];
R[2][0] = e1_cross_e2[0];	          R[2][1] = e1_cross_e2[1];	        R[2][2] = e1_cross_e2[2];

v_trans[0] = R[0][0]*v[0] + R[0][1]*v[1] + R[0][2]*v[2];
v_trans[1] = R[1][0]*v[0] + R[1][1]*v[1] + R[1][2]*v[2];
v_trans[2] = R[2][0]*v[0] + R[2][1]*v[1] + R[2][2]*v[2];

/* reassign */
v[0] = v_trans[0];  v[1] = v_trans[1];  v[2] = v_trans[2];
}

void special_transform_tensor(double v[3], double b[3], double T[3][3])
{

int i, j, k;
double R[3][3], trnsR[3][3], Z[3][3], Y[3][3];
double b_cross_v[3];
double e1_cross_e2[3];
double v_trans[3];
double norm1, norm2;
double bmag = sqrt(b[0]*b[0] + b[1]*b[1] + b[2]*b[2]);
double b_unit[3];

for(i=0; i<3; i++)
  b_unit[i] = b[i]/bmag;

for(i=0; i<3; i++)
  b[i] = b_unit[i];

norm1 = x_cross_y(b, v, b_cross_v);

for(i=0; i<3; i++)
b_cross_v[i] /= norm1;

norm2 = x_cross_y(b, b_cross_v, e1_cross_e2);

for(i=0; i<3; i++)
e1_cross_e2[i] /= norm2;


R[0][0] = b[0];                 R[0][1] = b[1];	                R[0][2] = b[2];
R[1][0] = b_cross_v[0];	        R[1][1] = b_cross_v[1]; 	R[1][2] = b_cross_v[2];
R[2][0] = e1_cross_e2[0];	R[2][1] = e1_cross_e2[1];	R[2][2] = e1_cross_e2[2];


/* get transpose */
trnsR[0][0] = R[0][0];	trnsR[0][1] = R[1][0];	trnsR[0][2] = R[2][0];
trnsR[1][0] = R[0][1];	trnsR[1][1] = R[1][1];	trnsR[1][2] = R[2][1];
trnsR[2][0] = R[0][2];	trnsR[2][1] = R[1][2];	trnsR[2][2] = R[2][2];

/* do the first part of the transformation */
for (i = 0; i<3; i++){ 
	for (j=0; j<3; j++){ Z[i][j] = 0.0; 
		for (k=0; k<3; k++){ Z[i][j] += T[i][k]*trnsR[k][j];}}}

/* do the second part of the transformation */
for (i = 0; i<3; i++){ 
	for (j=0; j<3; j++){ Y[i][j] = 0.0; 
		for (k=0; k<3; k++){ Y[i][j] += R[i][k]*Z[k][j];}}}

/* reassign */
T[0][0] = Y[0][0];	T[0][1] = Y[0][1];	T[0][2] = Y[0][2];
T[1][0] = Y[1][0];	T[1][1] = Y[1][1];	T[1][2] = Y[1][2];
T[2][0] = Y[2][0];	T[2][1] = Y[2][1];	T[2][2] = Y[2][2];

} 

void p_special_transform(TRUNC_STRUCT *tsr, double b_gse[3], double P_gse[3][3], double v_gse[3], double p[3][3])
{

int i,j;
double P_special[3][3], v_special[3], **p_trans;
double Nsc;

Nsc = tsr->Nsc;

for(i=0; i<3; i++)
   v_special[i] = v_gse[i];

for(i = 0; i<3; i++){ 
	for(j=0; j<3; j++){ P_special[i][j] = P_gse[i][j]; }} 
   
   
special_transform_vector(v_special, b_gse);
special_transform_tensor(v_special, b_gse, P_special);


// calculate p 
for(i = 0; i<3; i++){ 
	for(j=0; j<3; j++){p[i][j] = (P_special[i][j] - Nsc*M_ELECTRON*v_special[i]*v_special[j]);}}; 

}

double  LM1_LOWER_ENERGY_TABLE[4][93] = {{ 0.1 ,
 0.1 ,
 0.1 ,
 1.684 ,
 2.155 ,
 2.374 ,
 2.593 ,
 2.951 ,
 3.704 ,
 4.015 ,
 4.327 ,
 4.757 ,
 5.334 ,
 6.008 ,
 6.398 ,
 7.443 ,
 7.492 ,
 8.696 ,
 8.955 ,
 10.337 ,
 10.696 ,
 11.345 ,
 11.881 ,
 14.65 ,
 14.814 ,
 17.55 ,
 19.146 ,
 21.839 ,
 23.987 ,
 27.117 ,
 30.343 ,
 33.458 ,
 37.905 ,
 41.685 ,
 46.584 ,
 52.354 ,
 58.395 ,
 65.669 ,
 72.095 ,
 81.949 ,
 92.128 ,
 101.891 ,
 113.654 ,
 127.087 ,
 141.385 ,
 158.828 ,
 177.498 ,
 197.198 ,
 220.608 ,
 246.259 ,
 276.113 ,
 308.369 ,
 344.005 ,
 382.307 ,
 424.617 ,
 474.506 ,
 531.27 ,
 594.553 ,
 666.552 ,
 745.966 ,
 831.379 ,
 930.849 ,
 1042.238 ,
 1166.092 ,
 1300.825 ,
 1453.77 ,
 1627.083 ,
 1818.861 ,
 2031.517 ,
 2270.115 ,
 2537.567 ,
 2843.805 ,
 3178.92 ,
 3546.825 ,
 3968.744 ,
 4433.744 ,
 4932.634 ,
 5522.729 ,
 6168.654 ,
 6888.827 ,
 7699.794 ,
 8608.086 ,
 9616.917 ,
 10752.53 ,
 12008.52 ,
 13410.83 ,
 14997.91 ,
 16739.19 ,
 16739.19 ,
 16739.21 ,
 16739.22 ,
 16739.22 ,
 16739.24 },
{ 0.1 ,
 0.1 ,
 0.1 ,
 1.196 ,
 1.3 ,
 1.539 ,
 1.835 ,
 2.258 ,
 2.796 ,
 3.402 ,
 3.891 ,
 4.514 ,
 5.069 ,
 5.713 ,
 6.123 ,
 6.97 ,
 7.447 ,
 8.056 ,
 8.487 ,
 9.254 ,
 9.676 ,
 10.859 ,
 11.659 ,
 14.008 ,
 14.855 ,
 17.527 ,
 19.356 ,
 21.158 ,
 23.248 ,
 27.036 ,
 29.614 ,
 34.755 ,
 37.953 ,
 42.846 ,
 47.601 ,
 51.856 ,
 57.094 ,
 64.978 ,
 72.165 ,
 80.541 ,
 90.64 ,
 101.665 ,
 114.455 ,
 127.249 ,
 141.11 ,
 158.726 ,
 175.692 ,
 200.375 ,
 227.671 ,
 251.12 ,
 276.511 ,
 306.364 ,
 341.871 ,
 383.495 ,
 434.115 ,
 476.186 ,
 525.813 ,
 591.544 ,
 662.887 ,
 735.375 ,
 823.568 ,
 921.101 ,
 1028.67 ,
 1148.879 ,
 1284.465 ,
 1441.595 ,
 1621.006 ,
 1800.711 ,
 2007.957 ,
 2248.405 ,
 2518.75 ,
 2813.061 ,
 3139.887 ,
 3502.312 ,
 3908.861 ,
 4378.602 ,
 4888.947 ,
 5471.516 ,
 6101.403 ,
 6800.693 ,
 7584.717 ,
 8485.334 ,
 9474.633 ,
 10599.06 ,
 11824.42 ,
 13218 ,
 14774.61 ,
 16491.91 ,
 16491.92 ,
 16491.93 ,
 16491.94 ,
 16491.95 ,
 16491.96 },
{ 0.1 ,
 0.1 ,
 0.1 ,
 1.105 ,
 1.126 ,
 1.442 ,
 1.829 ,
 2.196 ,
 2.656 ,
 3.237 ,
 4.087 ,
 4.808 ,
 5.375 ,
 6.219 ,
 6.578 ,
 7.463 ,
 7.82 ,
 8.468 ,
 9.067 ,
 9.948 ,
 9.957 ,
 11.355 ,
 12.275 ,
 14.447 ,
 14.858 ,
 17.594 ,
 19.019 ,
 21.511 ,
 23.502 ,
 26.493 ,
 29.403 ,
 32.714 ,
 37.088 ,
 41.272 ,
 46.147 ,
 51.492 ,
 57.218 ,
 64.905 ,
 71.835 ,
 80.677 ,
 90.573 ,
 100.663 ,
 113.395 ,
 126.485 ,
 141.799 ,
 158.619 ,
 176.662 ,
 197.701 ,
 220.565 ,
 246.56 ,
 275.324 ,
 306.86 ,
 343.206 ,
 383.944 ,
 428.567 ,
 477.27 ,
 532.756 ,
 595.994 ,
 662.907 ,
 744.279 ,
 826.469 ,
 925.739 ,
 1037.976 ,
 1157.436 ,
 1286.109 ,
 1444.581 ,
 1599.024 ,
 1801.31 ,
 1991.348 ,
 2226.744 ,
 2493.686 ,
 2809.988 ,
 3118.842 ,
 3511.455 ,
 3904.726 ,
 4367.895 ,
 4857.024 ,
 5430.443 ,
 6046.971 ,
 6772.574 ,
 7557.233 ,
 8478.328 ,
 9479.216 ,
 10581.12 ,
 11788.31 ,
 13181.31 ,
 14694.19 ,
 16427.42 ,
 16427.43 ,
 16427.44 ,
 16427.45 ,
 16427.46 ,
 16427.47 },
{ 0.1 ,
 0.1 ,
 0.1 ,
 0.69 ,
 1.381 ,
 1.726 ,
 2.269 ,
 2.837 ,
 3.514 ,
 3.667 ,
 4.289 ,
 5.094 ,
 6.035 ,
 6.272 ,
 6.688 ,
 7.506 ,
 8.15 ,
 8.936 ,
 8.976 ,
 9.644 ,
 10.427 ,
 11.44 ,
 11.714 ,
 14.47 ,
 15.235 ,
 18.023 ,
 19.576 ,
 21.704 ,
 24.164 ,
 27.167 ,
 30.091 ,
 33.54 ,
 37.63 ,
 42.13 ,
 46.918 ,
 52.173 ,
 57.869 ,
 65.224 ,
 72.672 ,
 80.329 ,
 90.117 ,
 100.832 ,
 112.343 ,
 125.568 ,
 140.405 ,
 156.959 ,
 175.149 ,
 195.92 ,
 219.344 ,
 245.557 ,
 273.999 ,
 306.892 ,
 343.966 ,
 384.286 ,
 427.672 ,
 477.878 ,
 534.068 ,
 596.761 ,
 665.382 ,
 744.885 ,
 828.421 ,
 926.272 ,
 1034.637 ,
 1156.21 ,
 1294.149 ,
 1443.959 ,
 1614.25 ,
 1803.843 ,
 2008.837 ,
 2249.012 ,
 2515.748 ,
 2815.087 ,
 3134.684 ,
 3500.298 ,
 3916.025 ,
 4376.876 ,
 4880.32 ,
 5454.518 ,
 6102.091 ,
 6806.238 ,
 7630.422 ,
 8506.525 ,
 9490.093 ,
 10633.22 ,
 11860.05 ,
 13243.46 ,
 14795.1 ,
 16527.31 ,
 16527.32 ,
 16527.33 ,
 16527.34 ,
 16527.35 ,
 16527.36 }};

double  LM2_LOWER_ENERGY_TABLE[4][93] = {{ 0.1 ,
 1.369 ,
 1.935 ,
 1.961 ,
 2.021 ,
 2.581 ,
 2.583 ,
 3.276 ,
 3.37 ,
 4.095 ,
 4.353 ,
 5.148 ,
 5.761 ,
 6.703 ,
 7.397 ,
 8.321 ,
 8.591 ,
 8.772 ,
 10.157 ,
 11.033 ,
 12.319 ,
 14.452 ,
 15.543 ,
 17.23 ,
 19.127 ,
 21.602 ,
 24.462 ,
 26.422 ,
 29.684 ,
 32.736 ,
 37.863 ,
 41.22 ,
 46.636 ,
 51.814 ,
 57.469 ,
 65.023 ,
 72.567 ,
 81.684 ,
 91.657 ,
 102.177 ,
 113.386 ,
 126.461 ,
 141.257 ,
 158.369 ,
 177.518 ,
 197.635 ,
 221.086 ,
 248.217 ,
 276.334 ,
 308.384 ,
 344.41 ,
 386.368 ,
 430.741 ,
 479.847 ,
 537.43 ,
 601.266 ,
 673.551 ,
 754.268 ,
 840.468 ,
 940.296 ,
 1050.997 ,
 1169.39 ,
 1304.191 ,
 1463.504 ,
 1637.308 ,
 1825.38 ,
 2037.289 ,
 2283.398 ,
 2551.688 ,
 2852.128 ,
 3178.061 ,
 3542.79 ,
 3990.048 ,
 4439.191 ,
 4951.838 ,
 5538.84 ,
 6173.844 ,
 6873.49 ,
 7708.366 ,
 8616.582 ,
 9613.396 ,
 10739.55 ,
 12022.41 ,
 13379.21 ,
 15017.67 ,
 16774.28 ,
 18276.72 ,
 20779.99 ,
 20780 ,
 20780.01 ,
 20780.02 ,
 20780.03 ,
 20780.04 },
{ 0.1 ,
 1.627 ,
 1.943 ,
 2.157 ,
 2.246 ,
 2.655 ,
 2.809 ,
 3.386 ,
 3.448 ,
 4.298 ,
 4.512 ,
 5.336 ,
 5.755 ,
 6.624 ,
 6.861 ,
 7.901 ,
 8.008 ,
 8.453 ,
 9.356 ,
 10.599 ,
 11.653 ,
 13.416 ,
 14.98 ,
 16.791 ,
 18.519 ,
 20.945 ,
 23.28 ,
 25.829 ,
 28.84 ,
 32.364 ,
 36.61 ,
 39.065 ,
 44.001 ,
 49.655 ,
 56.348 ,
 63.926 ,
 70.124 ,
 77.805 ,
 87.026 ,
 97.472 ,
 109.464 ,
 123.331 ,
 139.41 ,
 153.63 ,
 172.161 ,
 192.596 ,
 217.149 ,
 242.503 ,
 272.108 ,
 305.552 ,
 342.051 ,
 381.501 ,
 426.647 ,
 476.22 ,
 532.845 ,
 595.768 ,
 666.754 ,
 747.229 ,
 827.857 ,
 930.153 ,
 1037.073 ,
 1156.801 ,
 1297.136 ,
 1442.921 ,
 1624.093 ,
 1809.938 ,
 2019.47 ,
 2269.644 ,
 2539.623 ,
 2831.75 ,
 3172.097 ,
 3523.901 ,
 3939.532 ,
 4396.614 ,
 4888.769 ,
 5479.709 ,
 6118.495 ,
 6812.605 ,
 7608.644 ,
 8566.398 ,
 9557.921 ,
 10651.17 ,
 11925.62 ,
 13286.67 ,
 14869.78 ,
 16610.07 ,
 18541.99 ,
 20730 ,
 20730.01 ,
 20730.02 ,
 20730.03 ,
 20730.04 ,
 20730.05 },
{ 0.1 ,
 2.605 ,
 3.068 ,
 3.15 ,
 3.44 ,
 3.464 ,
 3.915 ,
 3.948 ,
 4.477 ,
 4.529 ,
 5.248 ,
 5.523 ,
 6.151 ,
 6.511 ,
 7.301 ,
 7.973 ,
 9.314 ,
 9.966 ,
 11.36 ,
 12.017 ,
 13.543 ,
 14.692 ,
 16.486 ,
 17.864 ,
 19.919 ,
 21.779 ,
 24.459 ,
 26.693 ,
 29.902 ,
 33.245 ,
 37.001 ,
 40.848 ,
 45.723 ,
 51.305 ,
 57.847 ,
 64.631 ,
 72.223 ,
 80.892 ,
 90.94 ,
 101.619 ,
 113.742 ,
 126.493 ,
 142.296 ,
 158.786 ,
 177.82 ,
 198.694 ,
 222.566 ,
 248.046 ,
 277.771 ,
 310.024 ,
 346.974 ,
 388.332 ,
 433.2 ,
 481.661 ,
 538.802 ,
 601.85 ,
 673.824 ,
 750.249 ,
 837.484 ,
 934.01 ,
 1044.169 ,
 1166.317 ,
 1297.299 ,
 1450.058 ,
 1623.593 ,
 1797.342 ,
 1999.645 ,
 2239.353 ,
 2502.73 ,
 2792.863 ,
 3112.687 ,
 3473.137 ,
 3867.947 ,
 4281.959 ,
 4726.461 ,
 5280.814 ,
 5891.518 ,
 6490.217 ,
 7120.688 ,
 7997.798 ,
 8937.179 ,
 9992.163 ,
 11173.74 ,
 12473.07 ,
 13929.34 ,
 15553.26 ,
 17380.7 ,
 19440.55 ,
 19440.56 ,
 19440.57 ,
 19440.58 ,
 19440.59 ,
 19440.6 },
{ 0.1 ,
 2.544 ,
 2.852 ,
 3.257 ,
 3.588 ,
 3.666 ,
 4.145 ,
 4.533 ,
 4.964 ,
 5.679 ,
 6.041 ,
 6.645 ,
 7.211 ,
 7.756 ,
 8.336 ,
 8.965 ,
 9.398 ,
 9.784 ,
 10.671 ,
 11.76 ,
 13.1 ,
 14.656 ,
 16.384 ,
 18.135 ,
 20.104 ,
 22.305 ,
 24.719 ,
 27.269 ,
 30.54 ,
 34.343 ,
 38.673 ,
 42.057 ,
 47.953 ,
 53.556 ,
 60.49 ,
 67.594 ,
 75.349 ,
 84.03 ,
 93.862 ,
 104.438 ,
 116.972 ,
 129.878 ,
 145.152 ,
 161.825 ,
 180.953 ,
 201.159 ,
 225.672 ,
 251.553 ,
 281.813 ,
 314.061 ,
 350.282 ,
 391.143 ,
 437.271 ,
 486.978 ,
 542.869 ,
 605.969 ,
 678.277 ,
 757.43 ,
 845.581 ,
 945.251 ,
 1057.727 ,
 1183.898 ,
 1318.135 ,
 1477.292 ,
 1653.151 ,
 1844.262 ,
 2063.598 ,
 2304.574 ,
 2575.172 ,
 2879.57 ,
 3211.281 ,
 3583.296 ,
 4012.531 ,
 4486.886 ,
 5003.237 ,
 5602.161 ,
 6252.175 ,
 6977.28 ,
 7796.741 ,
 8712.124 ,
 9738.852 ,
 10879.01 ,
 12132.39 ,
 13542.93 ,
 15148.25 ,
 16914.5 ,
 18896.68 ,
 21042.45 ,
 21042.46 ,
 21042.47 ,
 21042.48 ,
 21042.49 ,
 21042.5 }};
 
double  LM3_LOWER_ENERGY_TABLE[4][93] = {{ 0.1 ,
 0.1 ,
 0.1 ,
 2.163 ,
 2.304 ,
 2.581 ,
 2.9 ,
 3.468 ,
 3.543 ,
 3.62 ,
 4.014 ,
 4.975 ,
 5.046 ,
 5.858 ,
 6.093 ,
 7.331 ,
 7.412 ,
 8.022 ,
 8.663 ,
 8.716 ,
 10.084 ,
 10.303 ,
 12.712 ,
 14.704 ,
 16.169 ,
 17.874 ,
 19.785 ,
 21.65 ,
 24.124 ,
 26.593 ,
 29.801 ,
 33.99 ,
 38.44 ,
 42.212 ,
 46.636 ,
 51.814 ,
 57.469 ,
 65.023 ,
 72.567 ,
 81.684 ,
 91.657 ,
 102.177 ,
 113.386 ,
 126.461 ,
 141.257 ,
 158.369 ,
 177.518 ,
 197.635 ,
 221.086 ,
 248.217 ,
 276.334 ,
 308.384 ,
 344.41 ,
 386.368 ,
 430.741 ,
 479.847 ,
 537.43 ,
 601.266 ,
 673.551 ,
 754.268 ,
 840.468 ,
 940.296 ,
 1050.997 ,
 1169.39 ,
 1304.191 ,
 1463.504 ,
 1637.308 ,
 1825.38 ,
 2037.289 ,
 2283.398 ,
 2551.688 ,
 2852.128 ,
 3178.061 ,
 3542.79 ,
 3990.048 ,
 4439.191 ,
 4951.838 ,
 5538.84 ,
 6173.844 ,
 6873.49 ,
 7708.366 ,
 8616.582 ,
 9613.396 ,
 10739.55 ,
 12022.41 ,
 13379.21 ,
 15017.67 ,
 16774.28 ,
 16774.29 ,
 16774.3 ,
 16774.31 ,
 16774.32 ,
 16774.33 },
{ 0.1 ,
 0.1 ,
 0.1 ,
 2.402 ,
 2.916 ,
 2.986 ,
 3.318 ,
 3.503 ,
 3.849 ,
 3.893 ,
 4.395 ,
 4.839 ,
 5.3 ,
 5.645 ,
 6.217 ,
 6.758 ,
 7.319 ,
 7.736 ,
 8.531 ,
 9.007 ,
 10.145 ,
 10.879 ,
 12.576 ,
 14.043 ,
 15.772 ,
 17.373 ,
 19.445 ,
 21.853 ,
 24.282 ,
 26.704 ,
 29.986 ,
 33.464 ,
 37.742 ,
 42.3 ,
 47.27 ,
 51.313 ,
 58.056 ,
 63.926 ,
 70.124 ,
 77.805 ,
 87.026 ,
 97.472 ,
 109.464 ,
 123.331 ,
 139.41 ,
 153.63 ,
 172.161 ,
 192.596 ,
 217.149 ,
 242.503 ,
 272.108 ,
 305.552 ,
 342.051 ,
 381.501 ,
 426.647 ,
 476.22 ,
 532.845 ,
 595.768 ,
 666.754 ,
 747.229 ,
 827.857 ,
 930.153 ,
 1037.073 ,
 1156.801 ,
 1297.136 ,
 1442.921 ,
 1624.093 ,
 1809.938 ,
 2019.47 ,
 2269.644 ,
 2539.623 ,
 2831.75 ,
 3172.097 ,
 3523.901 ,
 3939.532 ,
 4396.614 ,
 4888.769 ,
 5479.709 ,
 6118.495 ,
 6812.605 ,
 7608.644 ,
 8566.398 ,
 9557.921 ,
 10651.17 ,
 11925.62 ,
 13286.67 ,
 14869.78 ,
 16610.07 ,
 16610.08 ,
 16610.09 ,
 16610.1 ,
 16610.11 ,
 16610.12 },
{ 0.1 ,
 0.1 ,
 0.1 ,
 2.605 ,
 3.068 ,
 3.15 ,
 3.44 ,
 3.464 ,
 3.915 ,
 3.948 ,
 4.477 ,
 4.529 ,
 5.248 ,
 5.523 ,
 6.151 ,
 6.511 ,
 7.301 ,
 7.973 ,
 9.314 ,
 9.966 ,
 11.36 ,
 12.017 ,
 13.543 ,
 14.692 ,
 16.486 ,
 17.864 ,
 19.919 ,
 21.779 ,
 24.459 ,
 26.693 ,
 29.902 ,
 33.245 ,
 37.001 ,
 40.848 ,
 45.723 ,
 51.305 ,
 57.847 ,
 64.631 ,
 72.223 ,
 80.892 ,
 90.94 ,
 101.619 ,
 113.742 ,
 126.493 ,
 142.296 ,
 158.786 ,
 177.82 ,
 198.694 ,
 222.566 ,
 248.046 ,
 277.771 ,
 310.024 ,
 346.974 ,
 388.332 ,
 433.2 ,
 481.661 ,
 538.802 ,
 601.85 ,
 673.824 ,
 750.249 ,
 837.484 ,
 934.01 ,
 1044.169 ,
 1166.317 ,
 1297.299 ,
 1450.058 ,
 1623.593 ,
 1797.342 ,
 1999.645 ,
 2239.353 ,
 2502.73 ,
 2792.863 ,
 3112.687 ,
 3473.137 ,
 3867.947 ,
 4281.959 ,
 4726.461 ,
 5280.814 ,
 5891.518 ,
 6490.217 ,
 7120.688 ,
 7997.798 ,
 8937.179 ,
 9992.163 ,
 11173.74 ,
 12473.07 ,
 13929.34 ,
 15553.26 ,
 15553.27 ,
 15553.28 ,
 15553.29 ,
 15553.3 ,
 15553.31 },
{ 0.1 ,
 0.1 ,
 0.1 ,
 2.927 ,
 3.415 ,
 3.468 ,
 3.545 ,
 3.778 ,
 4.11 ,
 4.12 ,
 4.877 ,
 5.202 ,
 6.003 ,
 6.448 ,
 7.285 ,
 7.533 ,
 8.359 ,
 8.614 ,
 9.318 ,
 9.693 ,
 10.698 ,
 11.456 ,
 13.228 ,
 14.6 ,
 16.602 ,
 18.103 ,
 20.49 ,
 23.308 ,
 25.337 ,
 27.521 ,
 31.031 ,
 34.759 ,
 39.013 ,
 42.811 ,
 47.953 ,
 53.125 ,
 60.139 ,
 66.792 ,
 74.197 ,
 83.163 ,
 93.217 ,
 103.605 ,
 116.102 ,
 129.078 ,
 144.483 ,
 161.004 ,
 180.014 ,
 200.603 ,
 225.778 ,
 251.645 ,
 282.307 ,
 314.698 ,
 351.859 ,
 393.268 ,
 438.921 ,
 488.877 ,
 546.663 ,
 610.554 ,
 683.512 ,
 760.478 ,
 847.538 ,
 948.915 ,
 1060.566 ,
 1183.961 ,
 1320.804 ,
 1477.731 ,
 1653.196 ,
 1844.209 ,
 2064.883 ,
 2304.668 ,
 2580.801 ,
 2881.79 ,
 3213.828 ,
 3588.598 ,
 4022.145 ,
 4489.837 ,
 4997.905 ,
 5611.482 ,
 6254.412 ,
 6986.804 ,
 7781.023 ,
 8696.277 ,
 9750.665 ,
 10887.16 ,
 12153.86 ,
 13524.58 ,
 15162.61 ,
 16943.27 ,
 16943.28 ,
 16943.29 ,
 16943.3 ,
 16943.31 ,
 16943.32 }};

double  HM1_LOWER_ENERGY_TABLE[4][93] ={{ 0.1 ,
 0.1 ,
 0.1 ,
 0.406 ,
 0.452 ,
 0.498 ,
 0.805 ,
 1.137 ,
 1.551 ,
 2.434 ,
 3.059 ,
 3.233 ,
 3.961 ,
 4.448 ,
 4.819 ,
 5.586 ,
 6.418 ,
 6.571 ,
 7.23 ,
 7.871 ,
 8.243 ,
 9.714 ,
 10.305 ,
 12.53 ,
 13.426 ,
 15.794 ,
 17.384 ,
 19.984 ,
 22.309 ,
 25.282 ,
 27.785 ,
 31.492 ,
 35.727 ,
 39.599 ,
 44.272 ,
 49.582 ,
 55.357 ,
 62.523 ,
 69.704 ,
 78.484 ,
 88.627 ,
 98.761 ,
 110.239 ,
 123.265 ,
 137.95 ,
 154.438 ,
 172.611 ,
 193.619 ,
 216.14 ,
 241.74 ,
 270.475 ,
 303.652 ,
 338.384 ,
 377.699 ,
 423.18 ,
 472.753 ,
 527.002 ,
 588.496 ,
 659.981 ,
 737.175 ,
 822.461 ,
 916.478 ,
 1027.459 ,
 1146.699 ,
 1280.042 ,
 1429.855 ,
 1599.688 ,
 1790.054 ,
 1997.056 ,
 2235.024 ,
 2496.854 ,
 2789.58 ,
 3115.311 ,
 3473.251 ,
 3889.14 ,
 4348.213 ,
 4846.94 ,
 5420.532 ,
 6046.565 ,
 6770.292 ,
 7554.947 ,
 8463.781 ,
 9426.256 ,
 10529.85 ,
 11798.24 ,
 13141.84 ,
 14716.69 ,
 16395.41 ,
 16395.42 ,
 16395.43 ,
 16395.44 ,
 16395.45 ,
 16395.46 },
{ 0.1 ,
 0.1 ,
 0.1 ,
 1.177 ,
 1.28 ,
 1.515 ,
 1.806 ,
 2.223 ,
 2.752 ,
 3.349 ,
 3.83 ,
 4.443 ,
 4.989 ,
 5.624 ,
 6.027 ,
 6.861 ,
 7.331 ,
 7.93 ,
 8.355 ,
 9.109 ,
 9.525 ,
 10.689 ,
 11.477 ,
 13.79 ,
 14.398 ,
 16.978 ,
 18.022 ,
 20.327 ,
 21.917 ,
 24.459 ,
 28.533 ,
 31.185 ,
 35.768 ,
 40.072 ,
 45.489 ,
 49.689 ,
 54.83 ,
 61.522 ,
 69.194 ,
 77.343 ,
 86.128 ,
 96.96 ,
 107.619 ,
 121.906 ,
 134.964 ,
 155.293 ,
 174.521 ,
 193.15 ,
 215.649 ,
 241.922 ,
 270.901 ,
 301.131 ,
 336.649 ,
 378.112 ,
 421.124 ,
 467.542 ,
 521.121 ,
 581.572 ,
 651.571 ,
 726.875 ,
 811.729 ,
 906.563 ,
 1016.994 ,
 1131.734 ,
 1266.954 ,
 1416.499 ,
 1586.404 ,
 1769.62 ,
 1974.116 ,
 2205.149 ,
 2466.923 ,
 2761.835 ,
 3080.704 ,
 3436.451 ,
 3840.028 ,
 4302.321 ,
 4804.416 ,
 5374.37 ,
 5996.503 ,
 6683.876 ,
 7458.308 ,
 8350.431 ,
 9318.166 ,
 10429.5 ,
 11648.83 ,
 13017.33 ,
 14570.49 ,
 16265.16 ,
 16265.17 ,
 16265.18 ,
 16265.19 ,
 16265.2 ,
 16265.21 },
{ 0.1 ,
 0.1 ,
 0.1 ,
 1.187 ,
 1.253 ,
 1.509 ,
 1.723 ,
 2.247 ,
 2.553 ,
 3.176 ,
 3.71 ,
 4.25 ,
 4.78 ,
 5.479 ,
 6.079 ,
 6.62 ,
 7.121 ,
 7.728 ,
 8.211 ,
 8.757 ,
 9.202 ,
 10.483 ,
 11.587 ,
 13.648 ,
 14.303 ,
 17.099 ,
 18.714 ,
 21.153 ,
 23.505 ,
 26.379 ,
 29.26 ,
 32.778 ,
 36.809 ,
 40.666 ,
 45.713 ,
 50.864 ,
 56.545 ,
 63.797 ,
 71.095 ,
 79.057 ,
 89.606 ,
 100.054 ,
 112.155 ,
 125.272 ,
 140.014 ,
 156.536 ,
 174.827 ,
 194.711 ,
 217.167 ,
 242.246 ,
 271.504 ,
 303.289 ,
 339.112 ,
 378.027 ,
 423.074 ,
 472.068 ,
 527.849 ,
 590.428 ,
 657.27 ,
 737.154 ,
 821.18 ,
 920.72 ,
 1024.847 ,
 1142.369 ,
 1287.135 ,
 1424.504 ,
 1594.952 ,
 1781.161 ,
 1989.938 ,
 2227.723 ,
 2492.255 ,
 2779.266 ,
 3096.318 ,
 3457.207 ,
 3842.206 ,
 4311.168 ,
 4830.087 ,
 5410.813 ,
 6050.347 ,
 6744.177 ,
 7550.6 ,
 8486.155 ,
 9429.671 ,
 10599.08 ,
 11831.18 ,
 13212.26 ,
 14771.63 ,
 16623.55 ,
 16623.56 ,
 16623.57 ,
 16623.58 ,
 16623.59 ,
 16623.6 },
{ 0.1 ,
 0.1 ,
 0.1 ,
 0.573 ,
 1.147 ,
 1.368 ,
 1.795 ,
 2.107 ,
 2.819 ,
 3.254 ,
 3.765 ,
 4.515 ,
 4.959 ,
 5.412 ,
 5.969 ,
 6.507 ,
 7.252 ,
 7.735 ,
 8.219 ,
 8.666 ,
 9.26 ,
 10.441 ,
 11.787 ,
 13.769 ,
 14.642 ,
 17.017 ,
 18.513 ,
 20.942 ,
 23.169 ,
 26.113 ,
 28.742 ,
 32.153 ,
 36.094 ,
 40.018 ,
 44.962 ,
 50.856 ,
 56.233 ,
 63.659 ,
 70.165 ,
 78.214 ,
 88.581 ,
 98.318 ,
 109.328 ,
 123.892 ,
 139.13 ,
 155.655 ,
 173.546 ,
 193.635 ,
 215.997 ,
 241.988 ,
 270.502 ,
 302.776 ,
 337.51 ,
 377.49 ,
 421.895 ,
 471.769 ,
 524.325 ,
 584.922 ,
 654.715 ,
 730.286 ,
 814.744 ,
 910.256 ,
 1015.437 ,
 1135.157 ,
 1268.781 ,
 1412.154 ,
 1577.305 ,
 1764.062 ,
 1969.218 ,
 2188.332 ,
 2445.167 ,
 2733.953 ,
 3059.981 ,
 3415.027 ,
 3819.815 ,
 4272.717 ,
 4784.406 ,
 5359.418 ,
 5983.38 ,
 6676.256 ,
 7483.987 ,
 8356.659 ,
 9326.602 ,
 10443.36 ,
 11712.09 ,
 13025.54 ,
 14597.96 ,
 16348.1 ,
 16348.11 ,
 16348.12 ,
 16348.13 ,
 16348.14 ,
 16348.15 }};
 
double  HM2_LOWER_ENERGY_TABLE[4][93] = {{ 0.1 ,
 0.883 ,
 0.991 ,
 1.186 ,
 1.239 ,
 1.701 ,
 1.949 ,
 2.196 ,
 2.417 ,
 3.045 ,
 3.529 ,
 4.089 ,
 4.38 ,
 5.084 ,
 5.632 ,
 6.365 ,
 6.427 ,
 6.823 ,
 7.72 ,
 8.845 ,
 9.829 ,
 11.452 ,
 13.187 ,
 15.163 ,
 16.719 ,
 18.972 ,
 21.432 ,
 24.245 ,
 26.916 ,
 30.429 ,
 35.329 ,
 38.81 ,
 45.089 ,
 49.953 ,
 56.821 ,
 63.073 ,
 70.767 ,
 79.799 ,
 89.832 ,
 100.248 ,
 112.491 ,
 125.218 ,
 140.447 ,
 156.469 ,
 175.186 ,
 195.015 ,
 218.592 ,
 244.008 ,
 273.746 ,
 305.629 ,
 341.495 ,
 381.648 ,
 426.831 ,
 476.403 ,
 531.976 ,
 594.483 ,
 665.344 ,
 744.857 ,
 830.303 ,
 927.918 ,
 1035.821 ,
 1164.028 ,
 1295.773 ,
 1453.101 ,
 1624.604 ,
 1813.512 ,
 2024.206 ,
 2260.872 ,
 2529.131 ,
 2829.581 ,
 3159.374 ,
 3527.15 ,
 3946.052 ,
 4424.411 ,
 4932.217 ,
 5504.494 ,
 6146.428 ,
 6880.849 ,
 7705.032 ,
 8616.858 ,
 9613.37 ,
 10737.91 ,
 11997.06 ,
 13392 ,
 14971.23 ,
 16732.6 ,
 18571.45 ,
 20691.73 ,
 20691.74 ,
 20691.75 ,
 20691.76 ,
 20691.77 ,
 20691.78 },
{ 0.1 ,
 1.602 ,
 1.912 ,
 2.124 ,
 2.211 ,
 2.614 ,
 2.765 ,
 3.333 ,
 3.394 ,
 4.231 ,
 4.442 ,
 5.253 ,
 5.665 ,
 6.521 ,
 6.753 ,
 7.778 ,
 7.883 ,
 8.321 ,
 9.21 ,
 10.434 ,
 11.471 ,
 13.206 ,
 14.746 ,
 16.529 ,
 18.23 ,
 20.618 ,
 22.917 ,
 25.425 ,
 28.39 ,
 31.858 ,
 36.655 ,
 40.092 ,
 44.885 ,
 50.862 ,
 57.506 ,
 63.959 ,
 71.256 ,
 79.842 ,
 89.379 ,
 99.969 ,
 110.463 ,
 122.504 ,
 137.383 ,
 153.839 ,
 170.94 ,
 191.378 ,
 214.261 ,
 238.388 ,
 268.507 ,
 299.23 ,
 335.128 ,
 375.702 ,
 417.966 ,
 465.661 ,
 519.609 ,
 580.466 ,
 652.427 ,
 732.085 ,
 812.593 ,
 906.212 ,
 1019.44 ,
 1131.549 ,
 1271.327 ,
 1415.858 ,
 1595.356 ,
 1783.138 ,
 1984.155 ,
 2230.016 ,
 2489.921 ,
 2781.133 ,
 3114.377 ,
 3470.778 ,
 3881.435 ,
 4316.557 ,
 4834.884 ,
 5413.005 ,
 6017.351 ,
 6687.664 ,
 7495.449 ,
 8376.337 ,
 9405.591 ,
 10498.25 ,
 11748.96 ,
 13106.93 ,
 14677.41 ,
 16395.28 ,
 18319.16 ,
 20506.23 ,
 20506.24 ,
 20506.25 ,
 20506.26 ,
 20506.27 ,
 20506.28 },
{ 0.1 ,
 2.652 ,
 3.157 ,
 3.266 ,
 3.868 ,
 3.869 ,
 4.318 ,
 4.406 ,
 5.006 ,
 5.265 ,
 5.948 ,
 6.086 ,
 6.826 ,
 6.868 ,
 7.503 ,
 7.604 ,
 8.361 ,
 8.745 ,
 10.182 ,
 11.024 ,
 12.518 ,
 13.858 ,
 15.925 ,
 17.265 ,
 19.774 ,
 22.492 ,
 25.284 ,
 27.757 ,
 31.24 ,
 33.617 ,
 37.82 ,
 41.836 ,
 47.122 ,
 52.01 ,
 58.777 ,
 65.264 ,
 74.175 ,
 82.429 ,
 93.167 ,
 103.779 ,
 116.061 ,
 128.722 ,
 144.261 ,
 160.64 ,
 179.322 ,
 199.673 ,
 223.165 ,
 248.79 ,
 278.779 ,
 310.498 ,
 347.104 ,
 388.289 ,
 433.311 ,
 481.601 ,
 537.851 ,
 600.86 ,
 671.562 ,
 748.344 ,
 833.831 ,
 929.124 ,
 1040.29 ,
 1161.399 ,
 1295.252 ,
 1446.611 ,
 1613.969 ,
 1789.894 ,
 1996.648 ,
 2229.691 ,
 2491.068 ,
 2778.369 ,
 3090.201 ,
 3442.921 ,
 3859.307 ,
 4252.832 ,
 4675.555 ,
 5225.86 ,
 5847.025 ,
 6506.126 ,
 7211.853 ,
 8108.927 ,
 9056.352 ,
 10115.98 ,
 11305.42 ,
 12620.57 ,
 14101.9 ,
 15737.5 ,
 17563.65 ,
 19689.92 ,
 19689.93 ,
 19689.94 ,
 19689.95 ,
 19689.96 ,
 19689.97 },
{ 0.1 ,
 2.623 ,
 3.123 ,
 3.231 ,
 3.826 ,
 3.827 ,
 4.271 ,
 4.359 ,
 4.952 ,
 5.208 ,
 5.883 ,
 6.02 ,
 6.752 ,
 6.794 ,
 7.422 ,
 7.522 ,
 8.271 ,
 8.65 ,
 10.072 ,
 10.905 ,
 12.383 ,
 13.708 ,
 15.753 ,
 17.078 ,
 19.561 ,
 22.249 ,
 25.011 ,
 27.458 ,
 30.903 ,
 33.254 ,
 37.412 ,
 41.384 ,
 46.614 ,
 51.449 ,
 58.143 ,
 64.56 ,
 73.375 ,
 81.54 ,
 92.162 ,
 102.66 ,
 114.809 ,
 127.334 ,
 142.705 ,
 158.907 ,
 177.387 ,
 197.519 ,
 220.758 ,
 246.107 ,
 275.771 ,
 307.148 ,
 343.36 ,
 384.101 ,
 428.636 ,
 476.405 ,
 532.049 ,
 594.378 ,
 664.317 ,
 740.271 ,
 824.836 ,
 919.101 ,
 1029.068 ,
 1148.87 ,
 1281.28 ,
 1431.006 ,
 1596.558 ,
 1770.586 ,
 1975.109 ,
 2205.638 ,
 2464.196 ,
 2748.397 ,
 3056.865 ,
 3405.781 ,
 3817.675 ,
 4206.955 ,
 4625.117 ,
 5169.486 ,
 5783.95 ,
 6435.941 ,
 7134.055 ,
 8021.452 ,
 8958.656 ,
 10006.85 ,
 11183.46 ,
 12484.43 ,
 13949.78 ,
 15567.73 ,
 17374.18 ,
 19477.51 ,
 19477.53 ,
 19477.53 ,
 19477.54 ,
 19477.56 ,
 19477.56 }};
 
double  HM3_LOWER_ENERGY_TABLE[4][93] = {{ 0.1 ,
 0.1 ,
 0.1 ,
 2.185 ,
 2.563 ,
 2.606 ,
 2.993 ,
 3.168 ,
 3.361 ,
 3.381 ,
 3.913 ,
 4.193 ,
 4.363 ,
 4.552 ,
 5.162 ,
 5.709 ,
 6.321 ,
 6.516 ,
 7.315 ,
 7.997 ,
 9.172 ,
 9.799 ,
 11.24 ,
 12.687 ,
 14.405 ,
 15.774 ,
 18.092 ,
 20.188 ,
 22.858 ,
 25.108 ,
 28.728 ,
 32.139 ,
 36.25 ,
 39.977 ,
 45.089 ,
 50.144 ,
 56.969 ,
 63.352 ,
 70.926 ,
 79.568 ,
 89.513 ,
 99.82 ,
 111.838 ,
 124.719 ,
 140.21 ,
 156.253 ,
 174.909 ,
 194.922 ,
 218.913 ,
 244.213 ,
 273.996 ,
 306.001 ,
 341.938 ,
 381.695 ,
 427.608 ,
 476.602 ,
 532.58 ,
 594.672 ,
 667.852 ,
 746.472 ,
 835.684 ,
 935.035 ,
 1046.043 ,
 1167.494 ,
 1301.159 ,
 1456.914 ,
 1627.651 ,
 1821.631 ,
 2034.525 ,
 2275.262 ,
 2542.799 ,
 2841.988 ,
 3168.831 ,
 3540.141 ,
 3958.888 ,
 4428.466 ,
 4944.116 ,
 5520.153 ,
 6171.95 ,
 6888.565 ,
 7688.366 ,
 8596.479 ,
 9598.986 ,
 10727.66 ,
 11985.25 ,
 13396.4 ,
 14985.92 ,
 16760.76 ,
 16760.77 ,
 16760.78 ,
 16760.79 ,
 16760.8 ,
 16760.81 },
{ 0.1 ,
 0.1 ,
 0.1 ,
 2.364 ,
 2.871 ,
 2.94 ,
 3.266 ,
 3.448 ,
 3.789 ,
 3.833 ,
 4.326 ,
 4.764 ,
 5.217 ,
 5.557 ,
 6.12 ,
 6.652 ,
 7.204 ,
 7.615 ,
 8.398 ,
 8.867 ,
 9.987 ,
 10.71 ,
 12.38 ,
 13.824 ,
 15.526 ,
 17.599 ,
 19.352 ,
 22.242 ,
 24.905 ,
 27.484 ,
 29.607 ,
 33.546 ,
 37.182 ,
 40.868 ,
 44.885 ,
 51.173 ,
 58.089 ,
 63.001 ,
 70.216 ,
 80.121 ,
 89.953 ,
 98.702 ,
 109.238 ,
 122.263 ,
 136.694 ,
 153.558 ,
 171.338 ,
 194.169 ,
 215.101 ,
 239.779 ,
 268.518 ,
 299.507 ,
 335.475 ,
 374.95 ,
 419.015 ,
 469.832 ,
 524.955 ,
 586.991 ,
 656.241 ,
 733.225 ,
 817.53 ,
 914.271 ,
 1023.828 ,
 1145.929 ,
 1280.916 ,
 1430.318 ,
 1595.356 ,
 1783.138 ,
 1984.155 ,
 2230.016 ,
 2489.921 ,
 2781.133 ,
 3114.377 ,
 3470.778 ,
 3881.435 ,
 4316.557 ,
 4834.884 ,
 5413.005 ,
 6017.351 ,
 6687.664 ,
 7495.449 ,
 8376.337 ,
 9405.591 ,
 10498.25 ,
 11748.96 ,
 13106.93 ,
 14677.41 ,
 16395.28 ,
 16395.29 ,
 16395.3 ,
 16395.31 ,
 16395.32 ,
 16395.33 },
{ 0.1 ,
 0.1 ,
 0.1 ,
 2.652 ,
 3.157 ,
 3.266 ,
 3.868 ,
 3.869 ,
 4.318 ,
 4.406 ,
 5.006 ,
 5.265 ,
 5.948 ,
 6.086 ,
 6.826 ,
 6.868 ,
 7.503 ,
 7.604 ,
 8.361 ,
 8.745 ,
 10.182 ,
 11.024 ,
 12.518 ,
 13.858 ,
 15.925 ,
 17.265 ,
 19.774 ,
 22.492 ,
 25.284 ,
 27.757 ,
 31.24 ,
 33.617 ,
 37.82 ,
 41.836 ,
 47.122 ,
 52.01 ,
 58.777 ,
 65.264 ,
 74.175 ,
 82.429 ,
 93.167 ,
 103.779 ,
 116.061 ,
 128.722 ,
 144.261 ,
 160.64 ,
 179.322 ,
 199.673 ,
 223.165 ,
 248.79 ,
 278.779 ,
 310.498 ,
 347.104 ,
 388.289 ,
 433.311 ,
 481.601 ,
 537.851 ,
 600.86 ,
 671.562 ,
 748.344 ,
 833.831 ,
 929.124 ,
 1040.29 ,
 1161.399 ,
 1295.252 ,
 1446.611 ,
 1613.969 ,
 1789.894 ,
 1996.648 ,
 2229.691 ,
 2491.068 ,
 2778.369 ,
 3090.201 ,
 3442.921 ,
 3859.307 ,
 4252.832 ,
 4675.555 ,
 5225.86 ,
 5847.025 ,
 6506.126 ,
 7211.853 ,
 8108.927 ,
 9056.352 ,
 10115.98 ,
 11305.42 ,
 12620.57 ,
 14101.9 ,
 15737.5 ,
 15737.5 ,
 15737.51 ,
 15737.52 ,
 15737.53 ,
 15737.55 },
{ 0.1 ,
 0.1 ,
 0.1 ,
 2.623 ,
 3.123 ,
 3.231 ,
 3.826 ,
 3.827 ,
 4.271 ,
 4.359 ,
 4.952 ,
 5.208 ,
 5.883 ,
 6.02 ,
 6.752 ,
 6.794 ,
 7.422 ,
 7.522 ,
 8.271 ,
 8.65 ,
 10.072 ,
 10.905 ,
 12.383 ,
 13.708 ,
 15.753 ,
 17.078 ,
 19.561 ,
 22.249 ,
 25.011 ,
 27.458 ,
 30.903 ,
 33.254 ,
 37.412 ,
 41.384 ,
 46.614 ,
 51.449 ,
 58.143 ,
 64.56 ,
 73.375 ,
 81.54 ,
 92.162 ,
 102.66 ,
 114.809 ,
 127.334 ,
 142.705 ,
 158.907 ,
 177.387 ,
 197.519 ,
 220.758 ,
 246.107 ,
 275.771 ,
 307.148 ,
 343.36 ,
 384.101 ,
 428.636 ,
 476.405 ,
 532.049 ,
 594.378 ,
 664.317 ,
 740.271 ,
 824.836 ,
 919.101 ,
 1029.068 ,
 1148.87 ,
 1281.28 ,
 1431.006 ,
 1596.558 ,
 1770.586 ,
 1975.109 ,
 2205.638 ,
 2464.196 ,
 2748.397 ,
 3056.865 ,
 3405.781 ,
 3817.675 ,
 4206.955 ,
 4625.117 ,
 5169.486 ,
 5783.95 ,
 6435.941 ,
 7134.055 ,
 8021.452 ,
 8958.656 ,
 10006.85 ,
 11183.46 ,
 12484.43 ,
 13949.78 ,
 15567.73 ,
 15567.74 ,
 15567.75 ,
 15567.76 ,
 15567.77 ,
 15567.78 }};


double  LM1_UPPER_ENERGY_TABLE[4][93] = {{ 0.1 ,
 0.1 ,
 0.101 ,
 2.951 ,
 3.704 ,
 4.015 ,
 4.327 ,
 4.757 ,
 5.334 ,
 6.008 ,
 6.398 ,
 7.443 ,
 7.492 ,
 8.696 ,
 8.955 ,
 10.337 ,
 10.696 ,
 11.345 ,
 11.881 ,
 14.65 ,
 14.814 ,
 17.55 ,
 19.146 ,
 21.839 ,
 23.987 ,
 27.117 ,
 30.343 ,
 33.458 ,
 37.905 ,
 41.685 ,
 46.584 ,
 52.354 ,
 58.395 ,
 65.669 ,
 72.095 ,
 81.949 ,
 92.128 ,
 101.891 ,
 113.654 ,
 127.087 ,
 141.385 ,
 158.828 ,
 177.498 ,
 197.198 ,
 220.608 ,
 246.259 ,
 276.113 ,
 308.369 ,
 344.005 ,
 382.307 ,
 424.617 ,
 474.506 ,
 531.27 ,
 594.553 ,
 666.552 ,
 745.966 ,
 831.379 ,
 930.849 ,
 1042.238 ,
 1166.092 ,
 1300.825 ,
 1453.77 ,
 1627.083 ,
 1818.861 ,
 2031.517 ,
 2270.115 ,
 2537.567 ,
 2843.805 ,
 3178.92 ,
 3546.825 ,
 3968.744 ,
 4433.744 ,
 4932.634 ,
 5522.729 ,
 6168.654 ,
 6888.827 ,
 7699.794 ,
 8608.086 ,
 9616.917 ,
 10752.53 ,
 12008.52 ,
 13410.83 ,
 14997.91 ,
 16739.19 ,
 18690.83 ,
 20931.96 ,
 23389.51 ,
 26020.5 ,
 26020.51 ,
 26020.52 ,
 26020.53 ,
 26020.54 ,
 26020.55 },
{ 0.1 ,
 0.1 ,
 0.101 ,
 2.258 ,
 2.796 ,
 3.402 ,
 3.891 ,
 4.514 ,
 5.069 ,
 5.713 ,
 6.123 ,
 6.97 ,
 7.447 ,
 8.056 ,
 8.487 ,
 9.254 ,
 9.676 ,
 10.859 ,
 11.659 ,
 14.008 ,
 14.855 ,
 17.527 ,
 19.356 ,
 21.158 ,
 23.248 ,
 27.036 ,
 29.614 ,
 34.755 ,
 37.953 ,
 42.846 ,
 47.601 ,
 51.856 ,
 57.094 ,
 64.978 ,
 72.165 ,
 80.541 ,
 90.64 ,
 101.665 ,
 114.455 ,
 127.249 ,
 141.11 ,
 158.726 ,
 175.692 ,
 200.375 ,
 227.671 ,
 251.12 ,
 276.511 ,
 306.364 ,
 341.871 ,
 383.495 ,
 434.115 ,
 476.186 ,
 525.813 ,
 591.544 ,
 662.887 ,
 735.375 ,
 823.568 ,
 921.101 ,
 1028.67 ,
 1148.879 ,
 1284.465 ,
 1441.595 ,
 1621.006 ,
 1800.711 ,
 2007.957 ,
 2248.405 ,
 2518.75 ,
 2813.061 ,
 3139.887 ,
 3502.312 ,
 3908.861 ,
 4378.602 ,
 4888.947 ,
 5471.516 ,
 6101.403 ,
 6800.693 ,
 7584.717 ,
 8485.334 ,
 9474.633 ,
 10599.06 ,
 11824.42 ,
 13218 ,
 14774.61 ,
 16491.91 ,
 18422.62 ,
 20592.35 ,
 23001.91 ,
 25657.97 ,
 25657.98 ,
 25657.99 ,
 25658 ,
 25658.01 ,
 25658.02 },
{ 0.1 ,
 0.1 ,
 0.101 ,
 2.196 ,
 2.656 ,
 3.237 ,
 4.087 ,
 4.808 ,
 5.375 ,
 6.219 ,
 6.578 ,
 7.463 ,
 7.82 ,
 8.468 ,
 9.067 ,
 9.948 ,
 9.957 ,
 11.355 ,
 12.275 ,
 14.447 ,
 14.858 ,
 17.594 ,
 19.019 ,
 21.511 ,
 23.502 ,
 26.493 ,
 29.403 ,
 32.714 ,
 37.088 ,
 41.272 ,
 46.147 ,
 51.492 ,
 57.218 ,
 64.905 ,
 71.835 ,
 80.677 ,
 90.573 ,
 100.663 ,
 113.395 ,
 126.485 ,
 141.799 ,
 158.619 ,
 176.662 ,
 197.701 ,
 220.565 ,
 246.56 ,
 275.324 ,
 306.86 ,
 343.206 ,
 383.944 ,
 428.567 ,
 477.27 ,
 532.756 ,
 595.994 ,
 662.907 ,
 744.279 ,
 826.469 ,
 925.739 ,
 1037.976 ,
 1157.436 ,
 1286.109 ,
 1444.581 ,
 1599.024 ,
 1801.31 ,
 1991.348 ,
 2226.744 ,
 2493.686 ,
 2809.988 ,
 3118.842 ,
 3511.455 ,
 3904.726 ,
 4367.895 ,
 4857.024 ,
 5430.443 ,
 6046.971 ,
 6772.574 ,
 7557.233 ,
 8478.328 ,
 9479.216 ,
 10581.12 ,
 11788.31 ,
 13181.31 ,
 14694.19 ,
 16427.42 ,
 18418.48 ,
 20582.93 ,
 22993.46 ,
 25696.49 ,
 25696.51 ,
 25696.51 ,
 25696.53 ,
 25696.53 ,
 25696.54 },
{ 0.1 ,
 0.1 ,
 0.101 ,
 2.837 ,
 3.514 ,
 3.667 ,
 4.289 ,
 5.094 ,
 6.035 ,
 6.272 ,
 6.688 ,
 7.506 ,
 8.15 ,
 8.936 ,
 8.976 ,
 9.644 ,
 10.427 ,
 11.44 ,
 11.714 ,
 14.47 ,
 15.235 ,
 18.023 ,
 19.576 ,
 21.704 ,
 24.164 ,
 27.167 ,
 30.091 ,
 33.54 ,
 37.63 ,
 42.13 ,
 46.918 ,
 52.173 ,
 57.869 ,
 65.224 ,
 72.672 ,
 80.329 ,
 90.117 ,
 100.832 ,
 112.343 ,
 125.568 ,
 140.405 ,
 156.959 ,
 175.149 ,
 195.92 ,
 219.344 ,
 245.557 ,
 273.999 ,
 306.892 ,
 343.966 ,
 384.286 ,
 427.672 ,
 477.878 ,
 534.068 ,
 596.761 ,
 665.382 ,
 744.885 ,
 828.421 ,
 926.272 ,
 1034.637 ,
 1156.21 ,
 1294.149 ,
 1443.959 ,
 1614.25 ,
 1803.843 ,
 2008.837 ,
 2249.012 ,
 2515.748 ,
 2815.087 ,
 3134.684 ,
 3500.298 ,
 3916.025 ,
 4376.876 ,
 4880.32 ,
 5454.518 ,
 6102.091 ,
 6806.238 ,
 7630.422 ,
 8506.525 ,
 9490.093 ,
 10633.22 ,
 11860.05 ,
 13243.46 ,
 14795.1 ,
 16527.31 ,
 18466.2 ,
 20669.05 ,
 23068.08 ,
 25729.56 ,
 25729.57 ,
 25729.58 ,
 25729.59 ,
 25729.6 ,
 25729.61 }};
 
double  LM2_UPPER_ENERGY_TABLE[4][93] = {{ 0.1 ,
 1.961 ,
 2.021 ,
 2.581 ,
 2.583 ,
 3.276 ,
 3.37 ,
 4.095 ,
 4.353 ,
 5.148 ,
 5.761 ,
 6.703 ,
 7.397 ,
 8.321 ,
 8.591 ,
 8.772 ,
 10.157 ,
 11.033 ,
 12.319 ,
 14.452 ,
 15.543 ,
 17.23 ,
 19.127 ,
 21.602 ,
 24.462 ,
 26.422 ,
 29.684 ,
 32.736 ,
 37.863 ,
 41.22 ,
 46.636 ,
 51.814 ,
 57.469 ,
 65.023 ,
 72.567 ,
 81.684 ,
 91.657 ,
 102.177 ,
 113.386 ,
 126.461 ,
 141.257 ,
 158.369 ,
 177.518 ,
 197.635 ,
 221.086 ,
 248.217 ,
 276.334 ,
 308.384 ,
 344.41 ,
 386.368 ,
 430.741 ,
 479.847 ,
 537.43 ,
 601.266 ,
 673.551 ,
 754.268 ,
 840.468 ,
 940.296 ,
 1050.997 ,
 1169.39 ,
 1304.191 ,
 1463.504 ,
 1637.308 ,
 1825.38 ,
 2037.289 ,
 2283.398 ,
 2551.688 ,
 2852.128 ,
 3178.061 ,
 3542.79 ,
 3990.048 ,
 4439.191 ,
 4951.838 ,
 5538.84 ,
 6173.844 ,
 6873.49 ,
 7708.366 ,
 8616.582 ,
 9613.396 ,
 10739.55 ,
 12022.41 ,
 13379.21 ,
 15017.67 ,
 16774.28 ,
 18276.72 ,
 20779.99 ,
 23903.85 ,
 27682.07 ,
 27682.08 ,
 27682.09 ,
 27682.1 ,
 27682.11 ,
 27682.12 },
{ 0.1 ,
 2.157 ,
 2.246 ,
 2.655 ,
 2.809 ,
 3.386 ,
 3.448 ,
 4.298 ,
 4.512 ,
 5.336 ,
 5.755 ,
 6.624 ,
 6.861 ,
 7.901 ,
 8.008 ,
 8.453 ,
 9.356 ,
 10.599 ,
 11.653 ,
 13.416 ,
 14.98 ,
 16.791 ,
 18.519 ,
 20.945 ,
 23.28 ,
 25.829 ,
 28.84 ,
 32.364 ,
 36.61 ,
 39.065 ,
 44.001 ,
 49.655 ,
 56.348 ,
 63.926 ,
 70.124 ,
 77.805 ,
 87.026 ,
 97.472 ,
 109.464 ,
 123.331 ,
 139.41 ,
 153.63 ,
 172.161 ,
 192.596 ,
 217.149 ,
 242.503 ,
 272.108 ,
 305.552 ,
 342.051 ,
 381.501 ,
 426.647 ,
 476.22 ,
 532.845 ,
 595.768 ,
 666.754 ,
 747.229 ,
 827.857 ,
 930.153 ,
 1037.073 ,
 1156.801 ,
 1297.136 ,
 1442.921 ,
 1624.093 ,
 1809.938 ,
 2019.47 ,
 2269.644 ,
 2539.623 ,
 2831.75 ,
 3172.097 ,
 3523.901 ,
 3939.532 ,
 4396.614 ,
 4888.769 ,
 5479.709 ,
 6118.495 ,
 6812.605 ,
 7608.644 ,
 8566.398 ,
 9557.921 ,
 10651.17 ,
 11925.62 ,
 13286.67 ,
 14869.78 ,
 16610.07 ,
 18541.99 ,
 20730 ,
 23203.3 ,
 25819.9 ,
 25819.91 ,
 25819.92 ,
 25819.93 ,
 25819.94 ,
 25819.95 },
{ 0.1 ,
 3.068 ,
 3.44 ,
 3.464 ,
 3.915 ,
 3.948 ,
 4.477 ,
 4.529 ,
 5.248 ,
 5.523 ,
 6.151 ,
 6.511 ,
 7.301 ,
 7.973 ,
 9.314 ,
 9.966 ,
 11.36 ,
 12.017 ,
 13.543 ,
 14.692 ,
 16.486 ,
 17.864 ,
 19.919 ,
 21.779 ,
 24.459 ,
 26.693 ,
 29.902 ,
 33.245 ,
 37.001 ,
 40.848 ,
 45.723 ,
 51.305 ,
 57.847 ,
 64.631 ,
 72.223 ,
 80.892 ,
 90.94 ,
 101.619 ,
 113.742 ,
 126.493 ,
 142.296 ,
 158.786 ,
 177.82 ,
 198.694 ,
 222.566 ,
 248.046 ,
 277.771 ,
 310.024 ,
 346.974 ,
 388.332 ,
 433.2 ,
 481.661 ,
 538.802 ,
 601.85 ,
 673.824 ,
 750.249 ,
 837.484 ,
 934.01 ,
 1044.169 ,
 1166.317 ,
 1297.299 ,
 1450.058 ,
 1623.593 ,
 1797.342 ,
 1999.645 ,
 2239.353 ,
 2502.73 ,
 2792.863 ,
 3112.687 ,
 3473.137 ,
 3867.947 ,
 4281.959 ,
 4726.461 ,
 5280.814 ,
 5891.518 ,
 6490.217 ,
 7120.688 ,
 7997.798 ,
 8937.179 ,
 9992.163 ,
 11173.74 ,
 12473.07 ,
 13929.34 ,
 15553.26 ,
 17380.7 ,
 19440.55 ,
 21744.04 ,
 24222.02 ,
 24222.03 ,
 24222.04 ,
 24222.05 ,
 24222.06 ,
 24222.07 },
{ 0.1 ,
 3.257 ,
 3.588 ,
 3.666 ,
 4.145 ,
 4.533 ,
 4.964 ,
 5.679 ,
 6.041 ,
 6.645 ,
 7.211 ,
 7.756 ,
 8.336 ,
 8.965 ,
 9.398 ,
 9.784 ,
 10.671 ,
 11.76 ,
 13.1 ,
 14.656 ,
 16.384 ,
 18.135 ,
 20.104 ,
 22.305 ,
 24.719 ,
 27.269 ,
 30.54 ,
 34.343 ,
 38.673 ,
 42.057 ,
 47.953 ,
 53.556 ,
 60.49 ,
 67.594 ,
 75.349 ,
 84.03 ,
 93.862 ,
 104.438 ,
 116.972 ,
 129.878 ,
 145.152 ,
 161.825 ,
 180.953 ,
 201.159 ,
 225.672 ,
 251.553 ,
 281.813 ,
 314.061 ,
 350.282 ,
 391.143 ,
 437.271 ,
 486.978 ,
 542.869 ,
 605.969 ,
 678.277 ,
 757.43 ,
 845.581 ,
 945.251 ,
 1057.727 ,
 1183.898 ,
 1318.135 ,
 1477.292 ,
 1653.151 ,
 1844.262 ,
 2063.598 ,
 2304.574 ,
 2575.172 ,
 2879.57 ,
 3211.281 ,
 3583.296 ,
 4012.531 ,
 4486.886 ,
 5003.237 ,
 5602.161 ,
 6252.175 ,
 6977.28 ,
 7796.741 ,
 8712.124 ,
 9738.852 ,
 10879.01 ,
 12132.39 ,
 13542.93 ,
 15148.25 ,
 16914.5 ,
 18896.68 ,
 21042.45 ,
 23846.98 ,
 26838.63 ,
 26838.64 ,
 26838.65 ,
 26838.66 ,
 26838.67 ,
 26838.68 }};
 
double  LM3_UPPER_ENERGY_TABLE[4][93] = {{ 0.1 ,
 0.1 ,
 2.9 ,
 3.468 ,
 3.543 ,
 3.62 ,
 4.014 ,
 4.975 ,
 5.046 ,
 5.858 ,
 6.093 ,
 7.331 ,
 7.412 ,
 8.022 ,
 8.663 ,
 8.716 ,
 10.084 ,
 10.303 ,
 12.712 ,
 14.704 ,
 16.169 ,
 17.874 ,
 19.785 ,
 21.65 ,
 24.124 ,
 26.593 ,
 29.801 ,
 33.99 ,
 38.44 ,
 42.212 ,
 46.636 ,
 51.814 ,
 57.469 ,
 65.023 ,
 72.567 ,
 81.684 ,
 91.657 ,
 102.177 ,
 113.386 ,
 126.461 ,
 141.257 ,
 158.369 ,
 177.518 ,
 197.635 ,
 221.086 ,
 248.217 ,
 276.334 ,
 308.384 ,
 344.41 ,
 386.368 ,
 430.741 ,
 479.847 ,
 537.43 ,
 601.266 ,
 673.551 ,
 754.268 ,
 840.468 ,
 940.296 ,
 1050.997 ,
 1169.39 ,
 1304.191 ,
 1463.504 ,
 1637.308 ,
 1825.38 ,
 2037.289 ,
 2283.398 ,
 2551.688 ,
 2852.128 ,
 3178.061 ,
 3542.79 ,
 3990.048 ,
 4439.191 ,
 4951.838 ,
 5538.84 ,
 6173.844 ,
 6873.49 ,
 7708.366 ,
 8616.582 ,
 9613.396 ,
 10739.55 ,
 12022.41 ,
 13379.21 ,
 15017.67 ,
 16774.28 ,
 18276.72 ,
 20779.99 ,
 23903.85 ,
 27682.07 ,
 27682.08 ,
 27682.09 ,
 27682.1 ,
 27682.11 ,
 27682.12 },
{ 0.1 ,
 0.1 ,
 3.318 ,
 3.503 ,
 3.849 ,
 3.893 ,
 4.395 ,
 4.839 ,
 5.3 ,
 5.645 ,
 6.217 ,
 6.758 ,
 7.319 ,
 7.736 ,
 8.531 ,
 9.007 ,
 10.145 ,
 10.879 ,
 12.576 ,
 14.043 ,
 15.772 ,
 17.373 ,
 19.445 ,
 21.853 ,
 24.282 ,
 26.704 ,
 29.986 ,
 33.464 ,
 37.742 ,
 42.3 ,
 47.27 ,
 51.313 ,
 58.056 ,
 63.926 ,
 70.124 ,
 77.805 ,
 87.026 ,
 97.472 ,
 109.464 ,
 123.331 ,
 139.41 ,
 153.63 ,
 172.161 ,
 192.596 ,
 217.149 ,
 242.503 ,
 272.108 ,
 305.552 ,
 342.051 ,
 381.501 ,
 426.647 ,
 476.22 ,
 532.845 ,
 595.768 ,
 666.754 ,
 747.229 ,
 827.857 ,
 930.153 ,
 1037.073 ,
 1156.801 ,
 1297.136 ,
 1442.921 ,
 1624.093 ,
 1809.938 ,
 2019.47 ,
 2269.644 ,
 2539.623 ,
 2831.75 ,
 3172.097 ,
 3523.901 ,
 3939.532 ,
 4396.614 ,
 4888.769 ,
 5479.709 ,
 6118.495 ,
 6812.605 ,
 7608.644 ,
 8566.398 ,
 9557.921 ,
 10651.17 ,
 11925.62 ,
 13286.67 ,
 14869.78 ,
 16610.07 ,
 18541.99 ,
 20730 ,
 23203.3 ,
 25819.9 ,
 25819.91 ,
 25819.92 ,
 25819.93 ,
 25819.94 ,
 25819.95 },
{ 0.1 ,
 0.1 ,
 3.44 ,
 3.464 ,
 3.915 ,
 3.948 ,
 4.477 ,
 4.529 ,
 5.248 ,
 5.523 ,
 6.151 ,
 6.511 ,
 7.301 ,
 7.973 ,
 9.314 ,
 9.966 ,
 11.36 ,
 12.017 ,
 13.543 ,
 14.692 ,
 16.486 ,
 17.864 ,
 19.919 ,
 21.779 ,
 24.459 ,
 26.693 ,
 29.902 ,
 33.245 ,
 37.001 ,
 40.848 ,
 45.723 ,
 51.305 ,
 57.847 ,
 64.631 ,
 72.223 ,
 80.892 ,
 90.94 ,
 101.619 ,
 113.742 ,
 126.493 ,
 142.296 ,
 158.786 ,
 177.82 ,
 198.694 ,
 222.566 ,
 248.046 ,
 277.771 ,
 310.024 ,
 346.974 ,
 388.332 ,
 433.2 ,
 481.661 ,
 538.802 ,
 601.85 ,
 673.824 ,
 750.249 ,
 837.484 ,
 934.01 ,
 1044.169 ,
 1166.317 ,
 1297.299 ,
 1450.058 ,
 1623.593 ,
 1797.342 ,
 1999.645 ,
 2239.353 ,
 2502.73 ,
 2792.863 ,
 3112.687 ,
 3473.137 ,
 3867.947 ,
 4281.959 ,
 4726.461 ,
 5280.814 ,
 5891.518 ,
 6490.217 ,
 7120.688 ,
 7997.798 ,
 8937.179 ,
 9992.163 ,
 11173.74 ,
 12473.07 ,
 13929.34 ,
 15553.26 ,
 17380.7 ,
 19440.55 ,
 21744.04 ,
 24222.02 ,
 24222.03 ,
 24222.04 ,
 24222.05 ,
 24222.06 ,
 24222.07 },
{ 0.1 ,
 0.1 ,
 3.545 ,
 3.778 ,
 4.11 ,
 4.12 ,
 4.877 ,
 5.202 ,
 6.003 ,
 6.448 ,
 7.285 ,
 7.533 ,
 8.359 ,
 8.614 ,
 9.318 ,
 9.693 ,
 10.698 ,
 11.456 ,
 13.228 ,
 14.6 ,
 16.602 ,
 18.103 ,
 20.49 ,
 23.308 ,
 25.337 ,
 27.521 ,
 31.031 ,
 34.759 ,
 39.013 ,
 42.811 ,
 47.953 ,
 53.125 ,
 60.139 ,
 66.792 ,
 74.197 ,
 83.163 ,
 93.217 ,
 103.605 ,
 116.102 ,
 129.078 ,
 144.483 ,
 161.004 ,
 180.014 ,
 200.603 ,
 225.778 ,
 251.645 ,
 282.307 ,
 314.698 ,
 351.859 ,
 393.268 ,
 438.921 ,
 488.877 ,
 546.663 ,
 610.554 ,
 683.512 ,
 760.478 ,
 847.538 ,
 948.915 ,
 1060.566 ,
 1183.961 ,
 1320.804 ,
 1477.731 ,
 1653.196 ,
 1844.209 ,
 2064.883 ,
 2304.668 ,
 2580.801 ,
 2881.79 ,
 3213.828 ,
 3588.598 ,
 4022.145 ,
 4489.837 ,
 4997.905 ,
 5611.482 ,
 6254.412 ,
 6986.804 ,
 7781.023 ,
 8696.277 ,
 9750.665 ,
 10887.16 ,
 12153.86 ,
 13524.58 ,
 15162.61 ,
 16943.27 ,
 18930.72 ,
 21101.84 ,
 23997.6 ,
 26942.62 ,
 26942.63 ,
 26942.64 ,
 26942.65 ,
 26942.66 ,
 26942.67 }};

double  HM1_UPPER_ENERGY_TABLE[4][93] = {{ 0.1 ,
 0.1 ,
 0.101 ,
 1.137 ,
 1.551 ,
 2.434 ,
 3.059 ,
 3.233 ,
 3.961 ,
 4.448 ,
 4.819 ,
 5.586 ,
 6.418 ,
 6.571 ,
 7.23 ,
 7.871 ,
 8.243 ,
 9.714 ,
 10.305 ,
 12.53 ,
 13.426 ,
 15.794 ,
 17.384 ,
 19.984 ,
 22.309 ,
 25.282 ,
 27.785 ,
 31.492 ,
 35.727 ,
 39.599 ,
 44.272 ,
 49.582 ,
 55.357 ,
 62.523 ,
 69.704 ,
 78.484 ,
 88.627 ,
 98.761 ,
 110.239 ,
 123.265 ,
 137.95 ,
 154.438 ,
 172.611 ,
 193.619 ,
 216.14 ,
 241.74 ,
 270.475 ,
 303.652 ,
 338.384 ,
 377.699 ,
 423.18 ,
 472.753 ,
 527.002 ,
 588.496 ,
 659.981 ,
 737.175 ,
 822.461 ,
 916.478 ,
 1027.459 ,
 1146.699 ,
 1280.042 ,
 1429.855 ,
 1599.688 ,
 1790.054 ,
 1997.056 ,
 2235.024 ,
 2496.854 ,
 2789.58 ,
 3115.311 ,
 3473.251 ,
 3889.14 ,
 4348.213 ,
 4846.94 ,
 5420.532 ,
 6046.565 ,
 6770.292 ,
 7554.947 ,
 8463.781 ,
 9426.256 ,
 10529.85 ,
 11798.24 ,
 13141.84 ,
 14716.69 ,
 16395.41 ,
 18264.21 ,
 20436.68 ,
 22813.88 ,
 25503.91 ,
 25503.92 ,
 25503.93 ,
 25503.94 ,
 25503.95 ,
 25503.96 },
{ 0.1 ,
 0.1 ,
 0.101 ,
 2.223 ,
 2.752 ,
 3.349 ,
 3.83 ,
 4.443 ,
 4.989 ,
 5.624 ,
 6.027 ,
 6.861 ,
 7.331 ,
 7.93 ,
 8.355 ,
 9.109 ,
 9.525 ,
 10.689 ,
 11.477 ,
 13.79 ,
 14.398 ,
 16.978 ,
 18.022 ,
 20.327 ,
 21.917 ,
 24.459 ,
 28.533 ,
 31.185 ,
 35.768 ,
 40.072 ,
 45.489 ,
 49.689 ,
 54.83 ,
 61.522 ,
 69.194 ,
 77.343 ,
 86.128 ,
 96.96 ,
 107.619 ,
 121.906 ,
 134.964 ,
 155.293 ,
 174.521 ,
 193.15 ,
 215.649 ,
 241.922 ,
 270.901 ,
 301.131 ,
 336.649 ,
 378.112 ,
 421.124 ,
 467.542 ,
 521.121 ,
 581.572 ,
 651.571 ,
 726.875 ,
 811.729 ,
 906.563 ,
 1016.994 ,
 1131.734 ,
 1266.954 ,
 1416.499 ,
 1586.404 ,
 1769.62 ,
 1974.116 ,
 2205.149 ,
 2466.923 ,
 2761.835 ,
 3080.704 ,
 3436.451 ,
 3840.028 ,
 4302.321 ,
 4804.416 ,
 5374.37 ,
 5996.503 ,
 6683.876 ,
 7458.308 ,
 8350.431 ,
 9318.166 ,
 10429.5 ,
 11648.83 ,
 13017.33 ,
 14570.49 ,
 16265.16 ,
 18179.41 ,
 20357.41 ,
 22755.23 ,
 25436.02 ,
 25436.03 ,
 25436.04 ,
 25436.05 ,
 25436.06 ,
 25436.07 },
{ 0.1 ,
 0.1 ,
 0.101 ,
 2.247 ,
 2.553 ,
 3.176 ,
 3.71 ,
 4.25 ,
 4.78 ,
 5.479 ,
 6.079 ,
 6.62 ,
 7.121 ,
 7.728 ,
 8.211 ,
 8.757 ,
 9.202 ,
 10.483 ,
 11.587 ,
 13.648 ,
 14.303 ,
 17.099 ,
 18.714 ,
 21.153 ,
 23.505 ,
 26.379 ,
 29.26 ,
 32.778 ,
 36.809 ,
 40.666 ,
 45.713 ,
 50.864 ,
 56.545 ,
 63.797 ,
 71.095 ,
 79.057 ,
 89.606 ,
 100.054 ,
 112.155 ,
 125.272 ,
 140.014 ,
 156.536 ,
 174.827 ,
 194.711 ,
 217.167 ,
 242.246 ,
 271.504 ,
 303.289 ,
 339.112 ,
 378.027 ,
 423.074 ,
 472.068 ,
 527.849 ,
 590.428 ,
 657.27 ,
 737.154 ,
 821.18 ,
 920.72 ,
 1024.847 ,
 1142.369 ,
 1287.135 ,
 1424.504 ,
 1594.952 ,
 1781.161 ,
 1989.938 ,
 2227.723 ,
 2492.255 ,
 2779.266 ,
 3096.318 ,
 3457.207 ,
 3842.206 ,
 4311.168 ,
 4830.087 ,
 5410.813 ,
 6050.347 ,
 6744.177 ,
 7550.6 ,
 8486.155 ,
 9429.671 ,
 10599.08 ,
 11831.18 ,
 13212.26 ,
 14771.63 ,
 16623.55 ,
 18416.13 ,
 20575.73 ,
 22991.45 ,
 25618.25 ,
 25618.26 ,
 25618.27 ,
 25618.28 ,
 25618.29 ,
 25618.3 },
{ 0.1 ,
 0.1 ,
 0.101 ,
 2.107 ,
 2.819 ,
 3.254 ,
 3.765 ,
 4.515 ,
 4.959 ,
 5.412 ,
 5.969 ,
 6.507 ,
 7.252 ,
 7.735 ,
 8.219 ,
 8.666 ,
 9.26 ,
 10.441 ,
 11.787 ,
 13.769 ,
 14.642 ,
 17.017 ,
 18.513 ,
 20.942 ,
 23.169 ,
 26.113 ,
 28.742 ,
 32.153 ,
 36.094 ,
 40.018 ,
 44.962 ,
 50.856 ,
 56.233 ,
 63.659 ,
 70.165 ,
 78.214 ,
 88.581 ,
 98.318 ,
 109.328 ,
 123.892 ,
 139.13 ,
 155.655 ,
 173.546 ,
 193.635 ,
 215.997 ,
 241.988 ,
 270.502 ,
 302.776 ,
 337.51 ,
 377.49 ,
 421.895 ,
 471.769 ,
 524.325 ,
 584.922 ,
 654.715 ,
 730.286 ,
 814.744 ,
 910.256 ,
 1015.437 ,
 1135.157 ,
 1268.781 ,
 1412.154 ,
 1577.305 ,
 1764.062 ,
 1969.218 ,
 2188.332 ,
 2445.167 ,
 2733.953 ,
 3059.981 ,
 3415.027 ,
 3819.815 ,
 4272.717 ,
 4784.406 ,
 5359.418 ,
 5983.38 ,
 6676.256 ,
 7483.987 ,
 8356.659 ,
 9326.602 ,
 10443.36 ,
 11712.09 ,
 13025.54 ,
 14597.96 ,
 16348.1 ,
 18268.67 ,
 20433.04 ,
 22828.1 ,
 25470.06 ,
 25470.07 ,
 25470.08 ,
 25470.09 ,
 25470.1 ,
 25470.11 }};
 
double  HM2_UPPER_ENERGY_TABLE[4][93] = {{ 0.1 ,
 1.186 ,
 1.239 ,
 1.701 ,
 1.949 ,
 2.196 ,
 2.417 ,
 3.045 ,
 3.529 ,
 4.089 ,
 4.38 ,
 5.084 ,
 5.632 ,
 6.365 ,
 6.427 ,
 6.823 ,
 7.72 ,
 8.845 ,
 9.829 ,
 11.452 ,
 13.187 ,
 15.163 ,
 16.719 ,
 18.972 ,
 21.432 ,
 24.245 ,
 26.916 ,
 30.429 ,
 35.329 ,
 38.81 ,
 45.089 ,
 49.953 ,
 56.821 ,
 63.073 ,
 70.767 ,
 79.799 ,
 89.832 ,
 100.248 ,
 112.491 ,
 125.218 ,
 140.447 ,
 156.469 ,
 175.186 ,
 195.015 ,
 218.592 ,
 244.008 ,
 273.746 ,
 305.629 ,
 341.495 ,
 381.648 ,
 426.831 ,
 476.403 ,
 531.976 ,
 594.483 ,
 665.344 ,
 744.857 ,
 830.303 ,
 927.918 ,
 1035.821 ,
 1164.028 ,
 1295.773 ,
 1453.101 ,
 1624.604 ,
 1813.512 ,
 2024.206 ,
 2260.872 ,
 2529.131 ,
 2829.581 ,
 3159.374 ,
 3527.15 ,
 3946.052 ,
 4424.411 ,
 4932.217 ,
 5504.494 ,
 6146.428 ,
 6880.849 ,
 7705.032 ,
 8616.858 ,
 9613.37 ,
 10737.91 ,
 11997.06 ,
 13392 ,
 14971.23 ,
 16732.6 ,
 18571.45 ,
 20691.73 ,
 23708.77 ,
 26776.64 ,
 26776.65 ,
 26776.66 ,
 26776.67 ,
 26776.68 ,
 26776.69 },
{ 0.1 ,
 2.124 ,
 2.211 ,
 2.614 ,
 2.765 ,
 3.333 ,
 3.394 ,
 4.231 ,
 4.442 ,
 5.253 ,
 5.665 ,
 6.521 ,
 6.753 ,
 7.778 ,
 7.883 ,
 8.321 ,
 9.21 ,
 10.434 ,
 11.471 ,
 13.206 ,
 14.746 ,
 16.529 ,
 18.23 ,
 20.618 ,
 22.917 ,
 25.425 ,
 28.39 ,
 31.858 ,
 36.655 ,
 40.092 ,
 44.885 ,
 50.862 ,
 57.506 ,
 63.959 ,
 71.256 ,
 79.842 ,
 89.379 ,
 99.969 ,
 110.463 ,
 122.504 ,
 137.383 ,
 153.839 ,
 170.94 ,
 191.378 ,
 214.261 ,
 238.388 ,
 268.507 ,
 299.23 ,
 335.128 ,
 375.702 ,
 417.966 ,
 465.661 ,
 519.609 ,
 580.466 ,
 652.427 ,
 732.085 ,
 812.593 ,
 906.212 ,
 1019.44 ,
 1131.549 ,
 1271.327 ,
 1415.858 ,
 1595.356 ,
 1783.138 ,
 1984.155 ,
 2230.016 ,
 2489.921 ,
 2781.133 ,
 3114.377 ,
 3470.778 ,
 3881.435 ,
 4316.557 ,
 4834.884 ,
 5413.005 ,
 6017.351 ,
 6687.664 ,
 7495.449 ,
 8376.337 ,
 9405.591 ,
 10498.25 ,
 11748.96 ,
 13106.93 ,
 14677.41 ,
 16395.28 ,
 18319.16 ,
 20506.23 ,
 22962.5 ,
 25556.3 ,
 25556.31 ,
 25556.32 ,
 25556.33 ,
 25556.34 ,
 25556.35 },
{ 0.1 ,
 3.266 ,
 3.868 ,
 3.869 ,
 4.318 ,
 4.406 ,
 5.006 ,
 5.265 ,
 5.948 ,
 6.086 ,
 6.826 ,
 6.868 ,
 7.503 ,
 7.604 ,
 8.361 ,
 8.745 ,
 10.182 ,
 11.024 ,
 12.518 ,
 13.858 ,
 15.925 ,
 17.265 ,
 19.774 ,
 22.492 ,
 25.284 ,
 27.757 ,
 31.24 ,
 33.617 ,
 37.82 ,
 41.836 ,
 47.122 ,
 52.01 ,
 58.777 ,
 65.264 ,
 74.175 ,
 82.429 ,
 93.167 ,
 103.779 ,
 116.061 ,
 128.722 ,
 144.261 ,
 160.64 ,
 179.322 ,
 199.673 ,
 223.165 ,
 248.79 ,
 278.779 ,
 310.498 ,
 347.104 ,
 388.289 ,
 433.311 ,
 481.601 ,
 537.851 ,
 600.86 ,
 671.562 ,
 748.344 ,
 833.831 ,
 929.124 ,
 1040.29 ,
 1161.399 ,
 1295.252 ,
 1446.611 ,
 1613.969 ,
 1789.894 ,
 1996.648 ,
 2229.691 ,
 2491.068 ,
 2778.369 ,
 3090.201 ,
 3442.921 ,
 3859.307 ,
 4252.832 ,
 4675.555 ,
 5225.86 ,
 5847.025 ,
 6506.126 ,
 7211.853 ,
 8108.927 ,
 9056.352 ,
 10115.98 ,
 11305.42 ,
 12620.57 ,
 14101.9 ,
 15737.5 ,
 17563.65 ,
 19689.92 ,
 22012.93 ,
 24424.34 ,
 24424.35 ,
 24424.36 ,
 24424.37 ,
 24424.38 ,
 24424.39 },
{ 0.1 ,
 3.231 ,
 3.826 ,
 3.827 ,
 4.271 ,
 4.359 ,
 4.952 ,
 5.208 ,
 5.883 ,
 6.02 ,
 6.752 ,
 6.794 ,
 7.422 ,
 7.522 ,
 8.271 ,
 8.65 ,
 10.072 ,
 10.905 ,
 12.383 ,
 13.708 ,
 15.753 ,
 17.078 ,
 19.561 ,
 22.249 ,
 25.011 ,
 27.458 ,
 30.903 ,
 33.254 ,
 37.412 ,
 41.384 ,
 46.614 ,
 51.449 ,
 58.143 ,
 64.56 ,
 73.375 ,
 81.54 ,
 92.162 ,
 102.66 ,
 114.809 ,
 127.334 ,
 142.705 ,
 158.907 ,
 177.387 ,
 197.519 ,
 220.758 ,
 246.107 ,
 275.771 ,
 307.148 ,
 343.36 ,
 384.101 ,
 428.636 ,
 476.405 ,
 532.049 ,
 594.378 ,
 664.317 ,
 740.271 ,
 824.836 ,
 919.101 ,
 1029.068 ,
 1148.87 ,
 1281.28 ,
 1431.006 ,
 1596.558 ,
 1770.586 ,
 1975.109 ,
 2205.638 ,
 2464.196 ,
 2748.397 ,
 3056.865 ,
 3405.781 ,
 3817.675 ,
 4206.955 ,
 4625.117 ,
 5169.486 ,
 5783.95 ,
 6435.941 ,
 7134.055 ,
 8021.452 ,
 8958.656 ,
 10006.85 ,
 11183.46 ,
 12484.43 ,
 13949.78 ,
 15567.73 ,
 17374.18 ,
 19477.51 ,
 21775.47 ,
 24160.86 ,
 24160.87 ,
 24160.88 ,
 24160.89 ,
 24160.9 ,
 24160.91 }};

double  HM3_UPPER_ENERGY_TABLE[4][93] = {{ 0.1 ,
 0.1 ,
 2.993 ,
 3.168 ,
 3.361 ,
 3.381 ,
 3.913 ,
 4.193 ,
 4.363 ,
 4.552 ,
 5.162 ,
 5.709 ,
 6.321 ,
 6.516 ,
 7.315 ,
 7.997 ,
 9.172 ,
 9.799 ,
 11.24 ,
 12.687 ,
 14.405 ,
 15.774 ,
 18.092 ,
 20.188 ,
 22.858 ,
 25.108 ,
 28.728 ,
 32.139 ,
 36.25 ,
 39.977 ,
 45.089 ,
 50.144 ,
 56.969 ,
 63.352 ,
 70.926 ,
 79.568 ,
 89.513 ,
 99.82 ,
 111.838 ,
 124.719 ,
 140.21 ,
 156.253 ,
 174.909 ,
 194.922 ,
 218.913 ,
 244.213 ,
 273.996 ,
 306.001 ,
 341.938 ,
 381.695 ,
 427.608 ,
 476.602 ,
 532.58 ,
 594.672 ,
 667.852 ,
 746.472 ,
 835.684 ,
 935.035 ,
 1046.043 ,
 1167.494 ,
 1301.159 ,
 1456.914 ,
 1627.651 ,
 1821.631 ,
 2034.525 ,
 2275.262 ,
 2542.799 ,
 2841.988 ,
 3168.831 ,
 3540.141 ,
 3958.888 ,
 4428.466 ,
 4944.116 ,
 5520.153 ,
 6171.95 ,
 6888.565 ,
 7688.366 ,
 8596.479 ,
 9598.986 ,
 10727.66 ,
 11985.25 ,
 13396.4 ,
 14985.92 ,
 16760.76 ,
 18795.28 ,
 20856.26 ,
 23686.99 ,
 26700.78 ,
 26700.79 ,
 26700.8 ,
 26700.81 ,
 26700.82 ,
 26700.83 },
{ 0.1 ,
 0.1 ,
 3.266 ,
 3.448 ,
 3.789 ,
 3.833 ,
 4.326 ,
 4.764 ,
 5.217 ,
 5.557 ,
 6.12 ,
 6.652 ,
 7.204 ,
 7.615 ,
 8.398 ,
 8.867 ,
 9.987 ,
 10.71 ,
 12.38 ,
 13.824 ,
 15.526 ,
 17.599 ,
 19.352 ,
 22.242 ,
 24.905 ,
 27.484 ,
 29.607 ,
 33.546 ,
 37.182 ,
 40.868 ,
 44.885 ,
 51.173 ,
 58.089 ,
 63.001 ,
 70.216 ,
 80.121 ,
 89.953 ,
 98.702 ,
 109.238 ,
 122.263 ,
 136.694 ,
 153.558 ,
 171.338 ,
 194.169 ,
 215.101 ,
 239.779 ,
 268.518 ,
 299.507 ,
 335.475 ,
 374.95 ,
 419.015 ,
 469.832 ,
 524.955 ,
 586.991 ,
 656.241 ,
 733.225 ,
 817.53 ,
 914.271 ,
 1023.828 ,
 1145.929 ,
 1280.916 ,
 1430.318 ,
 1595.356 ,
 1783.138 ,
 1984.155 ,
 2230.016 ,
 2489.921 ,
 2781.133 ,
 3114.377 ,
 3470.778 ,
 3881.435 ,
 4316.557 ,
 4834.884 ,
 5413.005 ,
 6017.351 ,
 6687.664 ,
 7495.449 ,
 8376.337 ,
 9405.591 ,
 10498.25 ,
 11748.96 ,
 13106.93 ,
 14677.41 ,
 16395.28 ,
 18319.16 ,
 20506.23 ,
 22962.5 ,
 25556.3 ,
 25556.31 ,
 25556.32 ,
 25556.33 ,
 25556.34 ,
 25556.35 },
{ 0.1 ,
 0.1 ,
 3.868 ,
 3.869 ,
 4.318 ,
 4.406 ,
 5.006 ,
 5.265 ,
 5.948 ,
 6.086 ,
 6.826 ,
 6.868 ,
 7.503 ,
 7.604 ,
 8.361 ,
 8.745 ,
 10.182 ,
 11.024 ,
 12.518 ,
 13.858 ,
 15.925 ,
 17.265 ,
 19.774 ,
 22.492 ,
 25.284 ,
 27.757 ,
 31.24 ,
 33.617 ,
 37.82 ,
 41.836 ,
 47.122 ,
 52.01 ,
 58.777 ,
 65.264 ,
 74.175 ,
 82.429 ,
 93.167 ,
 103.779 ,
 116.061 ,
 128.722 ,
 144.261 ,
 160.64 ,
 179.322 ,
 199.673 ,
 223.165 ,
 248.79 ,
 278.779 ,
 310.498 ,
 347.104 ,
 388.289 ,
 433.311 ,
 481.601 ,
 537.851 ,
 600.86 ,
 671.562 ,
 748.344 ,
 833.831 ,
 929.124 ,
 1040.29 ,
 1161.399 ,
 1295.252 ,
 1446.611 ,
 1613.969 ,
 1789.894 ,
 1996.648 ,
 2229.691 ,
 2491.068 ,
 2778.369 ,
 3090.201 ,
 3442.921 ,
 3859.307 ,
 4252.832 ,
 4675.555 ,
 5225.86 ,
 5847.025 ,
 6506.126 ,
 7211.853 ,
 8108.927 ,
 9056.352 ,
 10115.98 ,
 11305.42 ,
 12620.57 ,
 14101.9 ,
 15737.5 ,
 17563.65 ,
 19689.92 ,
 22012.93 ,
 24424.34 ,
 24424.35 ,
 24424.36 ,
 24424.37 ,
 24424.38 ,
 24424.39 },
{ 0.1 ,
 0.1 ,
 3.826 ,
 3.827 ,
 4.271 ,
 4.359 ,
 4.952 ,
 5.208 ,
 5.883 ,
 6.02 ,
 6.752 ,
 6.794 ,
 7.422 ,
 7.522 ,
 8.271 ,
 8.65 ,
 10.072 ,
 10.905 ,
 12.383 ,
 13.708 ,
 15.753 ,
 17.078 ,
 19.561 ,
 22.249 ,
 25.011 ,
 27.458 ,
 30.903 ,
 33.254 ,
 37.412 ,
 41.384 ,
 46.614 ,
 51.449 ,
 58.143 ,
 64.56 ,
 73.375 ,
 81.54 ,
 92.162 ,
 102.66 ,
 114.809 ,
 127.334 ,
 142.705 ,
 158.907 ,
 177.387 ,
 197.519 ,
 220.758 ,
 246.107 ,
 275.771 ,
 307.148 ,
 343.36 ,
 384.101 ,
 428.636 ,
 476.405 ,
 532.049 ,
 594.378 ,
 664.317 ,
 740.271 ,
 824.836 ,
 919.101 ,
 1029.068 ,
 1148.87 ,
 1281.28 ,
 1431.006 ,
 1596.558 ,
 1770.586 ,
 1975.109 ,
 2205.638 ,
 2464.196 ,
 2748.397 ,
 3056.865 ,
 3405.781 ,
 3817.675 ,
 4206.955 ,
 4625.117 ,
 5169.486 ,
 5783.95 ,
 6435.941 ,
 7134.055 ,
 8021.452 ,
 8958.656 ,
 10006.85 ,
 11183.46 ,
 12484.43 ,
 13949.78 ,
 15567.73 ,
 17374.18 ,
 19477.51 ,
 21775.47 ,
 24160.86 ,
 24160.87 ,
 24160.88 ,
 24160.89 ,
 24160.9 ,
 24160.91 }};


double lower_energy_lookup(int lookupindex, int mode, int craft, int LEEA_or_HEEA)
{

int SC = craft-1; 

if(LEEA_or_HEEA == 0)
{
if(mode==1)
  return LM1_LOWER_ENERGY_TABLE[SC][lookupindex];
else if(mode==2)
  return LM2_LOWER_ENERGY_TABLE[SC][lookupindex];
else
  return LM3_LOWER_ENERGY_TABLE[SC][lookupindex];
}
else
{
if(mode==1)
  return HM1_LOWER_ENERGY_TABLE[SC][lookupindex];
else if(mode==2)
  return HM2_LOWER_ENERGY_TABLE[SC][lookupindex];
else
  return HM3_LOWER_ENERGY_TABLE[SC][lookupindex];
}
  
}

double upper_energy_lookup(int lookupindex, int mode, int craft, int LEEA_or_HEEA)
{

int SC = craft-1; 

if(LEEA_or_HEEA == 0)
{
if(mode==1)
  return LM1_UPPER_ENERGY_TABLE[SC][lookupindex];
else if(mode==2)
  return LM2_UPPER_ENERGY_TABLE[SC][lookupindex];
else
  return LM3_UPPER_ENERGY_TABLE[SC][lookupindex];
}
else
{
if(mode==1)
  return HM1_UPPER_ENERGY_TABLE[SC][lookupindex];
else if(mode==2)
  return HM2_UPPER_ENERGY_TABLE[SC][lookupindex];
else
  return HM3_UPPER_ENERGY_TABLE[SC][lookupindex];
}

}



void set_Nsc(TRUNC_STRUCT *tstr,  double N){ tstr->Nsc = N; }                         /* set measured density      m**-3          */
void set_vl(TRUNC_STRUCT *tstr, double E_l){ tstr->vl = ev2m_per_s(E_l); }            /* set lower integration limit  eV->m/s     */
void set_vu(TRUNC_STRUCT *tstr, double E_u){ tstr->vu = ev2m_per_s(E_u); }            /* set upper integration limit  eV->m/s 	  */
void set_phi_sc(TRUNC_STRUCT *tstr, double U){ tstr->phi_sc = 1 - U; }                /* set spacecraft potential (correction) V  */
void set_vxsc(TRUNC_STRUCT *tstr, double vx){ tstr->vxsc = vx; }                      /* set components of velocity x m/s         */
void set_vysc(TRUNC_STRUCT *tstr, double vy){ tstr->vysc = vy; }  	              /*             y                m/s         */
void set_vzsc(TRUNC_STRUCT *tstr, double vz){ tstr->vzsc = vz; }                      /*             z                m/s         */
void set_Psc(TRUNC_STRUCT *tstr, double Psr[3][3]){
int i, j;
for (i=0; i<3; i++){
	for (j=0; j<3; j++){tstr->Psc[i][j]=Psr[i][j];}};
		tstr->trPsc = trace(tstr->Psc); } 			              /* set pressure tensor & trace     Pa        */
void set_Vsc(TRUNC_STRUCT *tstr){ tstr->Vsc = sqrt(tstr->vxsc*tstr->vxsc + tstr->vysc*tstr->vysc + tstr->vzsc*tstr->vzsc);}  													      /* set space craft velocity (magnitude) m/s  */
void set_VTsc(TRUNC_STRUCT *tstr){ tstr->VTsc = sqrt(2.0/3.0*((tstr->trPsc/(tstr->Nsc*M_ELECTRON))-pow(tstr->Vsc, 2)));}
										      /* set thermal velocity m/s        	  */


/********************************************************/
void check_physics(TRUNC_STRUCT *tstr){
/********************************************************/
/* try to teach the code some physics - only warns the user though */

/* T */

if(tstr->T0 <= 0)
	printf("\n***Physics warning: zero or negative temperature\n");

/* n */

if(tstr->n0 <= 0)
	printf("\n***Physics warning: zero or negative density\n");

}


/********************************************************/
TRUNC_ERR check_p_diagonal(TRUNC_STRUCT *tstr){
/********************************************************/
/* check if any of p's diagonals are -ve - if so there are problems! */

double p00 = get_p_component(tstr, 0, 0);
double p11 = get_p_component(tstr, 1, 1);
double p22 = get_p_component(tstr, 2, 2);

if((p00 < 0) || (p11 < 0) || (p22 < 0)){return TRUNC_NEG_P_DIAG;}

return TRUNC_OK;

}

/********************************************************/
void print_tensor(double T[3][3], int output){
/********************************************************/

if(output==0 /*screen*/){ printf("%g\t%g\t%g \n %g\t%g\t%g \n %g\t%g\t%g\n",T[0][0],T[0][1],T[0][2],T[1][0],T[1][1],T[1][2],T[2][0],T[2][1],T[2][2]);};

if(output==1 /*comma delim*/){ printf("%g,%g,%g,%g,%g,%g,%g,%g,%g",T[0][0],T[0][1],T[0][2],T[1][0],T[1][1],T[1][2],T[2][0],T[2][1],T[2][2]);};

if(output > 1)exit(0);

}


/********************************************************/
double get_p_component(TRUNC_STRUCT *tstr, int i, int j){
/********************************************************/
/*
* returns the ijth component of the p (small p) tensor,
* given a trunc stuct and a component ij.
*
* benefit of this routine is the fact that only the trunc_struct is
* needed, and can be considered an access method. could come in
* useful.
*
* [e.g. i = 0, j = 0 : 00 or i = 2, j = 1 : 21 or i = 1, j = 1 : 11 etc.]
*/

double p[3][3], v[3], Nsc;

Nsc = tstr->Nsc;
v[0] = tstr->vxsc; v[1] = tstr->vysc; v[2] = tstr->vzsc;

p[i][j] = tstr->Psc[i][j] - Nsc*M_ELECTRON*v[i]*v[j];

return p[i][j];

}

/************************************************************************************************************/
double estimate_anisotropy_T0_para(TRUNC_STRUCT *tstr, double b_gse[3], double P_gse[3][3], double v_gse[3]){
/************************************************************************************************************/
/*
* estimates the anisotropy and returns the parallel component of T0
*
* note -- arguments are in gse coordinates, so remember to use transform routines below
* to process P, v before passing to this function. Or just keep track of sc and gse tensors
*/

int i, j;
double p[3][3];
double a[3];     /* intermediate vector */
double bmag;
double b_unit[3];
double Asc=1, T0, T0_para, trcp, p_para, Nsc;

bmag = sqrt(b_gse[0]*b_gse[0] + b_gse[1]*b_gse[1] + b_gse[2]*b_gse[2]);

for(i=0; i<3; i++)
  b_unit[i] = b_gse[i]/bmag;

Nsc = tstr->Nsc;
T0  = tstr->T0;

/* calculate p */
for (i = 0; i<3; i++){
	for (j=0; j<3; j++){p[i][j] = (P_gse[i][j] - Nsc*M_ELECTRON*v_gse[i]*v_gse[j]);}};

//print_tensor(p,0);

/* do the first part of the contraction */
a[0] = p[0][0]*b_unit[0] + p[0][1]*b_unit[1] + p[0][2]*b_unit[2];
a[1] = p[1][0]*b_unit[0] + p[1][1]*b_unit[1] + p[1][2]*b_unit[2];
a[2] = p[2][0]*b_unit[0] + p[2][1]*b_unit[1] + p[2][2]*b_unit[2];

/* do the second part of the contraction */
p_para = b_unit[0]*a[0] + b_unit[1]*a[1] + b_unit[2]*a[2];

trcp = trace(p);

Asc = (trcp - p_para)/(2*p_para);

T0_para = 3*T0/(1+2*Asc);

return T0_para;
}

/************************************************************************************************************/
double estimate_anisotropy_T0_perp(TRUNC_STRUCT *tstr, double b_gse[3], double P_gse[3][3], double v_gse[3]){
/************************************************************************************************************/
/*
* estimates the anisotropy and returns the perpendicular component of T0
*
* note -- arguments are in gse coordinates, so remember to use transform routines below
* to process P, v before passing to this function. Or just keep track of sc and gse tensors
*/

int i, j;
double p[3][3];
double a[3];     // intermediate vector 
double bmag;
double b_unit[3];
double Asc=1, T0, T0_para, T0_perp, trcp, p_para, Nsc;

bmag = sqrt(b_gse[0]*b_gse[0] + b_gse[1]*b_gse[1] + b_gse[2]*b_gse[2]);

for(i=0; i<3; i++)
  b_unit[i] = b_gse[i]/bmag;

Nsc = tstr->Nsc;
T0 = tstr->T0;

// calculate p 
for(i = 0; i<3; i++){ 
	for(j=0; j<3; j++){p[i][j] = (P_gse[i][j] - Nsc*M_ELECTRON*v_gse[i]*v_gse[j]);}};
                                                    
// do the first part of the contraction             
a[0] = p[0][0]*b_unit[0] + p[0][1]*b_unit[1] + p[0][2]*b_unit[2];
a[1] = p[1][0]*b_unit[0] + p[1][1]*b_unit[1] + p[1][2]*b_unit[2]; 
a[2] = p[2][0]*b_unit[0] + p[2][1]*b_unit[1] + p[2][2]*b_unit[2];

// do the second part of the contraction 
p_para = b_unit[0]*a[0] + b_unit[1]*a[1] + b_unit[2]*a[2];

trcp = trace(p);

Asc = (trcp - p_para)/(2*p_para);

T0_para = 3*T0/(1+2*Asc);

T0_perp = Asc*T0_para;

return T0_perp;
}

/****************************/
double trace(double P[3][3]){
/****************************/
/* returns the trace of the pressure (or any 3x3) tensor */
double trace;
trace = (P[0][0] + P[1][1] + P[2][2]);
return trace;
}

/*******************************************************************************/
void transform_to_gse_tensor(double latitude, double longitude, double sunaz,  double Q[3][3]){
/*******************************************************************************/
/*
* transforms the 3x3-tensor X to a new GSE coord system with the matrix R
* (which has been set with the function sc_to_gse -- see below)
* NB- the original tensor is not preserved.
*/

int i, j, k;
double R[3][3]; 	  /* transformation array */	
double Y[3][3], Z[3][3];  /* intermediate arrays */
double trnsR[3][3];       /* transpose of R */

initialize_tsr(R, 0.0);   /* initialize transformation array */
initialize_tsr(Y, 0.0);
initialize_tsr(Z, 0.0);
initialize_tsr(trnsR, 0.0);

sc_to_gse(R, latitude, longitude, sunaz, 0); /* create R based on lat, long input */

/* get transpose */
trnsR[0][0] = R[0][0];	trnsR[0][1] = R[1][0];	trnsR[0][2] = R[2][0];
trnsR[1][0] = R[0][1];	trnsR[1][1] = R[1][1];	trnsR[1][2] = R[2][1];
trnsR[2][0] = R[0][2];	trnsR[2][1] = R[1][2];	trnsR[2][2] = R[2][2];

/* do the first part of the transformation */
for (i = 0; i<3; i++){ 
	for (j=0; j<3; j++){ Z[i][j] = 0.0; 
		for (k=0; k<3; k++){ Z[i][j] += Q[i][k]*trnsR[k][j];}}}

/* do the second part of the transformation */
for (i = 0; i<3; i++){ 
	for (j=0; j<3; j++){ Y[i][j] = 0.0; 
		for (k=0; k<3; k++){ Y[i][j] += R[i][k]*Z[k][j];}}}

/* reassign */
Q[0][0] = Y[0][0];	Q[0][1] = Y[0][1];	Q[0][2] = Y[0][2];
Q[1][0] = Y[1][0];	Q[1][1] = Y[1][1];	Q[1][2] = Y[1][2];
Q[2][0] = Y[2][0];	Q[2][1] = Y[2][1];	Q[2][2] = Y[2][2];

}

/****************************************************************************/
void transform_to_gse_vector(double latitude, double longitude, double sunaz, double Q[3]){
/****************************************************************************/
/*
* transforms the 3-vector X to a new GSE coord system with the matrix R
* (which has been set with the function sc_to_gse -- see below)
* NB- the original vector is not preserved.
*/

double R[3][3]; 	        /* transformation array */
double Y[3] = {0.0, 0.0, 0.0};  /* intermediate array */

initialize_tsr(R, 0.0);         /* initialize array */

sc_to_gse(R, latitude, longitude, sunaz, 0); /* create R based on lat, long inputs */

/* do the transformation */
Y[0] = R[0][0]*Q[0] + R[0][1]*Q[1] + R[0][2]*Q[2]; 
Y[1] = R[1][0]*Q[0] + R[1][1]*Q[1] + R[1][2]*Q[2];
Y[2] = R[2][0]*Q[0] + R[2][1]*Q[1] + R[2][2]*Q[2];

/* reassign */
Q[0] = Y[0];  Q[1] = Y[1];  Q[2] = Y[2];

}

/**********************************************/
void initialize_tsr(double TR[3][3], double x){
/**********************************************/
/*
* given an uninitialized (double) 3x3 matrix, 
* initialize all it's components to x
*/
int i, j;
for(i=0; i<3; i++)
	for(j=0; j<3; j++)
		TR[i][j] = x;
}

/*****************************************************************************/
void sc_to_gse(double TR[3][3], double latitude, double longitude, double sunaz, int output){
/*****************************************************************************/  
/* adapted from TR_peace_sc2gse.c
*
*  by...
*
*  Steve Schwartz
*  s.j.schwartz@qmul.ac.uk
*  September 2002
*
*  up-dated July 2007 to take account of different sun-pulse offsets in sunaz
*
*  Calculate rotation matrix to go from PEACE spacecraft coords to gse
*  to transform velocity (and other higher-order moments) from on-board
*  ones returned in MOMENTS Virtual Instrument
*
*  Arguments:  spacecraft spin axis LATITUDE and LONGITUDE
*  Output (to stdout): cef file containing rotation matrix in C-order
*
*  Algorithm:
*              
*  Rgse = T(axis) Rot(sun) Rsc       
*       
*  where       
*  Rgse   vector in gse      
*  T      transformation matrix which depends on spin axis orientation      
*  Rot    rotation matrix to deal with offset between Yb and PEACE 0 deg az      
*  Rsc    vector in spacecraft body coords (Xb, Yb, Zb)      
*      
*  Parameters (all angles in degrees)
*
*  sunaz  22.4 Yb->sun position at PEACE 0 degree AZ     
*
*  sunaz = 26.2 - SPOS * 360/(16/1024) 
*            SPOS is sun pulse offset from CORE CPCDM, measured as defined above
*            Early in mission, SPOS was 176, which gives sunaz = 22.4 (or so). Later
*             it changed to 1200, which makes sunaz nearly zero (slightly negative)
*              
*  lambda argument LATITUDE of spin axis (e.g., from CSDS SP AUX files)     
*  phi argument LONGITUDE of spin axis     
*       
*  Derived parameters
*     
*  beta  Normalisation  1/sqrt ( sin^2 lambda + cos^2 lambda sin^2 phi )   
*
*  Matrices    
*       
*  Rot
*     1   0          0
*     0   cossunaz   sinsunaz
*     0  -sinsunaz   cossunaz
*             
*  T 
*     coslambda cosphi   beta[sin^2 lambda + cos^2 lambda sin^2 phi]      0
*     coslambda sinphi  -beta cos^2 lambda cosphi sinphi         	  beta sinlambda
*     sinlambda         -beta coslambda sinlambda cosphi     	         -beta coslambda sinphi
*
*  if output = 1 then output cef format data to stdout, else do not. 
*
*/

double TRDEG2RAD = 0.0174532925199432958;
// double sunaz = 22.4;  /*  Ybody->spin position at PEACE 0 deg AZ */ // now an argument
double lambda, phi, lambda_rad, phi_rad, sunaz_rad;
double RobGowen[3][3], Rot[3][3], T[3][3], intermediate[3][3], Rot2[3][3], RotBy90Deg[3][3];
double coslam, sinlam, cosphi, sinphi, beta, rg_angle;
int i, j, k;

lambda = latitude;
phi    = longitude;

/************************/
/*sunaz = 22.4; */ /* TESTING ONLY */
/*****************************/

lambda_rad = TRDEG2RAD * lambda;
phi_rad    = TRDEG2RAD * phi;
sunaz_rad  = TRDEG2RAD * sunaz;

rg_angle = TRDEG2RAD * 60;

/* **** Rotate to spacecraft coordinates **** */
        /* **** Py into Sy, 60 degrees clockwise about Pz **** */


//	First rotation: from CPCMOM (RAW) frame - a 60 degree rotation about Pz -- this is	
//	the Rob Gowen step	 

RobGowen[0][0] = cos(rg_angle);      RobGowen[0][1] = -sin(rg_angle);              RobGowen[0][2] = 0.0;
RobGowen[1][0] = sin(rg_angle);      RobGowen[1][1] = cos(rg_angle);               RobGowen[1][2] = 0.0;
RobGowen[2][0] = 0.0;                RobGowen[2][1] = 0.0;                         RobGowen[2][2] = 1.0;

/* **** Px into Sx, 90 degrees clockwise about Py(Sy)  **** */

//	Next, do a flip
//

RotBy90Deg[0][0] = 0.0;              RotBy90Deg[0][1] = 0.0;                       RotBy90Deg[0][2] = 1.0;
RotBy90Deg[1][0] = 0.0; 	           RotBy90Deg[1][1] = 1.0;                       RotBy90Deg[1][2] = 0.0;
RotBy90Deg[2][0] = -1.0;             RotBy90Deg[2][1] = 0.0;  	    	             RotBy90Deg[2][2] = 0.0;

//	make these into one matrix (intermediate)

for (i = 0; i<3; i++){
	 for (j=0; j<3; j++){ intermediate[i][j] = 0.0;
	  	for (k=0; k<3; k++){ intermediate[i][j] += RotBy90Deg[i][k]*RobGowen[k][j];}}}

/* **** Rotate from body coordinates to align Yb with  PEACE 0 deg AZ **** */

//	now we are in body coords, do the 22.4 degree rotation to align Yb with 0 deg AZ

Rot[0][0] = 1.0; Rot[0][1] = 0.0;              Rot[0][2] = 0.0;
Rot[1][0] = 0.0; Rot[1][1] = cos(sunaz_rad);   Rot[1][2] = sin(sunaz_rad);
Rot[2][0] = 0.0; Rot[2][1] = -sin(sunaz_rad);  Rot[2][2] = cos(sunaz_rad);

//	condense into a single matrix -- we are building up a matrix which will do the whole
//	transformation in the correct order

for (i = 0; i<3; i++){
	for (j=0; j<3; j++){Rot2[i][j] = 0.0;
		for (k=0; k<3; k++){ Rot2[i][j] += Rot[i][k]*intermediate[k][j];}}}

coslam = cos(lambda_rad);
sinlam = sin(lambda_rad);
cosphi = cos(phi_rad);
sinphi = sin(phi_rad);

beta = 1.0/(sqrt(sinlam*sinlam + coslam*coslam*sinphi*sinphi));

T[0][0] = coslam*cosphi;
T[0][1] = beta*(sinlam*sinlam+coslam*coslam*sinphi*sinphi);
T[0][2] = 0.0;

T[1][0] = coslam*sinphi;
T[1][1] = -beta*coslam*coslam*cosphi*sinphi;
T[1][2] = beta*sinlam;

T[2][0] = sinlam;
T[2][1] = -beta*coslam*sinlam*cosphi;
T[2][2] = -beta*coslam*sinphi;

/* multiply T * Rot to form complete rotation matrix */

//	get the final rotation matrix

for (i = 0; i<3; i++){
	for (j=0; j<3; j++){ TR[i][j] = 0.0;
		for (k=0; k<3; k++){ TR[i][j] += T[i][k]*Rot2[k][j];}}}

if(output!=0)
{
printf("!--------------- Cluster Exchange ASCII File -------------|\n");
printf("!- Tranformation from PEACE spacecraft (Xb,Yb,Zb) to gse  |\n");
printf("!- Row Major Order: TR_00, TR_01, TR_02, TR_10, TR_11, ...|\n");
printf("! V_gse = TR * V_sc                                       |\n");
printf("!                                                         |\n");
printf("! Steve Schwartz: S.J.Schwartz@qmul.ac.uk                 |\n");
printf("! September 2002                                          |\n");
printf("!---------------------------------------------------------|\n");
printf("File_name = T_lambda_phi.cef\n");
printf("Attribute_delimiter = ,\n");
printf("Data_delimiter = ,\n");
printf("End_of_record_marker = \\n\n");
printf("Record_numbering = off\n");
printf("Start_variable = TR\n");
printf("Sizes =  3, 3\n");
printf("Value_type = double\n");
printf("Data = ");

for (i=0; i<3; i++){ printf("   ");
	for(j=0; j<3; j++){ printf(" %g ", TR[i][j]); if ( i*j == 4 ) printf("\n"); else printf(",");}}

printf("FIELDNAM = TR_matrix PEA Body to gse\n");
printf("Frame = rot>sc_xyz\n");
printf("SI_conversion = 1>()\n");
printf("ToFrame = gse\n");
printf("UNITS =  \n");
printf("scat_lat (deg)= %g\n",lambda);
printf("scat_long (deg) = %g\n",phi);
printf("End_variable = TR\n");
printf("\n");
printf("! TR\n");
printf("Start_data = 1 ! Data records to follow\n");
}

}

/******************************************************************************/
void guess(double v[], double guess_one, double guess_two, double guess_three){
/******************************************************************************/
/* assign the guesses v1-3 the values guess_one-three. Note exponent form of guess_one */

double expo;

expo = guess_one; v[1] = pow(10, expo);
v[2] = guess_two;
v[3] = guess_three;

}

/*********************************************************************/
void set_defaults_tstruct(TRUNC_STRUCT *tstr){
/*********************************************************************/
/* initialise things to sensible values; most need to be overwritten */

int i,j;

tstr->n0 = 0.0;    /* Maxwellian vals; needs to be overwritten */
tstr->v0 = 0.0;
tstr->vT = 0.0;
tstr->T0 = 0.0;

/* measured moments and params */
tstr->vl      = 0.0;  tstr->vu      = 0.0;
tstr->phi_sc  = 0.0;
tstr->Nsc     = 0.0;
tstr->vxsc    = 0.0;  tstr->vysc    = 0.0;   tstr->vzsc    = 0.0;
for (i=0; i<3; i++){
   for (j=0; j<3; j++){
     tstr->Psc[i][j]=0.0;
   }
}
tstr->Vsc     = 0.0;
tstr->trPsc   = 0.0;
tstr->VTsc    = 0.0;

/* intermediary quantities */
tstr->curlyE  = 0.0;
tstr->vL      = 0.0;  tstr->vU      = 0.0;

/* normalised measured values */
tstr->zeta_sc = EPS;     /* a small number; zero will cause problems */
tstr->vl_norm = 0.0;     /* limits in measured thermal speeds */
tstr->vu_norm = 20.0;
tstr->eps_sc  = 0.0;

/* normalised Maxwellian moments */
tstr->eta     = ONE_OVER_SQRT_PI;    /* initial guess */
tstr->v0_norm = 1.0;
tstr->vt_norm = 1.0;

return;
}

/*********************************************************************/
void print_trunc_struct(TRUNC_STRUCT *t, TBOOL headings){
/*********************************************************************/
/* print out contents of trunct structure, with column headings if
*  headings == TTRUE */
if(headings==TTRUE){
printf("Truncated Moments Corrector Structure\n");
printf("=====================================\n");
printf("   n0(m3)  v0(m/s)  vT(m/s)  T0(K)  vl(m/s)   vu(m/s)   phi_sc\n");
printf("%8.03g %8.03g %8.03g %8.03g %8.03g %8.03g %8.03g\n",
         t->n0, t->v0, t->vT, t->T0, t->vl, t->vu, t->phi_sc);
printf(" Nsc(m3)  vxsc(m/s)   vysc     vzsc\n");
printf("%8.03g %8.03g %8.03g %8.03g\n",
      t->Nsc, t->vxsc,t->vysc,t->vzsc);
printf("Pxxsc(N/m2)  Pxysc    Pxzsc    Pyxsc    Pyysc    Pyzsc     ");
printf("Pzxsc     Pzysc    Pzzsc\n");
printf("%8.03g %8.03g %8.03g ",
      t->Psc[0][0],t->Psc[0][1],t->Psc[0][2]);
printf("%8.03g %8.03g %8.03g %8.03g %8.03g %8.03g\n",
      t->Psc[1][0],t->Psc[1][1], t->Psc[1][2],
      t->Psc[2][0],t->Psc[2][1],t->Psc[2][2]);

printf("  Vsc(m/s) trPsc   VTsc   curlyE(m/s)^2 vL(m/s)   vU\n");
printf("%8.03g %8.03g %8.03g %8.03g %8.03g %8.03g\n",
         t->Vsc,t->trPsc, t->VTsc, t->curlyE, t->vL, t->vU);
printf("zeta_sc   vl_norm   vu_norm   eps_sc    eta   v0_norm  vt_norm\n");
printf("%8.03g %8.03g %8.03g %8.03g %8.03g %8.03g %8.03g\n",
   t->zeta_sc,t->vl_norm,t->vu_norm, t->eps_sc, t->eta, t->v0_norm,
   t->vt_norm);
} else {
printf("%8.03g %8.03g %8.03g %8.03g %8.03g %8.03g %8.03g ",
         t->n0, t->v0, t->vT, t->T0, t->vl, t->vu, t->phi_sc);
printf("%8.03g %8.03g %8.03g %8.03g %8.03g %8.03g %8.03g ",
      t->Nsc, t->vxsc,t->vysc,t->vzsc,t->Psc[0][0],t->Psc[0][1],t->Psc[0][2]);
printf("%8.03g %8.03g %8.03g %8.03g %8.03g %8.03g %8.03g ",
      t->Psc[1][0],t->Psc[1][1], t->Psc[1][2],
      t->Psc[2][0],t->Psc[2][1],t->Psc[2][2],t->Vsc);
printf("%8.03g %8.03g %8.03g %8.03g %8.03g ",
         t->trPsc, t->VTsc, t->curlyE, t->vL, t->vU);
printf("%8.03g %8.03g %8.03g %8.03g %8.03g %8.03g %8.03g\n",
   t->zeta_sc,t->vl_norm,t->vu_norm, t->eps_sc, t->eta, t->v0_norm,
   t->vt_norm);
}
return ;
}
/*********************************************************************/
TRUNC_ERR normalised_nl_solver(TRUNC_STRUCT *tstruct, double Ga, double Gb, double Gr){
/*********************************************************************/
/*
*  Solve the nonlinear set in normalised quantities:
*
*  Inputs: all in tstrut structure
*  =======
*
*  zeta_sc    magnitude of measured bulk velocity normalised to
*                        measured thermal speed vt_sc, i.e. vz_sc/vt_sc
*  vl_norm    speed in free space arriving at detector at lower cutoff and
*                        normalised to vt_sc = sqrt( vl^2 -2e phi/me )/vt_sc^
*  vu_norm    ditto for upper limit; note ensure vl_norm and vu_norm >0
*                 and > eps_sc  [typically eps_sc <0]
*  eps_sc     negative sc potential in velocity units normalised to vt_sc^2
*
*  Outputs:
*  ========
*
*  eta        (n0/n_sc)/[  (sqrt(pi) (v0/vz_sc) (vt/vt_sc) ]
*  v0_norm    v0/vz_sc
*  vt_norm    vt/vt_sc
*
*  where n0, v0, and vt are the parameters of the underlying Maxwellian
*
*  Return Values:
*  ==============
*    TRUNC_OK                            success
*    TRUNC_TMNEWT_LU_ERR                 tmnewt lu decomp failed
*    TRUNC_TMNEWT_TRIALS_EXHAUSTED       self-explanatory
*    TRUNC_ERROR                         unspecified error
*    TRUNC_GUESSES
*
*/

double *x;
double gl = Ga, gu = Gb, gr = Gr, g1 = GUESS_ONE, g2 = GUESS_TWO, g3 = GUESS_THREE;
double tolx = 1.e-8, tolf = 1.e-8;
double mult = 0, n_guesses;
int n = 3, ntrials = NTRIES, i;
TRUNC_ERR ret_val;

int nr_ret;



n_guesses = (fabs(gl)+fabs(gu))/gr;

//printf("Lower = %g \t Upper = %g \t Resolution = %g\t N = %g \n",gl,gu,gr,n_guesses);

x = dvector(1,3);       			/* NR vector, indexed 1,2,3 */
ret_val = check_nl_inputs(tstruct);
guess(x, g1, g2, g3);  			       //initial guess

switch(ret_val){
   case TRUNC_VLVU_LT_VMIN:
   /*    printf("Warning: %s\n",trunc_err_messages(ret_val)); */ /* fall through */
   case TRUNC_OK:
       nr_ret = tmnewt(ntrials,x,n,tolx,tolf,tstruct);
       switch(nr_ret){
	 case TMNEWT_LU_ERR:
	           ret_val = TRUNC_TMNEWT_LU_ERR;           break;
	 case TMNEWT_TRIALS_EXHAUSTED:
	           ret_val = TRUNC_TMNEWT_TRIALS_EXHAUSTED; break;
	 case NR_OK:
	           ret_val = TRUNC_OK;                      break;
	 default:  ret_val = TRUNC_ERROR;                   break;
       }

       if (ret_val != TRUNC_OK){

	/*  printf("tmnewt returned error %d: %s\n",ret_val,trunc_err_messages(ret_val)); */
	/* loop for repeated guesses until nguesses is exhausted */

	mult = 0.01*gl;	/* reset the multiplicative factor to lower limit */

	/* invoke secondary guessing loop */
	for(i=0; i<=n_guesses; i++)
	{
  		guess(x, g1+mult*g1, g2+mult*g2, g3+mult*g3);	        /* try to guess again */
		nr_ret = tmnewt(ntrials, x, n, tolx, tolf, tstruct);	/* try the solver again */
		/* test if solver returned OK, if it did, then assign values and break the loop */
		if(nr_ret == NR_OK){tstruct->eta = x[1]; tstruct->v0_norm = x[2]; tstruct->vt_norm = sqrt(x[3]);
		                    ret_val = TRUNC_OK_GUESSES; GUESS_ONE=x[1];
				    GUESS_TWO = x[2]; GUESS_THREE = x[3];
				    break;}
		mult += 0.01*gr;		                        /* incremenls
		t the multiplicative factor to the guess */
		/*printf("G1: %g \t G2: %g \t G3: %g \t nr_ret: %d\n",g1+mult*g1, g2+mult*g2, g3+mult*g3, nr_ret);*/
	};


	/* if all else fails... */
	if(nr_ret != NR_OK){
	 			 tstruct->eta     = ONE_OVER_SQRT_PI;
	  			 tstruct->v0_norm = 1.0;
	  			 tstruct->vt_norm = 1.0;
				 ret_val = TRUNC_GUESSES_EXHAUSTED; break;}

       } else { tstruct->eta = x[1];  tstruct->v0_norm = x[2];  tstruct->vt_norm = sqrt(x[3]); ret_val = TRUNC_OK;  break;}
       break;
       default:  printf("  Failure: %s\r",trunc_err_messages(ret_val));
}

free_dvector(x,1,3);
return ret_val;
}  /* end normalised_nl_solver */

/*************************************************************************/
TRUNC_ERR check_nl_inputs(TRUNC_STRUCT *tstruct){
/*************************************************************************/
/* ensure vl_norm, vu_norm > eps_sc and > 0
*  vl and vu calculated from their squares, so should be > 0 but check
*  likewise, eps should always be such that vl[u]_norm^2 > eps
*  i.e., this check should always be a null check by construction
*/

double vl, vu, eps, vmin;
TRUNC_ERR ret_val;

vl = tstruct->vl_norm;
vu = tstruct->vu_norm;
eps = tstruct->eps_sc;

ret_val = TRUNC_OK;

vmin = (eps > 0.0) ? sqrt(eps) : 0.0 ; /* integ must start above this  */
                                        /* typically eps < 0 for +ve sc */
if (vl < vmin){
  /* printf("vl_norm < vmin on input; reset to vmin=%g\n",vmin); */
   tstruct->vl_norm = vmin;
   ret_val = TRUNC_VLVU_LT_VMIN;
}
if (vu < vmin){
  /* printf("vu_norm < vmin on input; reset to vmin=%g\n",vmin); */
   tstruct->vu_norm = vmin;
   ret_val = TRUNC_VLVU_LT_VMIN;
}
if (vl >= vu) {
  /* printf("zero or negative normalised integ. range (%g,%g)\n",vl,vu); */
   ret_val = TRUNC_VLVU_RANGE_ERROR;
}

//ret_val = check_p_diagonal(tstruct);

return ret_val;
}

/*************************************************************************/
const char *trunc_err_messages(TRUNC_ERR terr){  /* see header */
/*************************************************************************/
static const char *terr_str[]={"TMC OK",
			       "TMC Unspec Error",
	                       "TMC LU Decomp Error",
	                       "TMC Newton ntrials Exhausted",
                               "TMC Normalised Integ. Range Error",
                               "TMC Normalised VL, VU < min allowed",
	                       "TMC OK, But Needed Guessing Loop",
			       "TMC Guessing Loop Exhausted",
			       "TMC Negative Diagonal Elements in p",
		               "TMC OK, but didn't use B moments",
			       "TMC Unknown Error code"

};

if(terr<0 || terr>N_TRUNC_ERR -1) return terr_str[N_TRUNC_ERR];
else return terr_str[terr];
}


/*************************************************************************/
void tusrfun(double *x,int n,double *fvec,double **fjac,
                    TRUNC_STRUCT *tstruct){
/*************************************************************************/

/* use separate routines and numerical Jacobian */
tvecfun(n,x,fvec,tstruct);
tfdjac(n,x,fvec,fjac,tvecfun,tstruct);
return;
}


/*************************************************************************/
void tvecfun( int n, double *x, double *fvec, TRUNC_STRUCT  *tstruct){
/*************************************************************************/
/* calculate fvec; x, fvec declared [1-n] as
per NR conventions. Equations are fvec[i]=0
*/

/* x[1] = eta, x[2] = v0_norm, x[3]=vt_norm_sq */

double eta, v0, vt_sq;  		  /* all normalised */
double zeta, vl, vu, epssc;  	/* ditto - see TRUNCT_STRUCT */
double Nd = 1.;
int N = 1; /* number of sections to split integration into */
/* copy variables from input and tstruct into local copies */
zeta   = tstruct->zeta_sc;
vl     = tstruct->vl_norm;
vu     = tstruct->vu_norm;
epssc  = tstruct->eps_sc;
eta    = x[1];
v0     = x[2];
vt_sq  = x[3];

Nd = 0.5*tstruct->vu*tstruct->vu*M_ELECTRON/(1000*Q_ELECTRON); //one chunk per keV
N = ceil(Nd);
//printf("Nd = %g\tN = %d\n",Nd, N);

fvec[1] = 1.0 - (eta/zeta)*split_tqgauss100(N, integrand1, vl, vu, epssc, v0, vt_sq, zeta);
fvec[2] = zeta*zeta -  eta*split_tqgauss100(N, integrand2, vl, vu, epssc, v0, vt_sq, zeta);
fvec[3] = 1.5 + zeta*zeta - (eta/zeta)*split_tqgauss100(N, integrand3, vl, vu, epssc, v0, vt_sq, zeta);

return;
}

/*************************************************************************/
void tfdjac(int n, double x[], double fvec[], double **df,
   void (*tvecfunc)(int, double [], double [], TRUNC_STRUCT *),
   TRUNC_STRUCT  *tstruct)
/*************************************************************************/
/*Computes forward-difference approximation to Jacobian. On input, x[1..n]
is the point at which the Jacobian is to be evaluated, fvec[1..n] is the
vector of function values at the point, and vecfunc(n,x,f) is a user-supplied
routine that returns the vector of functions at x. On output, df[1..n][1..n]
is the Jacobian array. */
{
int i,j;
double h,temp,*f;
f=dvector(1,n);
for (j=1;j<=n;j++) {
   temp=x[j];
   h=EPS*fabs(temp);
   if (h == 0.0) h=EPS;
   x[j]=temp+h;   /* Trick to reduce  finite precision error. */
   h=x[j]-temp;
   (*tvecfunc)(n,x,f,tstruct);
   x[j]=temp;
   for (i=1;i<=n;i++) df[i][j]=(f[i]-fvec[i])/h; /* Forward difference */
}
free_dvector(f,1,n);
}

/*************************************************************************/
double integrand1(double v, double epssc, double v0, double vt_sq, double zeta){
/*************************************************************************/
/*   (v^2 -epssc)^(1/2) (eminus - eplus) dv
*
*  for fvec[1] from truncated density moment
*/
double eplus, eminus;
eplusminusfun(v, v0, zeta, vt_sq, &eplus, &eminus);

return (sqrt(v*v-epssc))*(eminus - eplus);

}

/*************************************************************************/
double integrand2(double v, double epssc, double v0, double vt_sq, double zeta){
/*************************************************************************/
/*   (v^2 -epssc)*[ (eminus + eplus) - (vt^2/2 v v0 zeta)*(eminus-eplus) dv
*
*  for fvec[2] from truncated density moment
*/
double eplus, eminus;
eplusminusfun(v, v0, zeta, vt_sq, &eplus, &eminus);

return (v*v-epssc)*(    (eminus + eplus) -
           ( vt_sq/(2.0 * v * v0 * zeta) )*(eminus - eplus)    );

}

/*************************************************************************/
double integrand3(double v, double epssc, double v0, double vt_sq, double zeta){
/*************************************************************************/
/*   (v^2 -epssc)^(3/2) (eminus - eplus) dv
*
*  for fvec[3] from truncated density moment
*/
double eplus, eminus, temp;
eplusminusfun(v, v0, zeta, vt_sq, &eplus, &eminus);
temp = v*v - epssc;

return (temp*sqrt(temp))*(eminus - eplus);

}

/*************************************************************************/
void eplusminusfun(double v, double v0, double zeta, double vt_sq,
      double *eplus, double *eminus){
/*************************************************************************/
/* eplus[minus] = exp - { (v +[-] v0*zeta_sc)^2 / vt^2  */

double  vzeta, plussq, minussq;

vzeta     = v0*zeta;
plussq    = (v+vzeta)/sqrt(vt_sq);
plussq   *= plussq;
minussq   = (v-vzeta)/sqrt(vt_sq);
minussq  *= minussq;
*eplus    = exp(-plussq);
*eminus   = exp(-minussq);

return;

}

/*************************************************************************/
double ev2m_per_s(double Eev){
/*************************************************************************/
/* convert energy in eV to velocity
*
*   1/2 m v^2 = e Eev   => v = sqrt (2 e Eev / m)
*/

if (Eev<=0.0) return 0.0;   /* error on input */
return sqrt(2.0*Q_ELECTRON*Eev/M_ELECTRON);
}

/*************************************************************************/
void calc_moments_from_maxwellian(TRUNC_STRUCT *tstr){
/*************************************************************************/
/* given Maxwellian n0, v0, vT, and intermediary quantities curlyE, vL, vU
*  calculate Nsc, Vsc, trPsc, VTsc
*/

double xl,xu,estar,v0_vT,zeta,vt_one;

double n0,v0,vT,Nsc,Vsc,trPsc,VTsc;   /* local copies */
double Nd = 1.;
int N = 1;

zeta   = 1.0;            /* normalise v0/vT, then with these values */
vt_one = 1.0;            /* Eplusminus = exp-{x +- v0_vT)^2/ 1^2    */
n0     = tstr->n0;       /* and integrands 1-3 can be re-used here  */
v0     = tstr->v0;
vT     = tstr->vT;
xl     = tstr->vL/vT;
xu     = tstr->vU/vT;
estar  = tstr->curlyE/(vT*vT);
v0_vT  = v0/vT;

Nd = 0.5*tstr->vu*tstr->vu*M_ELECTRON/(1000*Q_ELECTRON); //one chunk per keV
N = ceil(Nd);


Nsc  =  n0*vT*ONE_OVER_SQRT_PI/v0;
Nsc *=  split_tqgauss100(N, integrand1, xl, xu, estar, v0_vT, vt_one, zeta);

Vsc   = n0*ONE_OVER_SQRT_PI*vT*vT/v0;
Vsc  *= split_tqgauss100(N, integrand2, xl, xu, estar, v0_vT, vt_one, zeta);
Vsc  /= Nsc;
trPsc = M_ELECTRON*n0*vT*vT*vT*ONE_OVER_SQRT_PI/v0;
trPsc *= split_tqgauss100(N, integrand3, xl, xu, estar, v0_vT, vt_one, zeta);

VTsc        = sqrt( (2.0/3.0)*( (trPsc/(Nsc*M_ELECTRON)) - Vsc*Vsc ) );
tstr->Nsc   = Nsc;
tstr->Vsc   = Vsc;
tstr->trPsc = trPsc;
tstr->VTsc  = VTsc;

return;
}

/*************************************************************************/
void calc_maxwellian_nl_params(TRUNC_STRUCT *t,
   double *etatilda, double *v0tilda, double *vTtilda){
/*************************************************************************/
/* take values from Maxwellian params and truncated moments
*  in tstr and calculate expected eta,... roots */
*etatilda = ONE_OVER_SQRT_PI*(t->n0/t->Nsc)/(( t->v0/t->Vsc)*(t->vT/t->VTsc));
*v0tilda  = t->v0/t->Vsc;
*vTtilda  = t->vT/t->VTsc;
return;
}

/*************************************************************************/
void calc_intermediary(TRUNC_STRUCT *tstr){
/*************************************************************************/
/* calc curlyE, vL, vU from phi_sc, vl, vu */

double temp, curlyE;

curlyE = -2.0*Q_ELECTRON*tstr->phi_sc/M_ELECTRON;
tstr->curlyE = curlyE;

temp = tstr->vl*tstr->vl +curlyE;
if (temp <0.0) tstr->vL = 0.0; else tstr->vL=sqrt(temp);

temp = tstr->vu*tstr->vu +curlyE;
if (temp <0.0) tstr->vU = 0.0; else tstr->vU=sqrt(temp);
return;
}

/*************************************************************************/
void prepare_normalised_nl_inputs(TRUNC_STRUCT *tstr){
/*************************************************************************/
/* inputs Nsc, Vsc, trPsc, VTsc + curlyE,vL,vU from structure
*
*  outputs zeta_sc,vl_norm,vu_norm, eps_sc
*
*  This is the final preparation before passing to the nonlinear solver */

double VTsc;  /* local copy */

VTsc = tstr->VTsc;
tstr->vl_norm = tstr->vL / VTsc;
tstr->vu_norm = tstr->vU / VTsc;
tstr->eps_sc = tstr->curlyE/(VTsc*VTsc);
tstr->zeta_sc = tstr->Vsc/VTsc;
return;
}

/*************************************************************************/
void dimensional_from_nl_solution(TRUNC_STRUCT *tstr){
/*************************************************************************/
/* take eta, v0_norm, vt_norm and calculate n0, v0, vT, T0 of maxwellian
*  in physical units */

tstr->v0 = tstr->v0_norm*tstr->Vsc;
tstr->vT = tstr->vt_norm*tstr->VTsc;
tstr->n0 = tstr->eta*tstr->v0_norm*tstr->vt_norm*tstr->Nsc/ONE_OVER_SQRT_PI;
tstr->T0 = 0.5*M_ELECTRON*tstr->vT*tstr->vT/K_BOLTZ;
return;
}




